/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <string.h>
#include "color.h"
#include "draw-funcs.h"
#include "dibuja.h"
#include "util.h"
#include "dibuja-colorsel.h"

#define N_BUTTONS (20)
#ifndef MOUSE_RIGHT_BUTTON
	#define  MOUSE_RIGHT_BUTTON	(3)
#endif
#ifndef MOUSE_LEFT_BUTTON
	#define  MOUSE_LEFT_BUTTON	(1)
#endif

typedef struct{
	GdkColor color;
	guint16 alpha;
}Color;

void colorbutton_bg_color_set(GtkColorButton *widget, gpointer user_data);
void colorbutton_fg_color_set(GtkColorButton *widget, gpointer user_data);
gboolean tool_color_button_press_event(GtkWidget *button,
									GdkEventButton *event, gpointer data);


static GtkWidget *color_button_fg;
static GtkWidget *color_button_bg;
const gchar *default_colors[N_BUTTONS] = {"black", "white", "gray50", "red",
								   "purple", "blue", "light blue",
								   "green", "yellow", "orange",
								   "lavender", "brown", "goldenrod4",
								   "dodger blue", "pink", "light green",
								   "gray10", "gray30", "gray75",
								   "gray90"};

static void set_button_color(GtkWidget *button, GdkColor *color)
{
	GdkColor *col;

	col = (GdkColor *)g_object_get_data(G_OBJECT(button), "color");
	if(col){
		*col = *color;
		gtk_widget_queue_draw(button);
	}
}

gboolean buttonexpose(GtkWidget *button, GdkEventExpose *event, gpointer *data)
{
	GdkColor *col;
	guint16 *alpha = NULL;

	if((col = (GdkColor *)g_object_get_data(G_OBJECT(button), "color")) != NULL){
		cairo_t *cr = gdk_cairo_create(button->window);

		gdk_cairo_rectangle(cr, &event->area);
		alpha = (guint16 *)g_object_get_data(G_OBJECT(button), "alpha");
		if(alpha)
			cairo_set_source_rgba(cr, col->red / 65535.,
									  col->green / 65535.,
									  col->blue / 65535.,
									  *alpha / 65535.);
		else
			gdk_cairo_set_source_color(cr, col);

		cairo_fill(cr);

		cairo_destroy(cr);
	}

	return TRUE;
}
#define PALETTE_ENTRY_WIDTH		(25)
#define PALETTE_ENTRY_HEIGHT	(25)
static GtkWidget *create_button(const char *colordesc, char *name)
{
	GtkWidget *darea, *frame = NULL;
	GdkColor *color = NULL;
	guint16 *alpha = NULL;
	char *tooltip = 
					N_("Left click to set this color\n"
					"as the foreground color.\n\n"
					"Right click to set this color\n"
					"as the background color.");

	color = g_malloc(sizeof(GdkColor));
	if(color){
		frame = gtk_frame_new (NULL);
		gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);//GTK_SHADOW_IN);
		gtk_widget_set_size_request (frame, PALETTE_ENTRY_WIDTH, PALETTE_ENTRY_HEIGHT);

		darea = gtk_drawing_area_new();
		gtk_container_add (GTK_CONTAINER (frame), darea);
		gtk_widget_set_events (darea, GDK_BUTTON_PRESS_MASK
                         | GDK_EXPOSURE_MASK);

		g_object_set_data_full(G_OBJECT(darea), "color", (gpointer)color, g_free);

		alpha = g_malloc(sizeof(guint16));
		if(alpha){
			*alpha = 65535;
			g_object_set_data_full(G_OBJECT(darea), "alpha", (gpointer)alpha, g_free);
		}

		gdk_color_parse(colordesc, color);
		set_button_color(darea, color);

		/* Save widgets to call from color selection dlg */
		g_object_set_data(G_OBJECT(dibuja_get_canvas()), name, (gpointer)darea);
		g_signal_connect(G_OBJECT(darea), "expose-event", G_CALLBACK(buttonexpose), NULL);
		g_signal_connect(G_OBJECT(darea), "button-press-event",
							G_CALLBACK(tool_color_button_press_event), NULL);

		gtk_widget_set_tooltip_text(darea, _(tooltip));
	}
	return frame;
}

void color_init(GtkBuilder *builder)
{
	GtkWidget *button, *table;
	char name[40];
	int i = 0;
	int left, top;

	color_button_fg = get_widget("colorbutton_fg");
	color_button_bg = get_widget("colorbutton_bg");

	table = GTK_WIDGET (gtk_builder_get_object (builder, "table_col_pal"));

	gtk_table_resize(GTK_TABLE(table), 12, 2);
	for(top = 2; top < 12; top++){
		for(left = 0; left < 2; left++){
			sprintf(name, "button%d", i);
			button = create_button(default_colors[i], name);
			gtk_table_attach (GTK_TABLE (table), button,
							  left, left + 1,
							  top, top + 1,
							  GTK_FILL, GTK_FILL, 0, 0);
			i++;
		}
	}
}
// TODO: set alpha from palette
gboolean tool_color_button_press_event(GtkWidget *button, GdkEventButton *event, gpointer data)
{
	GdkColor *color = NULL;
	guint16 alpha;

	if(MOUSE_LEFT_BUTTON == event->button){
		color = (GdkColor *)g_object_get_data(G_OBJECT(button), "color");
		if(color){
			alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_fg")));
			dibuja_set_fg_color(color, alpha);
			gtk_color_button_set_color(GTK_COLOR_BUTTON(color_button_fg), color);
		}
	}
	else if(MOUSE_RIGHT_BUTTON == event->button){
		color = (GdkColor *)g_object_get_data(G_OBJECT(button), "color");
		if(color){
			alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_bg")));
			dibuja_set_bg_color(color, alpha);
			gtk_color_button_set_color(GTK_COLOR_BUTTON(color_button_bg), color);
		}
	}
	return TRUE;
}

void colorbutton_fg_color_set(GtkColorButton *widget, gpointer user_data)
{
	GdkColor color;
	guint16 alpha;

	gtk_color_button_get_color(widget, &color);
	alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_fg")));
	dibuja_set_fg_color(&color, alpha);
}

void colorbutton_bg_color_set(GtkColorButton *widget, gpointer user_data)
{
	GdkColor color;
	guint16 alpha;

	gtk_color_button_get_color(widget, &color);
	alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_bg")));
	dibuja_set_bg_color(&color, alpha);
}

gboolean colorbutton_right_click(GtkWidget *colorbutton, GdkEventButton *event, gpointer data)
{
	if(MOUSE_RIGHT_BUTTON == event->button){
		GdkColor fg, bg;
		guint16 alpha;

		gtk_color_button_get_color(GTK_COLOR_BUTTON(color_button_fg), &fg);
		gtk_color_button_get_color(GTK_COLOR_BUTTON(color_button_bg), &bg);

		gtk_color_button_set_color(GTK_COLOR_BUTTON(color_button_fg), &bg);
		alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_bg")));
		dibuja_set_fg_color(&bg, alpha);
		gtk_color_button_set_color(GTK_COLOR_BUTTON(color_button_bg), &fg);
		alpha = gtk_color_button_get_alpha(GTK_COLOR_BUTTON(get_widget("colorbutton_fg")));
		dibuja_set_bg_color(&fg, alpha);
	}
	return FALSE;
}
/* TODO */
/*static int color_edit_dialog_set_palette(gchar *palette)
{
	if(!palette) return 0;

	return 1;
}*/

/* Has default color palette on startup */
void color_edit_dialog(GtkWidget *widget)
{
	gint result;
	static GtkWidget *dialog = NULL;

	if(!GTK_IS_DIALOG(dialog)){
		dialog = dibuja_color_dialog_new(GTK_WINDOW(gtk_widget_get_toplevel(widget)));
	}

	result = gtk_dialog_run(GTK_DIALOG(dialog));

	if(GTK_RESPONSE_ACCEPT == result){
		GdkColor *gdkcolors = NULL;
		int i, ncolors;
		GtkWidget *button;
		char name[40];
		GtkWidget *canvas = dibuja_get_canvas();

		gdkcolors = dibuja_colorsel_get_colors(&ncolors);

		if(gdkcolors){
			for(i = 0; i < ncolors; i++){
				sprintf(name, "button%d", i);
				button = GTK_WIDGET (g_object_get_data(G_OBJECT(canvas), name));
				if(GTK_IS_WIDGET(button)){
					set_button_color(button, &(gdkcolors[i]));
					gtk_widget_queue_draw(button);
				}
			}
			g_free(gdkcolors);
		}

	}
	gtk_widget_hide(dialog);
}



