/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "sel-handle.h"
#include "tools.h"
#include "draw-funcs.h"
#include "dibuja.h"
#include "util.h"
#include "undo.h"

#define WIDTH	(DEF_HANDLE_WIDTH)
#define HEIGHT	(DEF_HANDLE_WIDTH)

static gboolean curve_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean curve_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data);
static gboolean curve_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean curve_expose(GtkWidget *widget, cairo_t *cr, gpointer data);
static void curve_finish(void);
static gboolean curve_cancel(void);

static DrawTool tool = {
				curve_button_press,
				curve_motion,
				curve_button_release,
				curve_expose,
				curve_finish,
				NULL,
				curve_cancel
			};

static double sx = 0;
static double sy = 0;
static double ex = 0;
static double ey = 0;
static GdkRectangle old, cur;
gboolean lock_adj_handles = TRUE;

static SelHandle *active_hdl = NULL;
static SelHandle *hdl_start = NULL;
static SelHandle *hdl_end = NULL;
static SelHandle *hdl_adj1 = NULL;
static SelHandle *hdl_adj2 = NULL;

DrawTool *curve_get_tool(void)
{
	return &tool;
}

void curve_init(GtkWidget *owner)
{
	GdkColor color = {0};
	int x = 10; int y = 40;

	if(GTK_IS_WIDGET(owner)){
		hdl_start = sel_handle_new();
		sel_handle_move (hdl_start, x, y);
		sel_handle_set_color (hdl_start, &color);
		g_object_set_data_full(G_OBJECT(owner), "curve1", (gpointer)hdl_start,
								(GDestroyNotify)g_object_unref);

		hdl_end = sel_handle_new();
		sel_handle_move (hdl_end, x, y);
		sel_handle_set_color (hdl_end, &color);
		g_object_set_data_full(G_OBJECT(owner), "curve2", (gpointer)hdl_end,
								(GDestroyNotify)g_object_unref);

		color.blue = 0xFFFF;

		hdl_adj1 = sel_handle_new();
		sel_handle_move (hdl_adj1, x, y);
		sel_handle_set_color (hdl_adj1, &color);
		sel_handle_set_type(hdl_adj1, SEL_HANDLE_ROUND);
		g_object_set_data_full(G_OBJECT(owner), "curve3", (gpointer)hdl_adj1,
								(GDestroyNotify)g_object_unref);

		hdl_adj2 = sel_handle_new();
		sel_handle_move (hdl_adj2, x, y);
		sel_handle_set_color (hdl_adj2, &color);
		sel_handle_set_type(hdl_adj2, SEL_HANDLE_ROUND);
		g_object_set_data_full(G_OBJECT(owner), "curve4", (gpointer)hdl_adj2,
								(GDestroyNotify)g_object_unref);
	
		g_object_set_data(G_OBJECT(hdl_start), "adj-handle",
								(gpointer)hdl_adj1);
		g_object_set_data(G_OBJECT(hdl_end), "adj-handle",
								(gpointer)hdl_adj2);
	}
}

/* TODO: Set this in preferences */
/*static void set_handles_dims(void)
{
	guint new_dim = WIDTH;

	new_dim = (guint)(MAX((int)((double)new_dim / canvas_get_zoom()), 1));

	sel_handle_set_dims (hdl_start, new_dim, new_dim);
	sel_handle_set_dims (hdl_end, new_dim, new_dim);
	sel_handle_set_dims (hdl_adj1, new_dim, new_dim);
	sel_handle_set_dims (hdl_adj2, new_dim, new_dim);
}*/

static void calc_update_area(GdkRectangle *area)
{
	GdkPoint pt1, pt2, pt3, pt4;
	double line_width = (double)dibuja_get_line_width();

	sel_handle_get_pos (hdl_start, &pt1);
	sel_handle_get_pos (hdl_end, &pt2);
	sel_handle_get_pos (hdl_adj1, &pt3);
	sel_handle_get_pos (hdl_adj2, &pt4);

	rect_from_pts(area, MIN(MIN(pt1.x, pt2.x), MIN(pt3.x, pt4.x)),
					MIN(MIN(pt1.y, pt2.y), MIN(pt3.y, pt4.y)),
					MAX(MAX(pt1.x, pt2.x), MAX(pt3.x, pt4.x)),
					MAX(MAX(pt1.y, pt2.y), MAX(pt3.y, pt4.y)) );

	resize_rect(area, WIDTH + line_width,
					  HEIGHT + line_width);
}

static void draw_handles(cairo_t *cr, double sx, double sy, double ex, double ey)
{
	if(active_hdl){
		GdkPoint pt1, pt2;

		//set_handles_dims();
		cairo_save(cr);
		cairo_set_antialias(cr, CAIRO_ANTIALIAS_DEFAULT);

		cairo_set_source_rgba(cr, 0., 0., 0., 0.5);
		sel_handle_draw (hdl_start, cr);
		sel_handle_draw (hdl_end, cr);
		sel_handle_draw (hdl_adj1, cr);
		sel_handle_draw (hdl_adj2, cr);

		cairo_set_line_width (cr, 1.);
		sel_handle_get_pos (hdl_start, &pt1);
		sel_handle_get_pos (hdl_adj1, &pt2);

		cairo_move_to(cr, (double)pt1.x + WIDTH / 2, (double)pt1.y + HEIGHT / 2);
		cairo_line_to(cr, (double)pt2.x + WIDTH / 2, (double)pt2.y + HEIGHT / 2);
		cairo_stroke(cr);

		sel_handle_get_pos (hdl_end, &pt1);
		sel_handle_get_pos (hdl_adj2, &pt2);

		cairo_set_source_rgba(cr, 0.2, 0.2, 1, 0.5);
		cairo_move_to(cr, (double)pt1.x + WIDTH / 2, (double)pt1.y + HEIGHT / 2);
		cairo_line_to(cr, (double)pt2.x + WIDTH / 2, (double)pt2.y + HEIGHT / 2);
		cairo_stroke(cr);
		cairo_restore(cr);
	}
	else{
		draw_line(cr, sx, sy, ex, ey);
	}
}

static void draw_curve(cairo_t *cr)
{
	int center;
	GdkPoint pt1, pt2, pt3, pt4;

	center = DEF_HANDLE_WIDTH / 2;

	sel_handle_get_pos (hdl_start, &pt1);
	sel_handle_get_pos (hdl_adj1, &pt2);
	sel_handle_get_pos (hdl_adj2, &pt3);
	sel_handle_get_pos (hdl_end, &pt4);

	cairo_move_to (cr, (double) (pt1.x + center), (double) (pt1.y + center));
	cairo_curve_to (cr, (double) (pt2.x + center), (double) (pt2.y + center),
						(double) (pt3.x + center), (double) (pt3.y + center),
						(double) (pt4.x + center), (double) (pt4.y + center) );
}

static void curve_tool_draw(cairo_t *cr, double sx, double sy, double ex, double ey)
{
	if(active_hdl)
		draw_curve(cr);
	cairo_stroke (cr);
	draw_handles(cr, sx, sy, ex, ey);
}
/* Check to see which handle has been clicked */
static SelHandle *curve_pt_in(int x, int y)
{
	if(sel_handle_pt_in (hdl_adj1, x, y))  { return hdl_adj1;  }
	if(sel_handle_pt_in (hdl_adj2, x, y))  { return hdl_adj2;  }
	if(sel_handle_pt_in (hdl_start, x, y)) { return hdl_start; }
	if(sel_handle_pt_in (hdl_end, x, y))   { return hdl_end;   }

	return NULL;
}

static void curve_move_handle(SelHandle *hdl, int x, int y)
{
	SelHandle *adj = NULL;
	int nx, ny;
	GdkPoint diff, pt;

	nx = x - WIDTH / 2;
	ny = y - HEIGHT / 2;

	adj = SEL_HANDLE(g_object_get_data(G_OBJECT(hdl), "adj-handle"));
	if(SEL_IS_HANDLE(adj) && lock_adj_handles){
		sel_handle_get_pos (hdl, &diff);
		sel_handle_get_pos (adj, &pt);
		diff.x = pt.x - diff.x;
		diff.y = pt.y - diff.y;
		sel_handle_move (adj, nx + diff.x, ny + diff.y);
	}
	sel_handle_move (hdl, nx, ny);
}

static gboolean curve_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	SelHandle *tmp = NULL;
	if(1 == event->button){
		sx = ex = event->x;
		sy = ey = event->y;

		tmp = curve_pt_in((int)sx, (int)sy);
		if(SEL_IS_HANDLE(tmp)){/* Clicked one of the handles */
			active_hdl = tmp;
			calc_update_area(&old);
			return TRUE;
		}
		else if(SEL_IS_HANDLE(active_hdl)){/* Clicked outside all the handle */
			curve_finish();	
		}
		active_hdl = NULL;
	}
	old.x = (int)sx;
	old.y = (int)sy;
	old.width = old.height = 0;
	sx = ex = event->x;
	sy = ey = event->y;

	return TRUE;
}

static gboolean curve_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	GdkRectangle area = {0};

	if (event->state & GDK_BUTTON1_MASK){
		ex = (double)event->x;
		ey = (double)event->y;

		if(SEL_IS_HANDLE(active_hdl)){
			curve_move_handle(active_hdl, (int)ex, (int)ey);
			calc_update_area(&cur);
			gdk_rectangle_union(&old, &cur, &area);
			old = cur;

		}
		else{
			int line_width = dibuja_get_line_width();
			if(event->state & GDK_CONTROL_MASK){
				draw_constrain_pt(sx, sy, ex, ey, &ex, &ey);
			}
			rect_from_pts(&cur, (int)sx, (int)sy, (int)ex, (int)ey );
			resize_rect(&cur, line_width * 2, line_width * 2);
			gdk_rectangle_union(&old, &cur, &area);
			old = cur;
		}

		simple_canvas_update_rect(SIMPLE_CANVAS(widget), &area);
	}

	return TRUE;
}

static void curve_finish(void)
{
	GdkRectangle area;
	cairo_t *cr;
	double l, t, r, b;
	int line_width = dibuja_get_line_width();
	int x, y;
	
	if(SEL_IS_HANDLE(active_hdl)){
		cr = dibuja_get_cr(CR_SURFACE);
		draw_curve(cr);

		cairo_path_extents(cr, &l, &t, &r, &b);
		rect_from_pts(&area, (int)l, (int)t, (int)r, (int)b);
		resize_rect(&area, line_width * 2, line_width * 2);
		undo_add(SIMPLE_CANVAS(dibuja_get_canvas()), area, _("Curve"));

		cairo_stroke (cr);
		cairo_destroy(cr);
	}

	/* Move them out of sight */
	x = y = -10;
	if(SEL_IS_HANDLE(active_hdl)){
		curve_move_handle(active_hdl, x, y);
		sel_handle_set_active (active_hdl, FALSE);
	}
	curve_move_handle(hdl_start, x, y);
	curve_move_handle(hdl_end, x, y);
	curve_move_handle(hdl_adj1, x, y);
	curve_move_handle(hdl_adj2, x, y);
	active_hdl = NULL;
	sx = ex = sy = ey = 0;
	gtk_widget_queue_draw(dibuja_get_canvas());
}

static gboolean curve_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(1 == event->button){

		if(event->state & GDK_CONTROL_MASK){
			draw_constrain_pt(sx, sy, ex, ey, &ex, &ey);
		}

		if((sx == ex && ey == sy)) return TRUE;

		if(!SEL_IS_HANDLE(active_hdl)){
			active_hdl = hdl_start;
			curve_move_handle(active_hdl, (int)ex, (int)ey);
			curve_move_handle(hdl_end, (int)sx, (int)sy);
			curve_move_handle(hdl_adj1, (int)ex, (int)ey);
			curve_move_handle(hdl_adj2, (int)sx, (int)sy);
			undo_add_temp(SIMPLE_CANVAS(widget), _("Curve"));
		}
		
		gtk_widget_queue_draw(widget);
	}

	return TRUE;
}

static gboolean curve_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	curve_tool_draw(cr, sx, sy, ex, ey);

	return TRUE;
}

static gboolean curve_cancel(void)
{
	if(SEL_IS_HANDLE(active_hdl)){
		active_hdl = NULL;
		old.x = (int)sx;
		old.y = (int)sy;
		old.width = old.height = 0;
		sx = ex = 0;
		sy = ey = 0;
		return TRUE;
	}

	return FALSE;
}

