/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <glib/gstdio.h>
#include <time.h>

#include "dibuja.h"
#include "util.h"
#include "draw-funcs.h"
#include "color.h"
#include "tools.h"
#include "gdkpixbuf-drawable.h"
#include "selection.h"
#include "file.h"
#include "simple-canvas.h"
#include "undo.h"
#include "rotate.h"
#include "menu.h"
#include "effects.h"
#include "fx-dlg.h"
#include "settings.h"
#include "uni-exiv2.hpp"


#define TARGET_IMG_FILENAME	(0)


static GtkWidget *window, *canvas;
static GtkWidget *statusbar_label, *statusbar_label_dims;
static DrawTool *cur_tool = NULL;
static ShapeFillType fill_mode = SHAPE_FILL_NONE;


static const GtkTargetEntry drop_targets[] = {
    {"text/uri-list", 0, TARGET_IMG_FILENAME},
};


double dibuja_get_zoom(void)
{
	return simple_canvas_get_zoom(SIMPLE_CANVAS(canvas));
}

static void dibuja_show_image_dims(double w, double h)
{
	char *text = NULL;

	text = g_strdup_printf("[%.02f x %.02f] %s", w, h, _("px"));
	gtk_label_set_text(GTK_LABEL(statusbar_label_dims), text);
	g_free(text);
}

void dibuja_show_sel_dims(double w, double h)
{
	dibuja_show_image_dims(w, h);
}

/******************************************************************************/
#ifdef HAVE_EXIV2
static void
add_label_pair(GtkWidget *table, int row, const char *l1, gchar const *l2, ...)
{
	GtkWidget *label;
	gchar *buffer;
	va_list args;

	label = util_create_label("<b>%s</b>", l1);
	gtk_misc_set_alignment(GTK_MISC(label), 0., .5);

	gtk_table_attach(GTK_TABLE(table), label, 0, 1, row, row + 1,
					 GTK_FILL, GTK_FILL, 12, 3);

	va_start(args, l2);
	buffer = g_strdup_vprintf(l2, args);
	va_end (args);

	label = gtk_label_new(buffer);
	g_free(buffer);

	gtk_misc_set_alignment(GTK_MISC(label), 0., .5);
	gtk_table_attach(GTK_TABLE(table), label, 1, 2, row, row + 1,
					 GTK_FILL, GTK_FILL, 12, 3);
}

static void exiv2_callback(const char *label, const char *value, void *data)
{
	if(label && value && GTK_IS_TABLE(data)){
		GtkWidget *table = GTK_WIDGET(data);
		int rows = 0;

		gtk_table_get_size(GTK_TABLE(table), (unsigned int *)&rows, NULL);
		add_label_pair(table, rows, label, value);
	}
}

void dibuja_file_info_dialog(void)
{
	GtkWidget *dialog, *vbox, *table;
	GStatBuf st;
	struct tm *info;
	char *fname = dibuja_get_filename();
	char buffer[80];
	int row = 0, ret, file_status = file_get_status();
	int w, h;
	GdkPixbufFormat *pbf;
	gdouble xres, yres;
	GdkDisplay *disp;

	ret = g_stat(fname, &st);

	if(file_status & FILE_NEW || 0 != ret){
		message_box_ok(GTK_MESSAGE_INFO,
					   _("The information could not be displayed"), NULL);
		return;
	}

	dialog = gtk_dialog_new_with_buttons(_("File Information"),
										 GTK_WINDOW(window),
										 GTK_DIALOG_DESTROY_WITH_PARENT,
										 GTK_STOCK_OK,
										 GTK_RESPONSE_OK,
										 NULL);

	vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	table = gtk_table_new(1, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, TRUE, 24);
	gtk_widget_show(table);

	info = localtime(&st.st_ctime);   /* time of last status change */
	strftime(buffer, sizeof(buffer), "%x %I:%M %p", info);
	add_label_pair(table, row++, _("File last saved"), buffer);

	g_stat(fname, &st);
	add_label_pair(table, row++, _("Size on disk"),
				   "%.01f KB", (double)st.st_size / 1000.);

	disp = gdk_display_get_default();
	gimp_get_screen_resolution(gdk_display_get_default_screen(disp),
							   &xres, &yres);
	add_label_pair(table, row++, _("Screen Resolution"),
				   "%d x %d", (int)xres, (int)yres);

	pbf = gdk_pixbuf_get_file_info((const char *)fname, &w, &h);
	if(pbf){
		add_label_pair(table, row++, _("Width"), "%d px", w);
		add_label_pair(table, row++, _("Height"), "%d px", h);

		char *n = gdk_pixbuf_format_get_name(pbf);
		add_label_pair(table, row++, _("File type"), n);
		g_free(n);
	}

	// Dummy labels for spacing
	add_label_pair(table, row++, "", "");

	uni_read_exiv2_map(fname, exiv2_callback, (void *)table);
	gtk_widget_show_all(table);

	gtk_dialog_run(GTK_DIALOG(dialog));

	gtk_widget_destroy(dialog);
}

#endif // HAVE_EXIV2

/******************************************************************************/
void dibuja_flip(int dir)
{
	                     /* Undo description */
	char *flip_horz = N_("Flip Horizontally");
	                     /* Undo description */
	char *flip_vert = N_("Flip Vertically");
	cairo_surface_t *surf = NULL;

	if(FLIP_HORZ != dir && FLIP_VERT != dir) return;

	surf = draw_surface_flip(dibuja_get_surface(), dir);

	if(surf)
		dibuja_resize(surf, FLIP_HORZ == dir ? _(flip_horz) : _(flip_vert));
}
// angle in degrees
void dibuja_rotate_image(double angle)
{
	cairo_surface_t *surf;

	surf = draw_rotate_image(dibuja_get_surface(), angle);

	if(surf)		/* Undo description */
		dibuja_resize(surf, _("Rotate"));
}

void dibuja_resize_image(gint wnew, gint hnew, gboolean centered)
{
	cairo_surface_t *new_surf;

	if(!(wnew > 0 && hnew > 0)) return ;

	new_surf = draw_resize_image(dibuja_get_surface(), wnew, hnew, centered);

	if(new_surf)			/* Undo description */
		dibuja_resize(new_surf, _("Resize"));
}

void dibuja_scale_image(gint wnew, gint hnew)
{
	cairo_surface_t *new_surf;

	if(!(wnew > 0 && hnew > 0)) return ;

	new_surf = draw_scale_image(dibuja_get_surface(), wnew, hnew);
	if(new_surf)				/* Undo description */
		dibuja_resize(new_surf, _("Scale"));
}

/******************************************************************************/
cairo_surface_t *dibuja_get_surface(void)
{
	return simple_canvas_get_image(SIMPLE_CANVAS(canvas));
}

/* Copy the drawing, or if have selection, copy that */
GdkPixbuf *dibuja_copy_drawing(void)
{
	cairo_surface_t *surf;
	if(sel_have()){
		return sel_copy();
	}

	surf = simple_canvas_get_image(SIMPLE_CANVAS(canvas));
	return gdk_pixbuf_get_from_surface (surf, 0, 0,
										cairo_image_surface_get_width(surf),
										cairo_image_surface_get_height(surf));
}

cairo_surface_t *dibuja_copy_from_surface(GdkRectangle *rect)
{
	if(!rect) return NULL;
	if(!(rect->width > 0) || !(rect->height > 0)) return NULL;

	return draw_copy_area(simple_canvas_get_image(SIMPLE_CANVAS(canvas)),
						  rect->x, rect->y, rect->width, rect->height);
}

GtkWidget *dibuja_get_canvas(void)
{
	return canvas;
}

void dibuja_surface_replace(cairo_surface_t *replacement)
{
	if(replacement){
		/* Destroy old, replace with replacement */
		simple_canvas_set_image(SIMPLE_CANVAS(canvas), replacement);
	}
}

cairo_t *dibuja_get_cr(CR_TYPE type)
{
	cairo_t *cr = NULL;

	switch(type){
		case CR_WIDGET:
			return simple_canvas_get_window_context(SIMPLE_CANVAS(canvas));
		case CR_SURFACE:
			return simple_canvas_get_context(SIMPLE_CANVAS(canvas));
		default:
			break;
	}
	return cr;
}

void dibuja_do_invert(void)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *surface = NULL;
	cairo_surface_t *src_surf = NULL;
	gboolean do_sel = sel_have();


	if(do_sel){
		src_surf = sel_get_cur_image();
	}
	else{
		src_surf = dibuja_get_surface();
	}

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_invert (src_buffer, src_buffer, &rect);

	surface = buffer_to_surface(src_buffer);

	if(do_sel){
		if(!sel_apply_effect(surface)){
			cairo_surface_destroy(surface);
		}
	}
	else						/* Undo description */
		dibuja_resize(surface, _("Invert"));

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_do_pixelize(void)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *surface = NULL;
	cairo_surface_t *src_surf = NULL;

	if(!sel_have()){
		return;
	}

	src_surf = sel_get_cur_image();

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_pixelize (src_buffer, src_buffer, &rect, 15, 15);

	surface = buffer_to_surface(src_buffer);

	if(!sel_apply_effect(surface)){
		cairo_surface_destroy(surface);
	}

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_do_grayscale(void)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *surface = NULL;
	cairo_surface_t *src_surf = NULL;
	gboolean do_sel = sel_have();


	if(do_sel){
		src_surf = sel_get_cur_image();
	}
	else{
		src_surf = dibuja_get_surface();
	}

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_grey (src_buffer, src_buffer, &rect);

	surface = buffer_to_surface(src_buffer);

	if(do_sel){
		if(!sel_apply_effect(surface)){
			cairo_surface_destroy(surface);
		}
	}
	else
							/* Undo description */
		dibuja_resize(surface, _("Grayscale"));

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_do_insta_brannan(char *op)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *src_surf, *surface = NULL;
	gboolean do_sel = sel_have();

	if(!op) return;

	if(do_sel){
		src_surf = sel_get_cur_image();
	}
	else{
		src_surf = dibuja_get_surface();
	}

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_insta_brannan (op, src_buffer, src_buffer, &rect);

	surface = buffer_to_surface(src_buffer);

	if(do_sel){
		if(!sel_apply_effect(surface)){
			cairo_surface_destroy(surface);
		}
	}
	else
							/* Undo description */
		dibuja_resize(surface, _("Insta Brannan"));

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_do_insta_1977(char *op)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *src_surf, *surface = NULL;
	gboolean do_sel = sel_have();

	if(!op) return;

	if(do_sel){
		src_surf = sel_get_cur_image();
	}
	else{
		src_surf = dibuja_get_surface();
	}

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_insta_1977 (op, src_buffer, src_buffer, &rect);

	surface = buffer_to_surface(src_buffer);

	if(do_sel){
		if(!sel_apply_effect(surface)){
			cairo_surface_destroy(surface);
		}
	}
	else
							/* Undo description */
		dibuja_resize(surface, _("Insta 1977"));

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_do_emboss(void)
{
	GeglBuffer *src_buffer;
	GeglRectangle rect = {0};
	cairo_surface_t *surface = NULL;
	cairo_surface_t *src_surf = NULL;
	gboolean do_sel = sel_have();


	if(do_sel){
		src_surf = sel_get_cur_image();
	}
	else{
		src_surf = dibuja_get_surface();
	}

	rect.width = cairo_image_surface_get_width(src_surf);
	rect.height = cairo_image_surface_get_height(src_surf);

	src_buffer = gegl_buffer_from_surface (src_surf);

	if(!src_buffer){
		return;
	}

	fx_apply_emboss (src_buffer, src_buffer, &rect);

	surface = buffer_to_surface(src_buffer);

	if(do_sel){
		if(!sel_apply_effect(surface)){
			cairo_surface_destroy(surface);
		}
	}
	else
							/* Undo description */
		dibuja_resize(surface, _("Emboss"));

	if(src_buffer){
		g_object_unref(src_buffer);
	}

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_run_fx_dilalog(char *op, char *desc)
{
	GtkWidget *dlg, *widget;
	int ret;

	widget = gtk_widget_get_toplevel(dibuja_get_canvas());
	dlg = effects_dialog_new(GTK_WINDOW(widget), op, dibuja_get_surface());
	if(GTK_IS_DIALOG(dlg)){

		ret = gtk_dialog_run(GTK_DIALOG(dlg));

		if(GTK_RESPONSE_APPLY == ret){
			cairo_surface_t *surf;

			surf = fx_dlg_get_cairo_surface(EFFECTS_DIALOG(dlg));
			dibuja_resize(surf, desc);
		}

		gtk_widget_destroy(dlg);
	}
}

static gboolean dibuja_end_take_screenshot(gpointer user_data)
{
	GdkPixbuf *pixbuf, *ret;
	GtkWidget *toplevel;
	GdkWindow *root = gdk_get_default_root_window ();
	int width, height;

	toplevel = gtk_widget_get_toplevel(dibuja_get_canvas());

	gdk_window_get_size(root, &width, &height);
	pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, TRUE, 8, width, height);
	ret = gdk_pixbuf_get_from_drawable(pixbuf, root, 0, 0, 0, 0, 0,
									   width, height);

	gtk_widget_show(toplevel);

	if(GDK_IS_PIXBUF(pixbuf)){
		if(GDK_IS_PIXBUF(ret)){
			tools_toggle_sel_button(TRUE);
			sel_create_from_pixbuf(pixbuf);
		}
		g_object_unref(pixbuf);
	}

	return FALSE;
}

gboolean dibuja_start_take_screenshot(gpointer user_data)
{
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(dibuja_get_canvas());
	gtk_widget_hide(toplevel);
	g_timeout_add_seconds(1, dibuja_end_take_screenshot, NULL);

	return FALSE;
}

int dibuja_set_filename(char *filename, char *format)
{
	int ret = 0;

	if(filename){
		char *p;

		p = g_strdup(filename);
		g_object_set_data_full(G_OBJECT(canvas), "filename", p, g_free);
		ret = 1;
	}
	if(format){
		char *p;

		p = g_strdup(format);
		g_object_set_data_full(G_OBJECT(canvas), "format", p, g_free);
		ret = 1;
	}

	return ret;
}

gchar *dibuja_get_filename(void)
{
	return (gchar *)g_object_get_data(G_OBJECT(canvas), "filename");
}

gchar *dibuja_get_file_format(void)
{
	return (gchar *)g_object_get_data(G_OBJECT(canvas), "format");
}

void dibuja_set_title(void)
{
	char *title = NULL;

	if(file_get_status() & FILE_NEED_SAVE)
		title = g_strdup_printf("*%s - %s", dibuja_get_filename(), PACKAGE_NAME);
	else
		title = g_strdup_printf("%s - %s", dibuja_get_filename(), PACKAGE_NAME);
	gtk_window_set_title(GTK_WINDOW(gtk_widget_get_toplevel(canvas)), title);
	g_free(title);
}

void dibuja_do_undo(void)
{
	undo_do_undo(SIMPLE_CANVAS(canvas));
}

void dibuja_do_redo(void)
{
	undo_do_redo(SIMPLE_CANVAS(canvas));
}

gboolean dibuja_tool_cancel(void)
{
	if(cur_tool){
		if(cur_tool->cancel){
			return cur_tool->cancel();
		}
	}
	return FALSE;
}

void dibuja_reset_canvas(void)
{
	GtkComboBox *zoom_combo = GTK_COMBO_BOX(get_widget("combo"));

	undo_clear(SIMPLE_CANVAS(canvas));

	if((GTK_IS_WIDGET(zoom_combo)) && GTK_WIDGET_REALIZED(zoom_combo)){
		int i = zoom_combo_default_index(zoom_combo);
		gtk_combo_box_set_active(zoom_combo, i);
	}
	// Set line as default tool
	//tools_toggle_line_button(TRUE);
	gtk_widget_queue_draw(dibuja_get_canvas());
}

void dibuja_clear_image(void)
{
	cairo_surface_t *new_surf;
	cairo_t *cr;
	int w, h, ret;

	if(sel_have()){
		tb_clear_selection_clicked(get_widget("tb_clear_selection"));
		return;
	}
	w = dibuja_get_width();
	h = dibuja_get_height();

	g_return_if_fail(w > 0 && h > 0);

	new_surf = cairo_image_surface_create(DEF_SURF_FORMAT, w, h);

	if(new_surf){
		cr = cairo_create(new_surf);
		cairo_set_source_rgb(cr, 1., 1., 1.);
		cairo_paint(cr);
		cairo_destroy(cr);

		ret = message_box_yes_no(_("Clear undo buffer also?"));
		if(GTK_RESPONSE_YES == ret){
			undo_clear(SIMPLE_CANVAS(dibuja_get_canvas()));
		}
					/* Undo description */
		dibuja_resize(new_surf, _("Clear"));
	}
}

void dibuja_set_scroll_pos(int x, int y)
{
	GtkScrolledWindow *scroll = GTK_SCROLLED_WINDOW(get_widget("scroll"));

	if(!GTK_IS_SCROLLED_WINDOW(scroll)) return;

	GtkAdjustment *hadj = gtk_scrolled_window_get_hadjustment(scroll);
	GtkAdjustment *vadj = gtk_scrolled_window_get_vadjustment(scroll);

	gtk_adjustment_set_value(hadj, (double)x);
	gtk_adjustment_set_value(vadj, (double)y);
}

void dibuja_get_visible_pos(GtkScrolledWindow *scroll, int *x, int *y)
{
	if(!GTK_IS_SCROLLED_WINDOW(scroll)) return;

	GtkAdjustment *hadj = gtk_scrolled_window_get_hadjustment(scroll);
	GtkAdjustment *vadj = gtk_scrolled_window_get_vadjustment(scroll);

	*x = (int)gtk_adjustment_get_value(hadj);
	*y = (int)gtk_adjustment_get_value(vadj);
}
// Hack to try to zoom in the same spot
void dibuja_set_zoom(int zoom_val)
{
	GdkWindow *gdkwindow = NULL;
	int x = 0;
	int y = 0;
	double nz = (double)zoom_val / 100.0;
	GtkScrolledWindow *scroll = GTK_SCROLLED_WINDOW(get_widget("scroll"));

	if(GTK_IS_SCROLLED_WINDOW(scroll)){
		gdkwindow = gtk_widget_get_window(GTK_WIDGET(scroll));
		gdk_window_freeze_updates(gdkwindow);
		dibuja_get_visible_pos(scroll, &x, &y);
		x = (int)win_to_img((double)x, dibuja_get_zoom());
		y = (int)win_to_img((double)y, dibuja_get_zoom());
	}

	gtk_widget_hide(canvas);
	simple_canvas_set_zoom(SIMPLE_CANVAS(canvas), nz);

	if(GTK_IS_SCROLLED_WINDOW(scroll)){
		x = (int)img_to_win((double)x, nz);
		y = (int)img_to_win((double)y, nz);

		dibuja_set_scroll_pos(x, y);
		gdk_window_thaw_updates(gdkwindow);
	}
	gtk_widget_show(canvas);
}

void canvas_image_changed(SimpleCanvas *canvas, gpointer data)
{
	dibuja_show_image_dims(dibuja_get_width(), dibuja_get_height());
}

gboolean canvas_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	if(cur_tool){
		if(cur_tool->expose){
			cur_tool->expose(widget, cr, data);
		}
	}

	return TRUE;
}

gboolean
canvas_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	cur_tool = tool_get();
	if(cur_tool){
		if(3 == event->button && !cur_tool->popup){
			return menu_default_popup(GTK_MENU(get_widget("menu2")), event);
		}
		if(cur_tool->button_press){
			cur_tool->button_press(widget, event, data);
		}
	}
	gtk_widget_queue_draw(widget);

	return TRUE;
}

gboolean
canvas_motion_notify(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	gchar mouse_pos[40];

	if(cur_tool){
		if(cur_tool->motion){
			cur_tool->motion(widget, event, data);
		}
	}

	g_snprintf(mouse_pos, sizeof(mouse_pos) - 1, "(%d,%d)",
			   (int)event->x, (int)event->y);
	gtk_label_set_text(GTK_LABEL(statusbar_label), mouse_pos);

	return TRUE;
}

gboolean
canvas_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(cur_tool){
		if(cur_tool->button_release){
			cur_tool->button_release(widget, event, data);
		}
	}
	gtk_widget_queue_draw(widget);

	return TRUE;
}

void canvas_drag_data_recieved(GtkWidget *widget, GdkDragContext *drag_context,
							   gint x, gint y, GtkSelectionData *data, guint info,
							   guint time, gpointer user_data)
{
	gchar *s = NULL;
	if(data && TARGET_IMG_FILENAME == info){
		if((s = (gchar *)data->data) != NULL){
			gchar *filename = g_strdup(s);
			if(NULL != filename){

				int i;

				for(i = 0; i < strlen(filename); i++){
					if(iscntrl(filename[i])){
						filename[i] = '\0';
					}
				}
				file_open_pixbuf(filename);
				g_free(filename);
			}
		}
	}
	gtk_drag_finish(drag_context, TRUE, FALSE, time);
}


gboolean do_quit(void)
{
	int file_status = file_get_status();
	GtkWidget *w;
	SCColor sc = {0};
	char *filename;
	Settings settings;

	if(file_status & FILE_NEED_SAVE){
		int ret = message_box_yes_no_cancel(
				_("The current file \"%s\" needs saving.\n" "Save?"),
								 dibuja_get_filename());
		switch(ret){
			case GTK_RESPONSE_YES:
				file_save_dialog(canvas, FALSE);
				break;
			case GTK_RESPONSE_CANCEL:
				return TRUE;
			default:
				break;
		}
	}

	// TODO: More elegant cleanup
	if(cur_tool){
		if(cur_tool->finish){
			cur_tool->finish();
		}
	}

	g_strlcpy(settings.version, PACKAGE_VERSION, VERSION_MAX);
	gtk_window_get_size(GTK_WINDOW(window), &settings.width, &settings.height);
	w = get_widget("fontbutton");
	if(GTK_IS_WIDGET(w)){
		g_strlcpy(settings.fontname,
				  gtk_font_button_get_font_name(GTK_FONT_BUTTON(w)),
				  FONTNAME_MAX);
	}

	dibuja_get_fg_sccolor(&sc);
	settings.fgcolor[0] = sc.r;
	settings.fgcolor[1] = sc.g;
	settings.fgcolor[2] = sc.b;
	settings.fgcolor[3] = sc.a;

	dibuja_get_bg_sccolor(&sc);
	settings.bgcolor[0] = sc.r;
	settings.bgcolor[1] = sc.g;
	settings.bgcolor[2] = sc.b;
	settings.bgcolor[3] = sc.a;

	settings.line_width = dibuja_get_line_width();
	settings.fill_mode = (int)dibuja_get_fill_mode();
	settings.antialias = simple_canvas_get_antialias(SIMPLE_CANVAS(canvas));
	settings.antialias = (settings.antialias != CAIRO_ANTIALIAS_NONE);

	w = get_widget("check_button_use_arrow");
	settings.use_arrow = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));

	// Selection
	w = get_widget("check_button_transparent_selection");
	settings.trans_sel = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w));

	filename = util_get_config_filename();
	settings_write(&settings, filename);
	g_free(filename);

	gtk_widget_destroy(window);

	return FALSE;
}

int dibuja_new_file(char *filename, char *format, int width, int height)
{
	cairo_surface_t *new;
	char *newfilename;
	char *newformat;
	cairo_t *cr;
	static int num = 0;

	new = cairo_image_surface_create(DEF_SURF_FORMAT, width, height);
	if(!new) return 0;

	// TODO: Get better way to finish tool
	if(cur_tool){
		if(cur_tool->finish){
			cur_tool->finish();
		}
	}

	cr = cairo_create(new);
	cairo_set_source_rgb(cr, 1., 1., 1.);
	cairo_paint(cr);
	cairo_destroy(cr);

	dibuja_surface_replace(new);

	if(!filename){				/* filename for new files */
		newfilename = g_strdup_printf(_("untitled %d"), num++);
	}
	else{
		newfilename = g_strdup(filename);
	}

	if(!format){
		newformat = g_strdup("png");
	}
	else{
		newformat = g_strdup(format);
	}

	file_set_status(FILE_NEW);
	dibuja_set_filename(newfilename, newformat);

	dibuja_set_title();

	dibuja_reset_canvas();

	g_free(newfilename);
	g_free(newformat);

	gtk_widget_queue_draw(canvas);
	return 1;
}

int dibuja_new_file_from_pixbuf(GdkPixbuf *pixbuf, char *filename, char *format)
{
	int ret = 0;
	int width, height;

	if(!GDK_IS_PIXBUF(pixbuf)) return ret;

	width = gdk_pixbuf_get_width(pixbuf);
	height = gdk_pixbuf_get_height(pixbuf);
	ret = dibuja_new_file(filename, format, width, height);
	if(ret){
		cairo_t *cr = dibuja_get_cr(CR_SURFACE);
		gdk_cairo_set_source_pixbuf(cr, pixbuf, 0., 0.);
		cairo_paint(cr);
		cairo_destroy(cr);
		ret = 1;
	}

	return ret;
}
/* Resize to new surface */
int dibuja_resize(cairo_surface_t *new, char *desc)
{
	GdkRectangle rect;
	cairo_surface_t *old;
	if(!new){
		g_warning("%s: %d - Cannot resize\n", __FILE__, __LINE__);
		return 0;
	}
	if(!desc){
		g_warning("%s: %d - Unkown resize operation\n", __FILE__, __LINE__);
		desc = g_strdup("Unkown resize operation");
	}

	rect.x = 0;
	rect.y = 0;
	rect.width = cairo_image_surface_get_width(dibuja_get_surface());
	rect.height = cairo_image_surface_get_height(dibuja_get_surface());

	old = simple_canvas_swap_image(SIMPLE_CANVAS(canvas), new);
	undo_add_from_surface(SIMPLE_CANVAS(canvas), old, rect, desc);

	return 1;
}

int dibuja_get_width(void)
{
	return cairo_image_surface_get_width(dibuja_get_surface());
}

int dibuja_get_height(void)
{
	return cairo_image_surface_get_height(dibuja_get_surface());
}

gboolean dibuja_get_use_trans_sel(void)
{
	return gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(
			get_widget("check_button_transparent_selection")));
}

void dibuja_run_preferences_dialog(void)
{
//	prefs_run_preferences_dialog(dibuja_get_canvas());
}

void dibuja_set_fg_color(GdkColor *color, guint16 alpha)
{
	if(color){
		simple_canvas_set_fg_rgba(SIMPLE_CANVAS(canvas),
								  (double)color->red / 65535.,
								  (double)color->green / 65535.,
								  (double)color->blue / 65535.,
								  (double)alpha / 65535.);
	}
}

void dibuja_set_bg_color(GdkColor *color, guint16 alpha)
{
	if(color){
		simple_canvas_set_bg_rgba(SIMPLE_CANVAS(canvas),
								  (double)color->red / 65535.,
								  (double)color->green / 65535.,
								  (double)color->blue / 65535.,
								  (double)alpha / 65535.);
	}
}

void dibuja_get_fg_color(GdkColor *color)
{
	if(color){
		simple_canvas_get_fg_color(SIMPLE_CANVAS(canvas), color);
	}
}

void dibuja_get_bg_color(GdkColor *color)
{
	if(color){
		simple_canvas_get_bg_color(SIMPLE_CANVAS(canvas), color);
	}
}

void dibuja_get_fg_sccolor(SCColor *color)
{
	if(color){
		simple_canvas_get_fg_rgba(SIMPLE_CANVAS(canvas),
							&(color->r), &(color->g),
							&(color->b), &(color->a) );
	}
}

void dibuja_get_bg_sccolor(SCColor *color)
{
	if(color){
		simple_canvas_get_bg_rgba(SIMPLE_CANVAS(canvas),
							&(color->r), &(color->g),
							&(color->b), &(color->a) );
	}
}

void dibuja_set_line_width(int lw)
{
	simple_canvas_set_line_width(SIMPLE_CANVAS(canvas), lw);
}

int dibuja_get_line_width(void)
{
	return simple_canvas_get_line_width(SIMPLE_CANVAS(canvas));
}

void dibuja_set_antialias(gboolean aa)
{
	simple_canvas_set_antialias(SIMPLE_CANVAS(canvas),
				FALSE == aa ? CAIRO_ANTIALIAS_NONE : CAIRO_ANTIALIAS_DEFAULT);
}

ShapeFillType dibuja_get_fill_mode(void)
{
	return fill_mode;
}

void dibuja_set_fill_mode(ShapeFillType mode)
{
	if(mode >= SHAPE_FILL_NONE && mode < SHAPE_FILL_LAST)
		fill_mode = mode;
}

GtkWidget *dibuja_create_canvas(GtkBuilder *builder, int w, int h)
{
	canvas = simple_canvas_new(w, h);
	window = GTK_WIDGET (gtk_builder_get_object (builder, "window1"));

	g_signal_connect(G_OBJECT(canvas), "simple-canvas-pointer-motion",
									G_CALLBACK(canvas_motion_notify), NULL);
	g_signal_connect(G_OBJECT(canvas), "simple-canvas-button-press",
									G_CALLBACK(canvas_button_press), NULL);
	g_signal_connect(G_OBJECT(canvas), "simple-canvas-button-release",
									G_CALLBACK(canvas_button_release), NULL);
	g_signal_connect(G_OBJECT(canvas), "simple-canvas-expose",
									G_CALLBACK(canvas_expose), NULL);
	g_signal_connect(G_OBJECT(canvas), "simple-canvas-image-changed",
									G_CALLBACK(canvas_image_changed), NULL);

	gtk_drag_dest_set(canvas,
					  GTK_DEST_DEFAULT_ALL,
					  drop_targets,
					  G_N_ELEMENTS(drop_targets),
					  GTK_DEST_DEFAULT_ALL );

	g_signal_connect(G_OBJECT(canvas), "drag-data-received",
									G_CALLBACK(canvas_drag_data_recieved), NULL);

	statusbar_label = GTK_WIDGET (gtk_builder_get_object (builder, "label1"));
	statusbar_label_dims = GTK_WIDGET (gtk_builder_get_object (builder,
									   "statusbar_label_dims"));
	cur_tool = tool_get();

	simple_canvas_set_line_width(SIMPLE_CANVAS(canvas), 1);
	//prefs_init(SIMPLE_CANVAS(canvas));

	return canvas;
}

#ifndef FILE_FORMAT_KEY
#define FILE_FORMAT_KEY	"pixbuf-format"
#endif

void dibuja_set_pixbuf_format(GdkPixbufFormat *format)
{
	g_object_set_data_full(G_OBJECT(dibuja_get_canvas()), FILE_FORMAT_KEY,
						   (gpointer)gdk_pixbuf_format_copy(format),
						   (GDestroyNotify)gdk_pixbuf_format_free);
}

GdkPixbufFormat *dibuja_get_pixbuf_format(void)
{
	GObject *obj = G_OBJECT(dibuja_get_canvas());

	return (GdkPixbufFormat *)g_object_get_data(obj, FILE_FORMAT_KEY);
}



