/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include "draw-funcs.h"
#include "dibuja.h"

typedef void (*ShapeFunc)(cairo_t *cr, gdouble x, gdouble y, gdouble w, gdouble h);


void rect_from_pts(GdkRectangle *rectangle, int p1x, int p1y, int p2x, int p2y)
{
	if(!rectangle) return;

	rectangle->x = MIN(p1x, p2x);
	rectangle->y = MIN(p1y, p2y);
	rectangle->width = ABS(p1x - p2x);
	rectangle->height = ABS(p1y - p2y);
}

void resize_rect(GdkRectangle *rectangle, int x, int y)
{
	if(!rectangle) return;

	rectangle->x -= x;
	rectangle->y -= y;
	rectangle->width += y * 2;
	rectangle->height += y * 2;
}

static void _draw_line(cairo_t *cr, gdouble x1, gdouble y1, gdouble x2, gdouble y2)
{
	int lw = cairo_get_line_width(cr);

	if(lw % 2 != 0){

		if(x1 == x2){
			x1 += .5;
			x2 += .5;
			cairo_move_to(cr, x1,  y1 > y2 ? y1 + 1. : y1);
			cairo_line_to(cr, x2,  y2 > y1 ? y2 + 1. : y2);
		}
		else if(y1 == y2){
			y1 += .5;
			y2 += .5;
			cairo_move_to(cr, (x1 > x2 ? x1 + 1. : x1),  y1);
			cairo_line_to(cr, (x2 > x1 ? x2 + 1. : x2),  y2);
		}
		else{
			cairo_move_to(cr, x1 + (x1 < x2 ? .49 : .51),
							  y1 + (y1 < y2 ? .49 : .51));
			cairo_line_to(cr, x2 + (x1 > x2 ? .49 : .51),
							  y2 + (y1 > y2 ? .49 : .51));
		}
	}
	else{
		cairo_move_to(cr, x1, y1);
		cairo_line_to(cr, x2, y2);
	}
}

void draw_line(cairo_t *cr, gdouble x1, gdouble y1, gdouble x2, gdouble y2)
{	
	_draw_line(cr, x1, y1, x2, y2);
	cairo_stroke(cr);
}

static void _draw_rectangle(cairo_t *cr, gdouble x, gdouble y, gdouble w, gdouble h)
{
	cairo_rectangle(cr, x, y, w, h);
}

static void _draw_ellipse(cairo_t *cr, gdouble x, gdouble y, gdouble w, gdouble h)
{
	cairo_save (cr);
	cairo_translate (cr, x + w / 2., y + h / 2.);
	cairo_scale (cr, w / 2., h / 2.);
	cairo_arc (cr, 0., 0., 1., 0., 2 * M_PI);
	cairo_restore (cr);
}

static void _draw_rrect(cairo_t *cr, gdouble x, gdouble y, gdouble width, gdouble height)
{
	gdouble corner_radius = MIN(height, width) / 10.0;
	gdouble radius = corner_radius;
	gdouble degrees = M_PI / 180.0;

	cairo_new_sub_path (cr);
	cairo_arc (cr, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
	cairo_arc (cr, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
	cairo_arc (cr, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
	cairo_arc (cr, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
	cairo_close_path (cr);
}

static void draw_shape(cairo_t *cr, ShapeFunc func, GdkRectangle r, SCColor *fg, SCColor *bg)
{
	int lw = (int)cairo_get_line_width(cr);
	double x = (double)r.x;
	double y = (double)r.y;
	double w = (double)r.width;
	double h = (double)r.height;

	if(lw > (int)w || lw > (int)h){
		lw = MIN(w, h);
		if(lw < 1) lw = 1;
	}

	if(lw % 2 != 0){
		x += .5;
		y += .5;
	}

	cairo_set_line_width (cr, lw);
	//cairo_rectangle(cr, x, y, w, h);
	func(cr, x, y, w, h);
	if(bg){
		cairo_set_source_rgba(cr, bg->r, bg->g, bg->b, bg->a);
		cairo_fill_preserve (cr);
	}
	cairo_set_source_rgba(cr, fg->r, fg->g, fg->b, fg->a);
	cairo_stroke (cr);
}

void draw_rectangle(cairo_t *cr, GdkRectangle r, SCColor *fg, SCColor *bg)
{
	draw_shape(cr, _draw_rectangle, r, fg, bg);
}

void draw_ellipse(cairo_t *cr, GdkRectangle r, SCColor *fg, SCColor *bg)
{
	draw_shape(cr, _draw_ellipse, r, fg, bg);
}

void draw_rounded_rectangle(cairo_t *cr, GdkRectangle r, SCColor *fg, SCColor *bg)
{
	draw_shape(cr, _draw_rrect, r, fg, bg);
}

void polygon(cairo_t *cr, int cx, int cy, int x, int y, int nsides,
			 SCColor *fg, SCColor *bg, GdkRectangle *rc)
{
	double a = 0.;
	double frac = (360. / nsides);
	int i;
	GdkPoint p;
	int lw;
	int width;
	
	if(nsides < 3){ return; }

	lw = (int)cairo_get_line_width(cr);
	width = sqrt(((cx - x) * (cx - x) +
					 (cy - y) * (cy - y))) * 2;

	if(lw > width){ lw = width; }

	if(lw % 2 != 0){
		cx += .5;
		cy += .5;
	}

	cairo_set_line_width(cr, (double)lw);
	cairo_new_sub_path (cr);
	for (i = 0; i < nsides; i++){
		p = draw_calc_rotate(a, (double)(x - cx), (double)(y - cy));
		cairo_line_to(cr, (double)p.x + cx, (double) p.y + cy);
		a += frac;
	}

	cairo_close_path (cr);

	if(rc){
		double dx1, dy1, dx2, dy2;
		cairo_path_extents(cr, &dx1, &dy1, &dx2, &dy2);
		rect_from_pts(rc, (int)dx1, (int)dy1, (int)dx2, (int)dy2);
	}

	if(bg){
        cairo_set_source_rgba(cr, bg->r, bg->g, bg->b, bg->a);
		cairo_fill_preserve (cr);
	}

	cairo_set_source_rgba(cr, fg->r, fg->g, fg->b, fg->a);

	cairo_stroke(cr);
}

double calc_angle(double p1x, double p1y, double p2x, double p2y)
{
	double angle = atan2 (p2y - p1y, p2x - p1x);
	return angle;
}

void draw_rotate_point(double sx, double sy, double *ex, double *ey, double angle)
{
	double line_len;

	if(ex && ey){
		line_len = draw_line_len(sx, sy, *ex, *ey);
		*ex = sx + line_len * cos(angle);
		*ey = sy + line_len * sin(angle);
	}
}

double draw_line_len(double sx, double sy, double ex, double ey)
{
	return sqrt((sx - ex) * (sx - ex) + (sy - ey) * (sy - ey));
}

void draw_arrow(cairo_t *cr, double sh_startx, double sh_starty, double sh_endx, double sh_endy, GdkRectangle *r)
{
	double angle = calc_angle(sh_endx, sh_endy, sh_startx, sh_starty);
	double xbarb1, ybarb1, xbarb2, ybarb2, xcap, ycap, xtip, ytip;
	double arrowh_len ;
	double arrowh_wid = cairo_get_line_width(cr);
	double barb_deg = DEG2RAD(45./2.);

	double ofst = .5;
	double x, y, len;

	len = cairo_get_line_width(cr) / 2.;

	arrowh_len = arrowh_wid * 4.;

	// Barb 1
	xbarb1 = sh_endx - arrowh_len;
	ybarb1 = sh_endy - arrowh_len;
	draw_rotate_point(sh_endx, sh_endy, &xbarb1, &ybarb1, angle - barb_deg);

	// Barb 2
	xbarb2 = sh_endx - arrowh_len;
	ybarb2 = sh_endy - arrowh_len;
	draw_rotate_point(sh_endx, sh_endy, &xbarb2, &ybarb2, angle + barb_deg);

	// Cap
	xcap = sh_endx - (arrowh_len - arrowh_wid);
	ycap = sh_endy - (arrowh_len - arrowh_wid);
	draw_rotate_point(sh_endx, sh_endy, &xcap, &ycap, angle);

	// Tip
	xtip = sh_endx;
	ytip = sh_endy;

	cairo_save(cr);
	//cairo_set_source_rgb(cr, 0., 0., 0.);
	cairo_set_line_width(cr, 1.0);

	cairo_move_to(cr, xtip + ofst, ytip + ofst);
	cairo_line_to(cr, xbarb1 + ofst, ybarb1 + ofst);

	// Shaft
	x = xcap + len;
	y = ycap + len;
	draw_rotate_point(xcap, ycap, &x, &y, angle - DEG2RAD(90.));
	cairo_line_to(cr, x + ofst, y + ofst);

	x = sh_startx + len;
	y = sh_starty + len;
	draw_rotate_point(sh_startx, sh_starty, &x, &y, angle - DEG2RAD(90.));
	cairo_line_to(cr, x + ofst, y + ofst);

	x = sh_startx + len;
	y = sh_starty + len;
	draw_rotate_point(sh_startx, sh_starty, &x, &y, angle + DEG2RAD(90.));
	cairo_line_to(cr, x + ofst, y + ofst);

	x = xcap + len;
	y = ycap + len;
	draw_rotate_point(xcap, ycap, &x, &y, angle + DEG2RAD(90.));
	cairo_line_to(cr, x + ofst, y + ofst);

	cairo_line_to(cr, xbarb2 + ofst, ybarb2 + ofst);

	cairo_close_path(cr);

	if(r){
		double x1, y1, x2, y2;

		cairo_stroke_extents(cr, &x1, &y1, &x2, &y2);
		rect_from_pts(r, (int)x1, (int)y1, (int)x2, (int)y2);
		r->x -= 2;
		r->y -= 2;
		r->width += 4;
		r->height += 4;
	}

	cairo_fill(cr);
	cairo_restore(cr);
}

cairo_surface_t *draw_copy_area(cairo_surface_t *source, int x, int y, int w, int h)
{
	cairo_surface_t *copy = NULL;

	if(source){
		cairo_t *cr;

		if(w < 1 || h < 1) return NULL;

		copy = cairo_image_surface_create(DEF_SURF_FORMAT, w, h);
		cr = cairo_create(copy);
		cairo_set_source_surface(cr, source, (gdouble)-x, (gdouble)-y);
		cairo_paint(cr);
		cairo_destroy(cr);
	}

	return copy;
}
/******************************************************************************/
GdkPoint draw_calc_rotate(double angle_deg, int x, int y)
{
	GdkPoint pt = {0, 0};
	double rad;
	double dx, dy;

	dx = (double)x;
	dy = (double)y;
	rad = M_PI / 180.0 * angle_deg ;

	pt.x = (int)((dx * cos(rad) - dy * sin(rad)) );
	pt.y = (int)((dx * sin(rad) + dy * cos(rad)) );

	return pt;
}

void draw_rotate_pts(double angle_deg, GdkPoint *p1, GdkPoint *p2,
									   GdkPoint *p3, GdkPoint *p4)
{
	*p1 = draw_calc_rotate(angle_deg, p1->x, p1->y);
	*p2 = draw_calc_rotate(angle_deg, p2->x, p2->y);
	*p3 = draw_calc_rotate(angle_deg, p3->x, p3->y);
	*p4 = draw_calc_rotate(angle_deg, p4->x, p4->y);
}

void draw_rotate_rectangle(double angle_deg, GdkRectangle *r)
{
	if(!r) return;

	int cx = r->x + r->width / 2;
	int cy = r->y + r->height / 2;

	r->x -= (r->width - r->width / 2);
	r->y -= (r->height - r->height / 2);
  {
	GdkPoint p1 = { r->x,             r->y};
	GdkPoint p2 = { r->x + r->width,  r->y};
	GdkPoint p3 = { r->x + r->width,  r->y + r->height};
	GdkPoint p4 = { r->x,             r->y + r->height};

	draw_rotate_pts(angle_deg, &p1, &p2, &p3, &p4);

	r->x = MIN(MIN(p1.x, p2.x), MIN(p3.x, p4.x));
	r->y = MIN(MIN(p1.y, p2.y), MIN(p3.y, p4.y));
	r->width = MAX(MAX(p1.x, p2.x), MAX(p3.x, p4.x));
	r->height = MAX(MAX(p1.y, p2.y), MAX(p3.y, p4.y));

	r->width = ABS(r->x - r->width);
	r->height = ABS(r->y - r->height);

	r->x = cx - r->width / 2;
	r->y = cy - r->height / 2;
  }
}

void draw_transform(cairo_t *dest_cr, cairo_surface_t *source,
							double xtr, double ytr, double xsc, double ysc,
							double x, double y)
{
	cairo_translate(dest_cr, xtr, ytr);
	cairo_scale(dest_cr, xsc, ysc);
	cairo_set_source_surface(dest_cr, source, x, y);
}

cairo_surface_t *draw_surface_flip(cairo_surface_t *source, int dir)
{
	cairo_surface_t *dest;
	cairo_t *cr_dest;
	double w, h;
	double xsc, ysc;
	
	if(FLIP_HORZ != dir && FLIP_VERT != dir) return NULL;
	if(!source) return NULL;

	w = (double)cairo_image_surface_get_width(source);
	h = (double)cairo_image_surface_get_height(source);

	dest = cairo_image_surface_create(DEF_SURF_FORMAT, (int)w, (int)h);
	if(!dest) return NULL;

	cr_dest = cairo_create(dest);
	if(FLIP_HORZ == dir){
		h = 0.;
		xsc = -1.0; ysc = 1.0;
	}
	else{
		w = 0.;
		xsc = 1.0; ysc = -1.0;
	}

	draw_transform(cr_dest, source, w, h,
				  xsc, ysc,
				  0., 0.);

	cairo_set_source_surface(cr_dest, source, 0., 0.);
	cairo_paint(cr_dest);

	cairo_destroy(cr_dest);

	return dest;
}

/*
	Create new rotated surface	
	angle in degrees
*/
cairo_surface_t *draw_rotate_image(cairo_surface_t *source, double angle)
{
	cairo_surface_t *dest;
	cairo_t *cr_dest;
	double w, h;
	GdkRectangle r = {0};

	if(!source) return NULL;

	w = (double)cairo_image_surface_get_width(source);
	h = (double)cairo_image_surface_get_height(source);

	r.width = (int)w;
	r.height = (int)h;
	draw_rotate_rectangle(angle, &r);

	dest = cairo_image_surface_create(DEF_SURF_FORMAT, r.width, r.height);
	if(!dest) return NULL;

	cr_dest = cairo_create(dest);

	angle = angle * M_PI / 180.;
	cairo_translate(cr_dest, (double)r.width / 2., (double)r.height / 2.);
	cairo_rotate(cr_dest, angle);
	cairo_translate(cr_dest, -w / 2.0, -h / 2.0);
	cairo_set_source_surface(cr_dest, source, 0, 0);

	cairo_paint(cr_dest);
	cairo_destroy(cr_dest);

	return dest;
}

cairo_surface_t *draw_scale_image(cairo_surface_t *source, gint wnew, gint hnew)
{
	cairo_surface_t *dest;
	cairo_t *cr;
	int w, h;
	double xcsale, yscale;

	if(!(wnew > 0 && hnew > 0)) return NULL;
	if(!source) return NULL;

	dest = cairo_image_surface_create(DEF_SURF_FORMAT, wnew, hnew);
	if(!dest) return NULL;

	cr = cairo_create(dest);
	cairo_set_source_rgb(cr, 1., 1., 1.);
	cairo_paint(cr);

	w = cairo_image_surface_get_width(source);
	h = cairo_image_surface_get_height(source);

	xcsale = (double)wnew / (double)w;
	yscale = (double)hnew / (double)h;

	draw_transform(cr, source,
			  0., 0.,
			  xcsale, yscale,
			  0., 0.);

	cairo_paint(cr);
	cairo_destroy(cr);

	return dest;
}

cairo_surface_t *draw_resize_image(cairo_surface_t *source,
								   gint wnew, gint hnew,
								   gboolean centered)
{
	cairo_surface_t *dest;
	cairo_t *cr;
	int w, h;

	if(!(wnew > 0 && hnew > 0)) return NULL;
	if(!source) return NULL;

	dest = cairo_image_surface_create(DEF_SURF_FORMAT, wnew, hnew);
	if(!dest) return NULL;

	cr = cairo_create(dest);
	// Set background color
	cairo_set_source_rgb(cr, 1., 1., 1.);
	cairo_paint(cr);

	w = cairo_image_surface_get_width(source);
	h = cairo_image_surface_get_height(source);

	if(centered){
		cairo_set_source_surface(cr, source,
								(double)wnew / 2. - (double)w / 2.,
								(double)hnew / 2. - (double)h / 2.);
	}
	else{
		cairo_set_source_surface(cr, source, 0., 0.);
	}

	cairo_paint(cr);
	cairo_destroy(cr);

	return dest;
}
/*
static double calc_angle(double p1x, double p1y, double p2x, double p2y)
{
	double angle = atan2 (p2y - p1y, p2x - p1x);
	return angle;
}*/

void draw_constrain_pt(double p1x, double p1y, double p2x,
							  double p2y, double *xret, double *yret)
{
	double angle, oldangle;
	GdkPoint p;
	double x, y;

	x = p2x - p1x;
	y = p2y - p1y;

	angle = calc_angle(p1x, p1y, p2x, p2y);

	// Convert to degrees. To understand better
	angle = angle * 180. / M_PI;
	oldangle = angle;

	// Find multiple of 45
	angle = (angle + 22.5) / 45;
	angle = floor(angle);
	angle = angle * 45.;

	p = draw_calc_rotate(angle - oldangle, (int)x, (int)y);

	*xret = (double)p.x + p1x;
	*yret = (double)p.y + p1y;
}





