/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "effects.h"

/*
Some parts from gimp
*/

#if GEGL_MAJOR_VERSION == 0 && GEGL_MINOR_VERSION == 3 && GEGL_MICRO_VERSION >= 8
  #define USE_INSTA		1
#elif GEGL_MAJOR_VERSION == 0 && GEGL_MINOR_VERSION == 4
  #define USE_INSTA		1
#endif

#if BABL_MAJOR_VERSION == 0 && BABL_MINOR_VERSION <= 1 && BABL_MICRO_VERSION <= 2
const Babl *fx_get_format_RGB24(void)
{
	const Babl *format = babl_format_new (
			"name", "cairo-RGB24",
			babl_model ("R'G'B'"),
			babl_type ("u8"),
			babl_component ("B'"),
			babl_component ("G'"),
			babl_component ("R'"),
			babl_component ("PAD"),
			NULL
			);
	return format;
}

const Babl *fx_get_format_ARGB32(void)
{
	const Babl *format = babl_format_new (
		"name", "cairo-ARGB32",
		babl_model ("R'G'B'A"),
		babl_type ("u8"),
		babl_component ("B'"),
		babl_component ("G'"),
		babl_component ("R'"),
		babl_component ("A"),
		NULL
      );
	return format;
}

#else

const Babl *fx_get_format_RGB24(void)
{
	return babl_format("cairo-RGB24");
}

const Babl *fx_get_format_ARGB32(void)
{
	return babl_format ("cairo-ARGB32");
}

#endif

GeglRectangle gegl_rectangle(int x, int y, int w, int h)
{
	GeglRectangle r = {x, y, w, h};
	return r;
}

GeglBuffer *fx_copy_buffer(GeglBuffer *src)
{
	GeglBuffer *dest = NULL;
	GeglRectangle  rect = {0};
	const Babl *format;

	g_return_val_if_fail(src != NULL, NULL);

	rect.width = gegl_buffer_get_width(src);
	rect.height = gegl_buffer_get_height(src);

	g_return_val_if_fail(rect.width > 0 && rect.height > 0, NULL);

	format = gegl_buffer_get_format(src);
	g_return_val_if_fail(format != NULL, NULL);

	dest = gegl_buffer_new(&rect, format);

	return dest;
}

GeglBuffer *gegl_buffer_from_surface (cairo_surface_t *ARGB32_surface)
{
	cairo_surface_t *surface = ARGB32_surface;
	GeglBuffer *gegl_buffer;
	int stride;
	const Babl *format;
	cairo_format_t cairo_format;
	GeglRectangle  rect = {0};

	g_return_val_if_fail(surface, NULL);

	cairo_format = cairo_image_surface_get_format(surface);
	g_return_val_if_fail(CAIRO_FORMAT_ARGB32 == cairo_format, NULL);

	rect.width = cairo_image_surface_get_width(surface);
	rect.height = cairo_image_surface_get_height(surface);
	g_return_val_if_fail(rect.width > 0 && rect.height > 0, NULL);


	format = fx_get_format_ARGB32();
	g_return_val_if_fail(format, NULL);

	gegl_buffer = gegl_buffer_new(&rect, format);
	if(!gegl_buffer)return NULL;

	stride = cairo_image_surface_get_stride(surface);

#if GEGL_MAJOR_VERSION == 0 && GEGL_MINOR_VERSION == 1
	gegl_buffer_set (gegl_buffer, &rect, format,
					cairo_image_surface_get_data(surface),
					stride);
#else
	gegl_buffer_set (gegl_buffer, &rect, 0, format,
					cairo_image_surface_get_data(surface),
					stride);
#endif
	return gegl_buffer;
}

cairo_surface_t *buffer_to_surface(GeglBuffer *buffer)
{
	cairo_surface_t *surface;
	GeglRectangle  rect = {0};
	guchar *p;
	int stride;
	const Babl *format;

	g_return_val_if_fail(buffer, NULL);

	rect.width = gegl_buffer_get_width(buffer);
	rect.height = gegl_buffer_get_height(buffer);

	surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32,
	    								  rect.width,
	    								  rect.height);

	p = cairo_image_surface_get_data(surface);
	stride = cairo_image_surface_get_stride(surface);
	format = fx_get_format_ARGB32();

#if GEGL_MAJOR_VERSION == 0 && GEGL_MINOR_VERSION == 1
	gegl_buffer_get(buffer, 1.0, &rect, format,
					(gpointer)p,  stride);
#else
	gegl_buffer_get(buffer, &rect, 1.0, format,
					(gpointer)p,  stride, GEGL_ABYSS_NONE);
#endif
	cairo_surface_mark_dirty(surface);

	return surface;
}

void
fx_apply_invert (GeglBuffer          *src_buffer,
                      GeglBuffer          *dest_buffer,
                      const GeglRectangle *dest_rect)
{
	GeglNode *operation_node;

	operation_node = gegl_node_new_child (NULL, "operation", "gegl:invert",
										NULL);
	if(!GEGL_IS_NODE(operation_node))
		operation_node = gegl_node_new_child(NULL, "operation",
											 "gegl:invert-linear", NULL);

	fx_apply_operation (src_buffer,
							   operation_node,
							   dest_buffer,
							   dest_rect);
	g_object_unref (operation_node);
}

void
fx_apply_grey (GeglBuffer          *src_buffer,
                      GeglBuffer          *dest_buffer,
                      const GeglRectangle *dest_rect)
{
	GeglNode *operation_node;

	operation_node = gegl_node_new_child (NULL, "operation", "gegl:grey",
										NULL);

	fx_apply_operation (src_buffer,
							   operation_node,
							   dest_buffer,
							   dest_rect);
	g_object_unref (operation_node);
}

/******************************************
	int size_x - Width of blocks in pixels	
	int size_y - Height of blocks in pixels
*/
void
fx_apply_pixelize (GeglBuffer *src_buffer,
                 GeglBuffer *dest_buffer,
                 const GeglRectangle *dest_rect,
                 int size_x, int size_y)
{
	GeglNode *node;

	g_return_if_fail (GEGL_IS_BUFFER (src_buffer));
	g_return_if_fail (GEGL_IS_BUFFER (dest_buffer));

	node = gegl_node_new_child (NULL,
								"operation",  "gegl:pixelize",
								"size-x",  size_x,
								"size-y",  size_y,
								NULL);

	fx_apply_operation (src_buffer,
							   node,
							   dest_buffer,
							   dest_rect);
	g_object_unref (node);
}

void
fx_apply_insta_1977 (char *op, GeglBuffer *src_buffer,
                 GeglBuffer *dest_buffer,
                 const GeglRectangle *dest_rect)
{
	GeglNode *node;

	g_return_if_fail (GEGL_IS_BUFFER (src_buffer));
	g_return_if_fail (GEGL_IS_BUFFER (dest_buffer));

#ifdef USE_INSTA
	node = gegl_node_new_child (NULL, "operation", op, "preset",
								GEGL_DIBUJA_INSTA_CURVE_PRESET_1977 ,NULL);
#else
	node = gegl_node_new_child (NULL, "operation", op, NULL);
#endif
	fx_apply_operation (src_buffer, node, dest_buffer, dest_rect);
	g_object_unref (node);
}

void
fx_apply_insta_brannan (char *op, GeglBuffer *src_buffer,
                 GeglBuffer *dest_buffer,
                 const GeglRectangle *dest_rect)
{
	GeglNode *node;

	g_return_if_fail (GEGL_IS_BUFFER (src_buffer));
	g_return_if_fail (GEGL_IS_BUFFER (dest_buffer));

#ifdef USE_INSTA
	node = gegl_node_new_child (NULL, "operation", op, "preset",
								GEGL_DIBUJA_INSTA_CURVE_PRESET_BRANNAN ,NULL);
#else
	node = gegl_node_new_child (NULL, "operation", op, NULL);
#endif
	fx_apply_operation (src_buffer, node, dest_buffer, dest_rect);
	g_object_unref (node);
}

void
fx_apply_emboss(GeglBuffer          *src_buffer,
                      GeglBuffer          *dest_buffer,
                      const GeglRectangle *dest_rect)
{
	GeglNode *operation_node;

	operation_node = gegl_node_new_child (NULL, "operation", "gegl:emboss",
										NULL);

	fx_apply_operation (src_buffer,
							   operation_node,
							   dest_buffer,
							   dest_rect);
	g_object_unref (operation_node);
}

/* 0 - 100 */
/*
void
fx_apply_red_eye_removal (GeglBuffer *src_buffer,
                 GeglBuffer *dest_buffer,
                 const GeglRectangle *dest_rect,
                 int threshold)
{
	GeglNode *node;
	double th = (double)threshold;

	g_return_if_fail (GEGL_IS_BUFFER (src_buffer));
	g_return_if_fail (GEGL_IS_BUFFER (dest_buffer));

	node = gegl_node_new_child (NULL,
								"operation",  "gegl:red-eye-removal",
								"Threshold",  th ,
								NULL);

	fx_apply_operation (src_buffer,
							   node,
							   dest_buffer,
							   dest_rect);
	g_object_unref (node);
}
*/
void
fx_apply_operation (GeglBuffer          *src_buffer,
						   GeglNode            *operation,
						   GeglBuffer          *dest_buffer,
						   const GeglRectangle *dest_rect)
{
	GeglNode      *gegl;
	GeglNode      *dest_node;
	GeglRectangle  rect = {0};
	gdouble        value;
	int progress = 0;

	g_return_if_fail (src_buffer == NULL || GEGL_IS_BUFFER (src_buffer));
	g_return_if_fail (GEGL_IS_NODE (operation));
	g_return_if_fail (GEGL_IS_BUFFER (dest_buffer));

	if (dest_rect) {
		  rect = *dest_rect;
	}
	else {
		  rect = gegl_rectangle (0, 0, gegl_buffer_get_width  (dest_buffer),
		                                gegl_buffer_get_height (dest_buffer));
	}

	gegl = gegl_node_new ();

	if (! gegl_node_get_parent (operation))
		gegl_node_add_child (gegl, operation);

	if (src_buffer) {
		  GeglNode *src_node;

		  src_node = gegl_node_new_child (gegl,
		                                  "operation", "gegl:buffer-source",
		                                  "buffer",    src_buffer,
		                                  NULL);

		  gegl_node_connect_to (src_node,  "output", operation, "input");
	}

	dest_node = gegl_node_new_child (gegl,
		                               "operation", "gegl:write-buffer",
		                               "buffer",    dest_buffer,
		                               NULL);


	gegl_node_connect_to (operation, "output",
		                    dest_node, "input");

	if (progress) {
		  GeglProcessor *processor;

		  processor = gegl_node_new_processor (dest_node, &rect);


		  while (gegl_processor_work (processor, &value))
		    

		  g_object_unref (processor);
	}
	else {
		  gegl_node_blit (dest_node, 1.0, &rect,
		                  NULL, NULL, 0, GEGL_BLIT_DEFAULT);
	}

	g_object_unref (gegl);

}




