/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "file.h"
#include "dibuja.h"
#include "util.h"
#include "gdkpixbuf-drawable.h"
#include "uni-exiv2.hpp"

#define file_get_pb_format(f)	((GdkPixbufFormat *)g_object_get_data( \
								(G_OBJECT(ff)),(FILE_FORMAT_KEY)))

static gint file_status = FILE_NEW;
static GSList *_pbformats = NULL;
static char bad_pixbuf_format_msg[] =
	N_("Dibuja could not determine a supported writable file format "
	"based on the filename.\n"
	"Please try a different file extension like .png or .jpg.");


void file_set_status(gint new_status)
{
	file_status = new_status;
	if(file_status & FILE_NEED_SAVE){
		dibuja_set_title();
	}
}

gint file_get_status()
{
	return file_status;
}

char *file_def_save_folder(void)
{
	char *folder = g_strdup(g_get_user_special_dir(G_USER_DIRECTORY_PICTURES));

	if(!(g_file_test(folder, G_FILE_TEST_IS_DIR) &&
	   g_file_test(folder, G_FILE_TEST_EXISTS))){
	   
		g_free(folder);
		folder = g_strdup((gchar *)g_get_home_dir());
	}

	return folder;
}

char *file_get_cur_dir(void)
{
	char *dir = g_path_get_dirname(dibuja_get_filename());

	if(!dir) dir = g_strdup(".");

	return dir;
}

/* From evince */
GdkPixbufFormat* get_gdk_pixbuf_format_by_extension(gchar *uri)
{
	GSList *iter;
	int i, j;

	j = 0;
	for(iter = _pbformats; iter; iter = iter->next, j++) {
		gchar **extension_list;
		GdkPixbufFormat *format = iter->data;

		if(gdk_pixbuf_format_is_disabled(format) ||
	    	    !gdk_pixbuf_format_is_writable(format))
		            continue;

	        extension_list = gdk_pixbuf_format_get_extensions(format);

		for(i = 0; extension_list[i] != NULL; i++) {
			if(g_str_has_suffix(uri, extension_list[i])) {
				g_strfreev(extension_list);
				return format;
			}
		}
		g_strfreev(extension_list);
	}

	return NULL;
}

static void save_dlg_response(GtkWidget *widget, gint id, gpointer data)
{
	if(GTK_RESPONSE_ACCEPT == id){
		GtkFileChooser *chooser;
		GtkFileFilter *ff = NULL;
		GdkPixbufFormat *format = NULL;
		int writable = 0;
		char *filename = NULL;

		chooser = GTK_FILE_CHOOSER(widget);

		ff = gtk_file_chooser_get_filter(chooser);
		//format =(GdkPixbufFormat *)g_object_get_data(G_OBJECT(ff), FILE_FORMAT_KEY);
		format = file_get_pb_format(ff);

		if(!format){
			filename = gtk_file_chooser_get_filename(chooser);
		}

		if(!format){
			format = gdk_pixbuf_get_file_info(filename, NULL, NULL);
		}
		if(!format){
			format = get_gdk_pixbuf_format_by_extension(filename);
		}

		if(format)
			writable = gdk_pixbuf_format_is_writable(format);

		if(filename)
			g_free(filename);

		if(!writable){
			message_box_ok(GTK_MESSAGE_ERROR, _(bad_pixbuf_format_msg));
			g_signal_stop_emission_by_name(widget, "response");
		}
	}

}

static GtkFileFilter *
file_create_filter(GdkPixbufFormat *format, GtkFileChooserAction action)
{
	GtkFileFilter *filter = NULL;
	char *desc, *ext, *name;
	char **exts;
	int i;

	if(!format)
		return NULL;

	if(GTK_FILE_CHOOSER_ACTION_SAVE == action){
		if(!gdk_pixbuf_format_is_writable(format))
			return NULL;
	}
	filter = gtk_file_filter_new();

	desc = gdk_pixbuf_format_get_description(format);
	exts = gdk_pixbuf_format_get_extensions(format);
	ext = g_strdup("");
	for(i = 0; exts[i] != NULL; i++){
		char *p1, *p2, *pattern;

		pattern = g_strdup_printf("*.%s", exts[i]);
		gtk_file_filter_add_pattern(filter, pattern);
		g_free(pattern);

		if(exts[i + 1] != NULL)
			p1 = g_strdup_printf("*.%s, ", exts[i]);
		else
			p1 = g_strdup_printf("*.%s", exts[i]);

		p2 = g_strconcat(ext, p1, NULL);
		g_free(p1);
		if(p2){
			g_free(ext);
			ext = p2;
		}
	}
	name = g_strdup_printf("%s(%s)", desc, ext);
	gtk_file_filter_set_name(filter, name);
	g_object_set_data(G_OBJECT(filter), FILE_FORMAT_KEY,(gpointer)format);

	g_free(name);
	g_free(ext);
	g_strfreev(exts);
	g_free(desc);

	return filter;
}

static void file_add_file_filters(GtkWidget *dlg, GtkFileChooserAction action)
{
	GSList *p;
	GtkFileFilter *filter;

	if(!_pbformats){
		_pbformats = gdk_pixbuf_get_formats();
		g_object_set_data_full(G_OBJECT(dibuja_get_canvas()),
							   "pixbuf-formats",
							  (gpointer)_pbformats,
							  (GDestroyNotify)g_slist_free);
	}
	p = _pbformats;

	while(p){
		filter = file_create_filter((GdkPixbufFormat *)p->data, action);
		if(filter){
			gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dlg), filter);
		}

		p = p->next;
	}
}

gchar *file_open_dialog(GtkWidget *widget)
{
	GtkWidget *dialog = NULL;
	GtkFileFilter *filter;
	gchar *filename = NULL;
	gchar *cur_dir = NULL;
	GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_OPEN;
											/* Dialog title */
	dialog = gtk_file_chooser_dialog_new(_("Open File"),
										  GTK_WINDOW(gtk_widget_get_toplevel(widget)),
										  action,
										  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
										  GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
										  NULL);
	
	filter = gtk_file_filter_new();

	gtk_file_filter_set_name(filter, _("Supported Formats"));
	gtk_file_filter_add_pixbuf_formats(filter);
	gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);

	file_add_file_filters(dialog, action);

	cur_dir = file_get_cur_dir();
	if(cur_dir){
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), cur_dir);
		g_free(cur_dir);
	}

	if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT){

		filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	}

	gtk_widget_destroy(dialog);

	return filename;
}

GdkPixbuf *file_load_pixbuf(char *filename, gboolean warn)
{
	GError *error = NULL;
	GdkPixbuf *pixbuf = NULL;

	pixbuf = gdk_pixbuf_new_from_file((const char *)filename, &error);
	if(NULL != error){
		g_message("[%d] Unable to read pixbuf file: %s.", __LINE__, error->message);
		if(warn)
			message_box_error(error->message);
		g_error_free(error);
	}
	else if(!GDK_IS_PIXBUF(pixbuf)) {
		g_message("[%d] Unable to read pixbuf file: %s\n", __LINE__, filename);
		if(warn)
			message_box_error("Unable to read pixbuf file: %s\n", filename);
	}

	return pixbuf;
}

int file_open_pixbuf(char *filename)
{
	GdkPixbufFormat *pbformat;
	gchar *format = NULL;
	gchar *pbfilename = NULL;
	GdkPixbuf *pixbuf = NULL;
	int ret = 0;

	if(!filename) return 0;

	pbfilename = g_filename_from_uri((const gchar *)filename, NULL, NULL);
	if(!pbfilename){
		pbfilename = g_strdup(filename);
	}

	pixbuf = file_load_pixbuf(pbfilename, TRUE);
	if(GDK_IS_PIXBUF(pixbuf)){
		pbformat = gdk_pixbuf_get_file_info(pbfilename, NULL, NULL);
		format = g_strdup(gdk_pixbuf_format_get_name(pbformat));

		dibuja_new_file_from_pixbuf(pixbuf, pbfilename, format);
		dibuja_set_pixbuf_format(pbformat);
		g_object_unref(pixbuf);
	
		file_status = FILE_EXISTING;
		ret = 1;
	}

	if(format)
		g_free(format);
	if(pbfilename)
		g_free(pbfilename);

	return ret;
}

int file_open(GtkWidget *widget)
{
	gchar *filename = NULL;
	int ret = 0;

	if(file_status & FILE_NEED_SAVE){
		int resp = message_box_yes_no_cancel(_("The current file \"%s\" needs saving.\n"
											"Save?"),
											 dibuja_get_filename());
		if(GTK_RESPONSE_CANCEL == resp)
			return ret;
	}

	filename = file_open_dialog(widget);
	if(NULL != filename){
		GtkWidget *main_wnd = GTK_WIDGET(gtk_widget_get_toplevel(dibuja_get_canvas()));
		GdkCursor *cursor = gdk_window_get_cursor(main_wnd->window);

		canvas_set_wait_cursor(main_wnd);
		canvas_set_wait_cursor(dibuja_get_canvas());

		ret = file_open_pixbuf(filename);
		g_free(filename);

		canvas_set_cursor(dibuja_get_canvas(), DIBUJA_CURSOR);
		gdk_window_set_cursor(main_wnd->window, cursor);
	}

	return ret;
}

static int file_save_pixbuf(char *filename, char *format)
{
	GdkPixbuf *pixbuf;
	cairo_surface_t *surface;

	surface = dibuja_get_surface();

#ifdef HAVE_EXIV2
	uni_read_exiv2_to_cache(filename);
#endif // HAVE_EXIV2

	pixbuf = gdk_pixbuf_get_from_surface(surface, 0, 0, 
								cairo_image_surface_get_width(surface),
								cairo_image_surface_get_height(surface));

	gdk_pixbuf_save(pixbuf, filename, format, NULL, NULL);

#ifdef HAVE_EXIV2
	uni_write_exiv2_from_cache(filename);
#endif // HAVE_EXIV2

	g_object_unref(pixbuf);

	return 0;
}

static int file_save_as_pixbuf(char *filename, char *as_filename, char *format)
{
	GdkPixbuf *pixbuf;
	cairo_surface_t *surface;

	surface = dibuja_get_surface();

#ifdef HAVE_EXIV2
	if(g_file_test(filename, G_FILE_TEST_IS_REGULAR))
		uni_read_exiv2_to_cache(filename);
#endif // HAVE_EXIV2

	pixbuf = gdk_pixbuf_get_from_surface(surface, 0, 0, 
								cairo_image_surface_get_width(surface),
								cairo_image_surface_get_height(surface));

	if(gdk_pixbuf_save(pixbuf, as_filename, format, NULL, NULL)){
		GdkPixbufFormat *pbformat;

		pbformat = gdk_pixbuf_get_file_info(as_filename, NULL, NULL);
		dibuja_set_pixbuf_format(pbformat);
	}

#ifdef HAVE_EXIV2
	if(g_file_test(filename, G_FILE_TEST_IS_REGULAR))
		uni_write_exiv2_from_cache(as_filename);
#endif // HAVE_EXIV2

	g_object_unref(pixbuf);

	return 0;
}

int file_save_dialog(GtkWidget *widget, gboolean save_as)
{
	GtkWidget *dialog;
	// TODO: get current user
	gchar *filename = NULL;
	gchar *format = NULL;
	GdkPixbufFormat *pbformat;
	int response = 0;
	char by_extension[] = N_("By extension");
	GtkFileChooserAction action = GTK_FILE_CHOOSER_ACTION_SAVE;
	GtkFileChooser *chooser;
	GtkFileFilter *ff = NULL;

	if(!(file_status & FILE_NEW) && !save_as){
		pbformat = dibuja_get_pixbuf_format();
		if(!gdk_pixbuf_format_is_writable(pbformat)){
			message_box_ok(GTK_MESSAGE_ERROR, _(bad_pixbuf_format_msg));
//printf("*********************[%s - %s(): %d]\n", __FILE__, __func__, __LINE__);
			return 0;
		}
		response = file_save_pixbuf(dibuja_get_filename(),
									dibuja_get_file_format());
		file_status = FILE_EXISTING;
		dibuja_set_title();

		return response;
	}

	dialog = gtk_file_chooser_dialog_new(!save_as ? _("Save File") : _("Save File As"),
										 GTK_WINDOW(gtk_widget_get_toplevel(widget)),
										 action,
										 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
										 GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
										 NULL);
	chooser = GTK_FILE_CHOOSER(dialog);

	g_signal_connect(G_OBJECT(chooser), "response",
					 G_CALLBACK(save_dlg_response), NULL);

	ff = gtk_file_filter_new();
	gtk_file_filter_set_name(ff, _(by_extension));
	g_object_set_data(G_OBJECT(ff), FILE_FORMAT_KEY, NULL);
	gtk_file_chooser_add_filter(chooser, ff);

	file_add_file_filters(dialog, action);

	gtk_file_chooser_set_do_overwrite_confirmation(chooser, TRUE);
//	gtk_file_chooser_set_filename(chooser, dibuja_get_filename());
	gtk_file_chooser_set_filter(chooser, ff);

	if(file_status & FILE_NEW){
		gchar *default_folder_for_saving = file_def_save_folder();
		gtk_file_chooser_set_current_folder(chooser,
											default_folder_for_saving);
		gtk_file_chooser_set_current_name(chooser, dibuja_get_filename());
		g_free(default_folder_for_saving);
	}
	else{
		char *base, *folder;
		filename = dibuja_get_filename();
		base = g_path_get_basename(filename);
		gtk_file_chooser_set_current_name(chooser, base);
		folder = g_path_get_dirname(filename);
		gtk_file_chooser_set_current_folder(chooser, folder);
		g_free(base);
		g_free(folder);
	}

	if((response = gtk_dialog_run(GTK_DIALOG(dialog))) != GTK_RESPONSE_ACCEPT){
		gtk_widget_destroy(dialog);
		return 0;
	}

	filename = gtk_file_chooser_get_filename(chooser);

	ff = gtk_file_chooser_get_filter(chooser);

	if(g_strstr_len(gtk_file_filter_get_name(ff), -1, _(by_extension))){
		pbformat = get_gdk_pixbuf_format_by_extension(filename);
	}
	else{
		pbformat = file_get_pb_format(ff);
	}

	if(pbformat){
		format = gdk_pixbuf_format_get_name(pbformat);
	}
	else{
		format = g_strdup("png");
	}

	gtk_widget_destroy(dialog);

	file_save_as_pixbuf(dibuja_get_filename(), filename, format);
	file_status = FILE_EXISTING;
	dibuja_set_filename(filename, format);
	dibuja_set_title();

	g_free(filename);
	g_free(format);

	return 1;
}

void file_new_file(void)
{
	if(file_status & FILE_NEED_SAVE){
		int ret = message_box_yes_no_cancel(_("The current file \"%s\" needs saving.\n"
											"Save?"),
											dibuja_get_filename());
		if(GTK_RESPONSE_CANCEL == ret)
			return;
	}
	dibuja_new_file(NULL, NULL, CANVAS_DEF_WIDTH, CANVAS_DEF_HEIGHT);
}

void filesize_type_combo_active(GtkComboBox *combo, gpointer data)
{
	gint index = gtk_combo_box_get_active(combo);
	guint digits = 2;
	gdouble step = .10, page = 10.;
	
	GtkWidget *spin_width = get_widget("new_image_width_spin");
	GtkWidget *spin_height = get_widget("new_image_height_spin");

	if(0 == index){
		digits = 0;
		step = 1.;
	}

	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin_width), digits);
	gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin_height), digits);

	gtk_spin_button_set_increments(GTK_SPIN_BUTTON(spin_width), step, page);
	gtk_spin_button_set_increments(GTK_SPIN_BUTTON(spin_height), step, page);
}

void file_new_file_at_size(void)
{
	gint result;
	GtkWidget *dialog = get_widget("dialog_new_image");
	GtkWidget *spin_width = get_widget("new_image_width_spin");
	GtkWidget *spin_height = get_widget("new_image_height_spin");
	GtkComboBox *combo_box = GTK_COMBO_BOX(get_widget("filesize_type_combo"));
	gdouble width, height;
	gint meas;

	if(file_status & FILE_NEED_SAVE){
		int ret = message_box_yes_no_cancel(_("The current file \"%s\" needs saving.\n"
											"Save?"),
											dibuja_get_filename());
		if(GTK_RESPONSE_CANCEL == ret)
			return;
	}

	width = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spin_width));
	if(width < 1.0){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_width), 600.0);
	}
	height = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spin_height));
	if(height < 1.0){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin_height), 400.0);
	}
	/* On reshow, one of the buttons has the focus */
	gtk_widget_grab_focus(spin_width);

	result = gtk_dialog_run(GTK_DIALOG(dialog));
	if(GTK_RESPONSE_OK == result){
		width = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spin_width));
		height = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spin_height));
		meas = gtk_combo_box_get_active(GTK_COMBO_BOX(combo_box));
		if(width > 0 && height > 0){
			gdouble xres, yres;
			gdouble mm_per_in = 25.4;

			gimp_get_screen_resolution(gdk_display_get_default_screen(gdk_display_get_default()),
							   &xres, &yres);
			switch(meas){
				case 0:
					//pixels
					{
					int w, h;
					width = (double)(w = (int)width);
					height = (double)(h = (int)height);
					}
					break;
				case 1:
					// Convert inches to pixels
					width = ROUND(width * xres);
					height = ROUND(height * yres);
					break;
				case 2:
					// Convert centimeters to pixels
					width = ROUND((double)width / mm_per_in * xres * 10.);
					height = ROUND((double)height / mm_per_in * yres * 10.);
					printf("width: %f, height: %f\n", width, height);
					break;
				default:
					return;
			}
			dibuja_new_file(NULL, NULL, width, height);
		}
	}

	gtk_widget_hide(dialog);
}

