/* 
 * Flood fill routine copied from gpaint,
 * (c) 2000 -2003, 2007 Li-Cheng (Andy) Tai, Micheal A. Meffie III
 * 
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "tools.h"
#include "flood-fill.h"
#include "dibuja.h"
#include "util.h"
#include "draw-funcs.h"
#include "file.h"
#include "undo.h"

#define STACKSIZE 10000

struct fillinfo
{
   unsigned char *rgb; 
   int rowstride;
   int pixelsize; 
   int width;
   int height;
   unsigned char or, og, ob, r, g, b;
};

struct fillpixelinfo
{
   int y, xl, xr, dy;
};

static gboolean flood_fill_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean flood_fill_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);

static __inline__ void set_new_pixel_value(struct fillinfo *info, int x, int y);
static void flood_fill_algo(struct fillinfo *info, int x, int y);
static void fill_bounded_area(cairo_surface_t *surface, int x, int y);
static __inline__ int is_old_pixel_value(struct fillinfo *info, int x, int y);

static cairo_surface_t *tmp_surface = NULL;
static int xpos = 0;
static int ypos = 0;

static int xmin, ymin, xmax, ymax;

static DrawTool tool = {
				flood_fill_button_press,
				NULL,
				flood_fill_button_release,
				NULL,
				NULL,
				NULL
			};


#define GPAINT_FILL(tool)  ((gpaint_fill*)(tool))

#define PUSH(py, pxl, pxr, pdy) \
{ \
    struct fillpixelinfo *p = sp;\
    if (((py) + (pdy) >= 0) && ((py) + (pdy) < info->height))\
    {\
        p->y = (py);\
        p->xl = (pxl);\
        p->xr = (pxr);\
        p->dy = (pdy);\
        sp++; \
    }\
}
   
#define POP(py, pxl, pxr, pdy) \
{\
    sp--;\
    (py) = sp->y + sp->dy;\
    (pxl) = sp->xl;\
    (pxr) = sp->xr;\
    (pdy) = sp->dy;\
}

static void
fill_bounded_area(cairo_surface_t *surface, int x, int y)
{
    GdkColor color;
    unsigned char r, g, b;
    struct fillinfo fillinfo;
    unsigned char *p ;

    dibuja_get_fg_color(&color);

    r = color.red / 256;
    g = color.green / 256;
    b = color.blue / 256;

    fillinfo.rgb = cairo_image_surface_get_data (surface);     
    fillinfo.width = cairo_image_surface_get_width (surface);
    fillinfo.height = cairo_image_surface_get_height (surface);
    fillinfo.rowstride = cairo_image_surface_get_stride (surface);
    fillinfo.pixelsize = sizeof(guint32);

    fillinfo.r = r;
    fillinfo.g = g;
    fillinfo.b = b;
    p = fillinfo.rgb + y * fillinfo.rowstride + x * fillinfo.pixelsize;

    fillinfo.ob = *p;
    fillinfo.og = *(p + 1);
	fillinfo.or = *(p + 2);

    flood_fill_algo(&fillinfo, x, y);
}


static __inline__ int
is_old_pixel_value(struct fillinfo *info, int x, int y)
{
    unsigned char *p = info->rgb + y * info->rowstride + x * info->pixelsize;
    unsigned char or, og, ob;
    ob = *p;
    og = *(p + 1);
    or = *(p + 2);
    if ((or == info->or) && (og == info->og) && (ob == info->ob))
    {
        return 1;
    }
    return 0;
}


static __inline__ void
set_new_pixel_value(struct fillinfo *info, int x, int y)
{
    unsigned char *p = info->rgb + y * info->rowstride + x * info->pixelsize;
    *p = info->b;
    *(p + 1) = info->g;
    *(p + 2) = info->r;

	if(x < xmin) xmin = x;
	if(x > xmax) xmax = x;
	if(y < ymin) ymin = y;
	if(y > ymax) ymax = y;
}

/*
 * algorithm based on SeedFill.c from GraphicsGems 1
 */
static void
flood_fill_algo(struct fillinfo *info, int x, int y)
{
    /* TODO: check for stack overflow? */
    /* TODO: that's a lot of memory! esp if we never use it */
    struct fillpixelinfo stack[STACKSIZE];
    struct fillpixelinfo * sp = stack;
    int l, x1, x2, dy;
      
    if ((x >= 0) && (x < info->width) && (y >= 0) && (y < info->height))
    {
        if ((info->or == info->r) && (info->og == info->g) && (info->ob == info->b))
        {
            return;
        }
        PUSH(y, x, x, 1);
        PUSH(y + 1, x, x, -1);
        while (sp > stack)  
        {
            POP(y, x1, x2, dy);
            for (x = x1; (x >= 0) && is_old_pixel_value(info, x, y); x--)
            { 
                set_new_pixel_value(info, x, y);
            }
            if (x >= x1)
            {
                goto skip;
            }
            l = x + 1;
            if (l < x1)
            {
                PUSH(y, l, x1 - 1, -dy);
            }
            x = x1 + 1;
            do
            {
                for (; (x < info->width) && is_old_pixel_value(info, x, y); x++)
                {
                    set_new_pixel_value(info, x, y);
                }
                PUSH(y, l, x - 1, dy);
                if (x > x2 + 1)
                {
                    PUSH(y, x2 + 1, x - 1, -dy);
                }
skip:
                for (x++; x <= x2 && !is_old_pixel_value(info, x, y); x++) 
                {
                    /* empty */ ;
                }
                l = x;
            } while (x <= x2);
        }
    }
}  



DrawTool *flood_fill_get_tool(void)
{
	return &tool;
}



static gboolean flood_fill_button_press(GtkWidget *canvas, GdkEventButton *event, gpointer data)
{

	if(1 == event->button){
		GdkRectangle copy;

		copy.x = copy.y = 0;
		copy.width = cairo_image_surface_get_width(dibuja_get_surface());
		copy.height = cairo_image_surface_get_height(dibuja_get_surface());
		tmp_surface = dibuja_copy_from_surface(&copy);
 
		xpos = event->x;
		ypos = event->y;

		xmin = xmax = xpos;
		ymin = ymax = ypos;
	}
	return TRUE;
}

static gboolean flood_fill_button_release(GtkWidget *canvas, GdkEventButton *event, gpointer data)
{

	if(1 == event->button){
		if(xpos == (int)event->x && ypos == (int)event->y){
			GdkRectangle update;
			cairo_surface_t *copy_from_tmp;

			fill_bounded_area(dibuja_get_surface(), xpos, ypos);

			rect_from_pts(&update, xmin, ymin, xmax, ymax);
			resize_rect(&update, 2, 2);//  + 2 for alias

			copy_from_tmp = draw_copy_area(tmp_surface, update.x, update.y, update.width, update.height);
			undo_add_from_surface(SIMPLE_CANVAS(canvas), copy_from_tmp, update, _("Flood Fill"));
			
			gtk_widget_queue_draw(canvas);
		}
	}

	if(tmp_surface){
		cairo_surface_destroy(tmp_surface);
		tmp_surface = NULL;
	}
	
	return TRUE;
}



