/* 
 * (c) 2015 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib/gi18n.h>
#include <string.h>
#include <ctype.h>

#include "fx-dlg.h"
#include "effects.h"

#define EFFECTS_DIALOG_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), EFFECTS_TYPE_DIALOG, EffectsDialogPrivate))

typedef void (*ResetFunc)(GtkWidget *widget, EffectsDialog *ed);


static const char *handled_ops[] = {
		"gegl:brightness-contrast", "gegl:gaussian-blur",
		"gegl:threshold", "gegl:vignette",  "gegl:posterize",
		"gegl:unsharp-mask", NULL
		};

static gboolean disp_expose(GtkWidget *disp, GdkEventExpose *event, EffectsDialog *ed);

static const cairo_format_t deffmt = CAIRO_FORMAT_ARGB32;

enum {
	DEFAULT_CLICKED,
	UPDATE_IMAGE,
	LAST_SIGNAL
};

//typedef struct _EffectsDialogPrivate EffectsDialogPrivate;

struct _EffectsDialogPrivate
{
	GtkWidget *table;
	GtkWidget *disp;
	GeglNode *node;
	GeglNode *op_node;
	double scale;
	guchar *buf;
	int imgw;
	int imgh;
	int freeze_update;
};


G_DEFINE_TYPE (EffectsDialog, effects_dialog, GTK_TYPE_DIALOG)

static guint effects_dialog_signals[LAST_SIGNAL] = { 0 };

static guchar *create_buf(int w, int h)
{
	guchar *b = NULL;
	int stride;

	stride = cairo_format_stride_for_width(deffmt, w);
	b = g_malloc(stride * h);

	return b;
}

static void effects_dialog_finalize(GObject *object)
{
	EffectsDialog *ed = EFFECTS_DIALOG(object);
	if(GEGL_IS_NODE(ed->priv->node)){
		g_object_unref(ed->priv->node);
	}
	if(ed->priv->buf){
		g_free(ed->priv->buf);
		ed->priv->buf = NULL;
	}

	G_OBJECT_CLASS (effects_dialog_parent_class)->finalize (object);
}

static void
effects_dialog_class_init (EffectsDialogClass *class)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (class);

	object_class->finalize = effects_dialog_finalize;

	effects_dialog_signals[DEFAULT_CLICKED] =
	  g_signal_new ("default-clicked",
		  G_OBJECT_CLASS_TYPE (object_class),
		  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
		  G_STRUCT_OFFSET (EffectsDialogClass, default_clicked),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);

	effects_dialog_signals[UPDATE_IMAGE] =
	  g_signal_new ("update-image",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (EffectsDialogClass, update_image),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

	g_type_class_add_private (object_class,
			    sizeof (EffectsDialogPrivate));
}

static void
effects_dialog_init (EffectsDialog *dialog)
{
	EffectsDialogPrivate *priv;

	priv = EFFECTS_DIALOG_GET_PRIVATE (dialog);
	dialog->priv = priv;

	dialog->priv->table  = NULL;
	dialog->priv->disp   = NULL;
	dialog->priv->node   = NULL;
	dialog->priv->scale  = 1.;
	dialog->priv->buf    = NULL;
	dialog->priv->imgw   = 0;
	dialog->priv->imgh   = 0;
	dialog->priv->freeze_update = 0;
}

void table_for_each(GtkWidget *widget, EffectsDialog *ed)
{
	ResetFunc reset_func = g_object_get_data(G_OBJECT(widget), "reset-func");

	if(reset_func){
		reset_func(widget, ed);
	}
}

static void button_reset(GtkWidget *button, EffectsDialog *ed)
{
	g_return_if_fail(EFFECTS_IS_DIALOG(ed));

	ed->priv->freeze_update = 1;

	gtk_container_forall(GTK_CONTAINER(ed->priv->table),
						 (GtkCallback)table_for_each, (gpointer)ed);
	ed->priv->freeze_update = 0;
	gtk_widget_queue_draw(ed->priv->disp);
}

void scale_set_int_prop(GtkRange *range, EffectsDialog *ed)
{
	int val = (int)gtk_range_get_value(range);
	char *prop = g_object_get_data(G_OBJECT(range), "prop");

	if(prop){
		gegl_node_set(ed->priv->op_node, prop, val, NULL);
		gtk_widget_queue_draw(ed->priv->disp);
	}
}

void scale_set_double_prop(GtkRange *range, EffectsDialog *ed)
{
	double val = gtk_range_get_value(range);
	char *prop = g_object_get_data(G_OBJECT(range), "prop");

	if(prop){
		gegl_node_set(ed->priv->op_node, prop, val, NULL);
		gtk_widget_queue_draw(ed->priv->disp);
	}
}

void scale_reset(GtkWidget *w, EffectsDialog *ed)
{
	double *def_val = g_object_get_data(G_OBJECT(w), "def-val");

	if(def_val)
		gtk_range_set_value(GTK_RANGE(w), *def_val);
}

void color_button_set(GtkColorButton *button, EffectsDialog *ed)
{
	GdkColor color;
	char *prop = g_object_get_data(G_OBJECT(button), "prop");

	if(prop){
		GeglColor *gc;
		guint16 alpha;
		char *str = NULL;

		gtk_color_button_get_color(button, &color);
		alpha = gtk_color_button_get_alpha(button);
		str = g_strdup_printf("rgba(%f,%f,%f,%f)",
							  (double)color.red / 65535.,
							  (double)color.green / 65535.,
							  (double)color.green / 65535.,
							  (double)alpha / 65535.
							  );
		gc = gegl_color_new(str);
		g_free(str);
		gegl_node_set(ed->priv->op_node, prop, gc, NULL);
		gtk_widget_queue_draw(ed->priv->disp);
	}
}

void vignette_shape_toggled(GtkToggleButton *button, EffectsDialog *ed)
{
	int *shape = NULL;

	if(gtk_toggle_button_get_active(button)){
		shape = (int *)g_object_get_data(G_OBJECT(button), "vignette-shape");
		if(shape){
			gegl_node_set(ed->priv->op_node, "shape", *shape, NULL);
			gtk_widget_queue_draw(ed->priv->disp);
		}
	}
}

void set_tool_tip(EffectsDialog *ed, GtkWidget *widget, char *prop)
{
	GParamSpec *pspec;

	pspec = gegl_node_find_property(ed->priv->op_node, prop);
	if(!pspec)
		g_warning("no property named \"%s\"\n", prop);

	gtk_widget_set_tooltip_text(widget, g_param_spec_get_blurb(pspec));
}

GtkWidget *create_label(EffectsDialog *ed, char *prop)
{
	GtkWidget *label;
	GParamSpec *pspec;
	char *text = NULL;

	pspec = gegl_node_find_property(ed->priv->op_node, prop);
	if(!pspec)
		g_warning("no property named %s\n", prop);
	text = g_strdup_printf("%s:", g_param_spec_get_nick(pspec));
	label = gtk_label_new(text);
	g_free(text);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 1.);

	return label;
}

GtkWidget *create_hscale2(EffectsDialog *ed, double min, double max, double step, double val,
						 char *prop, double def_val,
						 GCallback cb, gpointer data)
{
	char *p;
	double *dv;
	GtkWidget *scale = gtk_hscale_new_with_range (min, max, step);

	g_signal_connect(G_OBJECT(scale), "value-changed", cb, data);
	gtk_range_set_value(GTK_RANGE(scale), val);

	g_object_set_data(G_OBJECT(scale), "reset-func", (gpointer)scale_reset);
	p = g_strdup(prop);
	g_object_set_data_full(G_OBJECT(scale), "prop", (gpointer)p, g_free);

	dv = g_malloc(sizeof(double));
	*dv = def_val;
	g_object_set_data_full(G_OBJECT(scale), "def-val", (gpointer)dv, g_free);

	set_tool_tip(ed, scale, prop);

	return scale;
}

static void brightness_contrast(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "brightness"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, -3., 3., .10, 0., "brightness", 0.,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "contrast"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, -5., 5., .10, 1., "contrast", 1.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
}

static void gaussian_blur(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "std-dev-x"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 0., 100., 1., 4., "std-dev-x", 0.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "std-dev-y"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 0., 100., 1., 4., "std-dev-y", 0.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
}

static void threshold(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "value"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, -10., 10., .5, .5, "value", .5,
				   G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
}

static void posterize(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "levels"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 1., 64., 1., .8, "levels", 8.,
				   G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	g_signal_handlers_disconnect_by_func(G_OBJECT(scale),
										 scale_set_double_prop, ed);
	g_signal_connect(G_OBJECT(scale), "value-changed",
					 G_CALLBACK(scale_set_int_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
}

static void vignette(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *radio1, *radio;
	GtkWidget *scale;
	int xpad = 6;
	int *shape = NULL;

	l++;

	radio1 = gtk_radio_button_new_with_label(NULL, "Circle");
	shape = g_malloc(sizeof(int));
	*shape = 0;
	g_object_set_data_full(G_OBJECT(radio1), "vignette-shape",
							(gpointer)shape, g_free);
	g_signal_connect(G_OBJECT(radio1), "toggled",
							G_CALLBACK(vignette_shape_toggled), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), radio1, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "shape"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	radio = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio1), "Diamond");
	shape = g_malloc(sizeof(int));
	*shape = 1;
	g_object_set_data_full(G_OBJECT(radio), "vignette-shape",
							(gpointer)shape, g_free);
	g_signal_connect(G_OBJECT(radio), "toggled", 
							G_CALLBACK(vignette_shape_toggled), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), radio, l, l + 1, t, t + 1);

	t++;
	radio = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(radio1), "Square");
	shape = g_malloc(sizeof(int));
	*shape = 2;
	g_object_set_data_full(G_OBJECT(radio), "vignette-shape",
							(gpointer)shape, g_free);
	g_signal_connect(G_OBJECT(radio), "toggled", 
							G_CALLBACK(vignette_shape_toggled), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), radio, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "color"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = gtk_color_button_new();
	gtk_color_button_set_use_alpha(GTK_COLOR_BUTTON(scale), TRUE);
	g_object_set_data_full(G_OBJECT(scale), "prop", g_strdup("color"), g_free);
	g_signal_connect(G_OBJECT(scale), "color-set", G_CALLBACK(color_button_set),
							(gpointer)ed);
	gtk_table_attach(GTK_TABLE(table), scale,
							  l, l + 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	set_tool_tip(ed, scale, "color");

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "radius"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, 0., 3., .05, 1.5,
						   "radius", 1.5,
						 G_CALLBACK(scale_set_double_prop), ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "softness"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, 0., 1., .05, .8,
						   "softness", 0.8,
						   G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "gamma"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, 1., 20., .5, 2.,
						  "gamma", 2.,
						  G_CALLBACK(scale_set_double_prop),
									 (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "proportion"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, 0., 1., .01, 1.,
						  "proportion", 1.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "squeeze"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, -1., 1., .05, 0.,
							"squeeze", 0.,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "x"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, -1, 2, .5, .5,
							"x", .5,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "y"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, -1, 2, .5, .5,
							"y", .5,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "rotation"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, xpad, 0);
	scale = create_hscale2(ed, 0., 359., 1., 0., 
							"rotation", 0.,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
	gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_RIGHT);
}

/* Unsharp Mask */
static void sharpen(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "std-dev"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed,
						   0., //  min
						   300., // max
						   .01, // step
						   3., // val
						   "std-dev", 0.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "scale"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 
						   0., //  min
						   300., // max
						   .01, // step
						   .5, // val
						   "scale", 0.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
#if 0
	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "threshold"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 
						   0., //  min
						   1., // max
						   .10, // step
						   0., // val
						   "threshold", 0.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);
#endif
}

/*
static void mantiuk06(EffectsDialog *ed, GtkWidget *table, int l, int t)
{
	GtkWidget *scale, *label;

	l++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "contrast"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 0., 1., .10, .10, "contrast", .10,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "saturation"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 0., 2., .10, 80., "saturation", 80.,
							G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++;
	gtk_table_attach(GTK_TABLE(table), create_label(ed, "detail"),
							  0, 1, t, t + 1, GTK_FILL, GTK_FILL, 6, 0);
	scale = create_hscale2(ed, 1., 99., 1., 1., "detail", 1.,
						  G_CALLBACK(scale_set_double_prop), (gpointer)ed);
	gtk_table_attach_defaults(GTK_TABLE(table), scale, l, l + 1, t, t + 1);

	t++; t++;
	label = gtk_label_new("Depending on your computer's processing power,\n"
						  "and the size of the image, this could be a slow\n"
						  "operation. Please be patient.");
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 2, t, t + 1);
}
*/
static void add_widgets(EffectsDialog *ed, GtkWidget *table, char *op)
{

	if(strcmp(op, "gegl:brightness-contrast") == 0)
		brightness_contrast(ed, table, 0, 0);
	else if(strcmp(op, "gegl:gaussian-blur") == 0)
		gaussian_blur(ed, table, 0, 0);
	else if(strcmp(op, "gegl:threshold") == 0)
		threshold(ed, table, 0, 0);
	else if(strcmp(op, "gegl:vignette") == 0)
		vignette(ed, table, 0, 0);
	else if(strcmp(op, "gegl:posterize") == 0)
		posterize(ed, table, 0, 0);
	else if(strcmp(op, "gegl:unsharp-mask") == 0)
		sharpen(ed, table, 0, 0);
	else
		g_warning("Unknown operation %s\n", op);
}

static void dlg_set_title(GtkWindow *window, char *op)
{
	char *title;

	title = (char *)gegl_operation_get_key(op, "title");

	if(strcmp("gegl:gaussian-blur", op) == 0 && !title){
		gtk_window_set_title(window, _("Gaussian Blur"));
		return;
	}

	if(!title)
		title = op;
		
	if(title)
		gtk_window_set_title(window, title);
}

static gboolean create_nodes(EffectsDialog *ed, char *op, cairo_surface_t *s)
{
	EffectsDialogPrivate *priv = ed->priv;
	GeglNode *root, *photo;
	GeglBuffer *src_buffer;

	src_buffer = gegl_buffer_from_surface (s);
	if(!GEGL_IS_BUFFER(src_buffer))
		return FALSE;

	priv->node = gegl_node_new();

	root = gegl_node_create_child(priv->node, "gegl:over");
	priv->op_node = gegl_node_create_child(priv->node, op);

	gegl_node_connect_to(priv->op_node , "output", root, "input");

	photo = gegl_node_new_child (priv->node, "operation", "gegl:buffer-source",
								 "buffer", src_buffer, NULL);

	gegl_node_connect_to(photo , "output", priv->op_node, "input");

	g_object_unref(src_buffer);

	return TRUE;
}

gboolean fx_is_handled_op(char *op)
{
	int i = 0;

	if(op){
		while(handled_ops[i] != NULL){
			if(strcmp(handled_ops[i], op) == 0){
				return TRUE;
			}
				i++;
		}
	}
	return FALSE;
}

GtkWidget *effects_dialog_new(GtkWindow *window,
							  char *op,
							  cairo_surface_t *surface)
{
	GtkWidget *dialog, *vbox, *button2, *hbox, *scroll, *widget ;
	EffectsDialogPrivate *priv;
	int expand = FALSE;
	int fill = FALSE;
	guint padding = 0;
	guint nrows = 2;
	guint ncols = 3;
	int homogeneous = FALSE;
	int w, h;

	g_return_val_if_fail(fx_is_handled_op(op), NULL);
	g_return_val_if_fail(surface != NULL, NULL);

	w = cairo_image_surface_get_width(surface);
	h = cairo_image_surface_get_height(surface);
	g_return_val_if_fail(w > 0 && h > 0, NULL);

	dialog = g_object_new ( EFFECTS_TYPE_DIALOG, NULL);
	priv = EFFECTS_DIALOG_GET_PRIVATE (dialog);

	if(!create_nodes(EFFECTS_DIALOG(dialog), op, surface)){
		gtk_widget_destroy(dialog);
		return NULL;
	}

	priv->buf = create_buf(w, h);
	priv->imgw = w;
	priv->imgh = h;

	dlg_set_title (GTK_WINDOW (dialog), op);

	gtk_window_set_transient_for (GTK_WINDOW (dialog),
	                              GTK_WINDOW (window));
	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);

	vbox = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	gtk_box_set_homogeneous(GTK_BOX(vbox), FALSE);
	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, padding);
	widget = gtk_alignment_new(.5, .5, 0., 0.);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scroll), widget);

	priv->disp = gtk_drawing_area_new();
	g_signal_connect(G_OBJECT(priv->disp), "expose-event",
					 G_CALLBACK(disp_expose), (gpointer)dialog);
	gtk_widget_show(priv->disp);
	gtk_widget_set_size_request(priv->disp, w, h);
	gtk_container_add(GTK_CONTAINER(widget), priv->disp);

	priv->table = gtk_table_new(nrows, ncols, homogeneous);

	gtk_box_pack_start(GTK_BOX(vbox), priv->table, expand, fill, padding);
	gtk_widget_show (priv->table);

	add_widgets(EFFECTS_DIALOG(dialog), priv->table, op);

	hbox = gtk_hbox_new(homogeneous, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, expand, fill, 12);

	expand = TRUE; fill = TRUE;
	button2 = gtk_button_new_with_label("Reset");
	gtk_button_set_image(GTK_BUTTON(button2),
							gtk_image_new_from_stock(GTK_STOCK_REVERT_TO_SAVED,
												GTK_ICON_SIZE_SMALL_TOOLBAR));
	gtk_box_pack_start(GTK_BOX(hbox), button2, 0, fill, 12);

	g_signal_connect(G_OBJECT(button2), "clicked",
					 G_CALLBACK(button_reset), (gpointer)dialog);

//	widget = gtk_check_button_new_with_label("Preview");
//	gtk_box_pack_start(GTK_BOX(hbox), widget, expand, fill, 12);

	gtk_dialog_add_buttons(GTK_DIALOG(dialog),
						   GTK_STOCK_CANCEL , GTK_RESPONSE_CANCEL ,
						   GTK_STOCK_APPLY  , GTK_RESPONSE_APPLY  ,
						   NULL);

	gtk_window_set_default_size(GTK_WINDOW(dialog), 500, 500);
	gtk_widget_show_all(dialog);

	return dialog;
}
static void
fx_draw(EffectsDialog *ed, cairo_t *cr, GdkRectangle *r)
{
	cairo_surface_t *surface = NULL;
	GeglRectangle   gr;
	const Babl *fmt = babl_format("cairo-ARGB32");
	int stride;
	int imgw = ed->priv->imgw;
	int imgh = ed->priv->imgh;

	gr.x      = r->x;
	gr.y      = r->y;
	gr.width  = r->width;
	gr.height = r->height;

	if(gr.width > imgw){
		gr.width = imgw;
	}
	if(gr.height > imgh){
		gr.height = imgh;
	}

	if(!(gr.width > 0 && gr.height > 0)) return;

	if(!cr){
		g_warning("[%s: %d] Bad cr\n", __FILE__, __LINE__);
		return;
	}

	if(!ed->priv->buf){
		g_warning("[%s: %d] Could not allocate buffer\n", __FILE__, __LINE__);
		return;
	}

	if(!fmt){
		g_warning("[%s: %d] Bad fmt\n", __FILE__, __LINE__);
		return;
	}

	gegl_node_blit(ed->priv->op_node, 1., &gr, fmt, (gpointer)ed->priv->buf,
	               GEGL_AUTO_ROWSTRIDE, GEGL_BLIT_DEFAULT);

	stride = cairo_format_stride_for_width(deffmt, gr.width);

	surface = cairo_image_surface_create_for_data(ed->priv->buf, deffmt,
												gr.width, gr.height, stride);

//    cairo_scale(cr, .1, .1);
	if(surface){
	    cairo_set_source_surface(cr, surface, r->x, r->y);
	    gdk_cairo_rectangle(cr, r);
	    cairo_clip(cr);
	    cairo_paint(cr);

	    cairo_surface_destroy(surface);
	}
}

static gboolean disp_expose(GtkWidget *disp, GdkEventExpose *event, EffectsDialog *ed)
{
	cairo_t *cr;

	if(!gtk_widget_get_mapped(disp)) return TRUE;
	if(ed->priv->freeze_update)
		return TRUE;

	cr = gdk_cairo_create(disp->window);

	fx_draw(ed, cr, &(event->area));

	cairo_destroy(cr);

	return TRUE;
}

cairo_surface_t *fx_dlg_get_cairo_surface(EffectsDialog *ed)
{
	GdkRectangle r = {0};
	cairo_surface_t *surface;
	cairo_t *cr;

	r.width = ed->priv->imgw;
	r.height = ed->priv->imgh;

	surface = cairo_image_surface_create(deffmt, r.width, r.height);

	cr = cairo_create(surface);

	fx_draw(ed, cr, &r);

	cairo_destroy(cr);

	return surface;
}



