/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "tools.h"
#include "lines.h"
#include "dibuja.h"
#include "draw-funcs.h"
#include "undo.h"


static gboolean line_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean line_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data);
static gboolean line_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean line_expose(GtkWidget *widget, cairo_t *cr, gpointer data);
static void line_draw_line(cairo_t *cr, double sx, double sy, double ex, double ey);

static DrawTool tool = {
				line_button_press,
				line_motion,
				line_button_release,
				line_expose,
				NULL,
				NULL
			};

static GdkRectangle old, cur, start, arrowhead;
static int have_arrow = 0;

DrawTool *line_get_tool(void)
{
	return &tool;
}

gboolean line_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(1 == event->button){
		GdkRectangle tmp;
		int extra = dibuja_get_line_width() * 2;

		if((have_arrow = tools_get_draw_arrow()) != 0){
			cairo_t *cr = dibuja_get_cr(CR_WIDGET);

			draw_arrow(cr, event->x, event->y,
					   event->x, event->y, &arrowhead);
			cairo_destroy(cr);
			extra = MAX(arrowhead.width, arrowhead.height);

			tmp = start = arrowhead;
			simple_canvas_update_rect(SIMPLE_CANVAS(widget), &tmp);
		}
		else{
			rect_from_pts(&start, event->x - extra, event->y - extra,
					  event->x + extra, event->y + extra );
		}
		cur = old = start;
	}

	return TRUE;
}

gboolean line_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	GdkRectangle area = {0};

	if (event->state & GDK_BUTTON1_MASK){
		int extra = (dibuja_get_line_width() + 2);
		double x = event->x;
		double y = event->y;

		if(event->state & GDK_CONTROL_MASK){
			draw_constrain_pt(start.x + start.width / 2,
						   start.y + start.height / 2,
						   x, y, &x, &y);
		}
		rect_from_pts(&cur, x, y, x, y);

		gdk_rectangle_union(&start, &cur, &area);
		gdk_rectangle_union(&area, &old, &area);
		if(tools_get_draw_arrow())
			gdk_rectangle_union(&area, &arrowhead, &area);

		resize_rect(&area, extra * 2, extra * 2);

		old = cur;

		simple_canvas_update_rect(SIMPLE_CANVAS(widget), &area);
	}

	return TRUE;
}

gboolean line_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(1 == event->button){
		GdkRectangle area = {0};
		cairo_t *cr;
		int extra = dibuja_get_line_width();
		int sx, sy, ex, ey;

		sx = start.x + start.width / 2;
		sy = start.y + start.height / 2;
		ex = (int)event->x;
		ey = (int)event->y;

		if(event->state & GDK_CONTROL_MASK){
			double x, y;

			draw_constrain_pt((double)sx, (double)sy, (double)ex, (double)ey,
						   (double *)&x, (double *)&y);
			ex = (int)x;
			ey = (int)y;
		}

		if(ex == sx && sy == ey) return TRUE;

		rect_from_pts(&area, sx, sy, ex, ey);
		extra += 2;// antialias
		if(have_arrow){
			extra += MAX(arrowhead.width, arrowhead.height);
		}
		resize_rect(&area, extra, extra);
		undo_add(SIMPLE_CANVAS(widget), area, _("Line"));

		cr = dibuja_get_cr(CR_SURFACE);

		line_draw_line(cr, sx, sy, ex, ey);

		cairo_destroy(cr);

		//set_update_rect(&cur, sx, sy, ex, ey, simple_canvas_get_zoom(SIMPLE_CANVAS(widget)));

		simple_canvas_update_rect(SIMPLE_CANVAS(widget), &area);

		old.x = old.y = old.width = old.height = 0;
		arrowhead = cur = start = old;
		have_arrow = 0;
	}

	return TRUE;
}

static void line_draw_line(cairo_t *cr, double sx, double sy, double ex, double ey)
{
	if(have_arrow){
		draw_arrow(cr, sx, sy, ex, ey, &arrowhead);
	}
	else{
		draw_line(cr, sx, sy, ex, ey);
	}
}

static gboolean line_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	double sx, sy, ex, ey;

	ex = cur.x + cur.width / 2;
	ey = cur.y + cur.height / 2;
	sx = start.x + start.width / 2;
	sy = start.y + start.height / 2;

	line_draw_line(cr, sx, sy, ex, ey);

	return TRUE;
}

