/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <glib/gi18n.h>

#include "menu.h"
#include "dibuja.h"
#include "util.h"
#include "file.h"
#include "selection.h"
#include "simple-canvas.h"
#include "undo.h"
#include "color.h"
#include "rotate.h"
#include "resize.h"
#include "scale.h"
#include "effects.h"
#include "draw-funcs.h"
#include "brushes.h"

/* File Menu */
void menu_file_new(GtkWidget *widget, gpointer data)
{
	file_new_file();
}

void menu_file_at_size(GtkWidget *widget, gpointer data)
{
	file_new_file_at_size();
}

void menu_file_open(GtkWidget *widget, gpointer data)
{
	file_open(widget);
}

void menu_file_save(GtkWidget *widget, gpointer data)
{
	file_save_dialog(widget, FALSE);
}

void menu_file_save_as(GtkWidget *widget, gpointer data)
{
	file_save_dialog(widget, TRUE);
}

#ifdef HAVE_EXIV2

void menu_file_info_activate(GtkWidget *widget, gpointer data)
{
	dibuja_file_info_dialog();
}

#endif // HAVE_EXIV2

void menu_file_quit(GtkWidget *widget, gpointer data)
{
	do_quit();
}

/* Edit menu */
void menu_edit_undo(GtkWidget *widget, gpointer data)
{
	dibuja_do_undo();
}

void menu_edit_redo(GtkWidget *widget, gpointer data)
{
	dibuja_do_redo();
}

void menu_color_edit_activate(GtkWidget *widget, gpointer data)
{
	color_edit_dialog(widget);
}

void menu_edit_preferences_activate(GtkWidget *widget, gpointer data)
{
	dibuja_run_preferences_dialog();
}

void menu_edit_cut_activate(GtkWidget *widget, gpointer data)
{
	sel_cut();
}

void menu_edit_copy(GtkWidget *widget, gpointer data)
{
	GdkPixbuf *pixbuf;

	pixbuf = dibuja_copy_drawing();

	if(GDK_IS_PIXBUF(pixbuf)){
		GtkClipboard *clipboard;

		clipboard = gtk_clipboard_get_for_display(gdk_display_get_default(), GDK_SELECTION_CLIPBOARD);
		gtk_clipboard_set_image(clipboard, pixbuf);
		g_object_unref(pixbuf);
	}
}

void menu_edit_paste(GtkWidget *widget, gpointer data)
{
	GdkPixbuf *pixbuf;
	GtkClipboard *clipboard;

	clipboard = gtk_clipboard_get_for_display(gdk_display_get_default(),
											  GDK_SELECTION_CLIPBOARD);

	if(gtk_clipboard_wait_is_image_available(clipboard)){
		GtkWidget *main_wnd = GTK_WIDGET(gtk_widget_get_toplevel(widget));
		GdkCursor *cursor = gdk_window_get_cursor(main_wnd->window);

		canvas_set_wait_cursor(dibuja_get_canvas());
		canvas_set_wait_cursor(main_wnd);

		pixbuf = gtk_clipboard_wait_for_image(clipboard);
		tools_toggle_sel_button(TRUE);
		sel_create_from_pixbuf(pixbuf);
		g_object_unref(pixbuf);
		gtk_widget_queue_draw(dibuja_get_canvas());

		canvas_set_cursor(dibuja_get_canvas(), DIBUJA_CURSOR);
		gdk_window_set_cursor(main_wnd->window, cursor);
	}
}

void menu_edit_copy_desktop(GtkWidget *widget, gpointer data)
{
	g_timeout_add(500, dibuja_start_take_screenshot, NULL);
}

void menu_edit_paste_from_file(GtkWidget *widget, gpointer data)
{
	char *filename = file_open_dialog(widget);

	if(NULL != filename){
		GdkPixbuf *pixbuf = file_load_pixbuf(filename, TRUE);

		if(GDK_IS_PIXBUF(pixbuf)){
			tools_toggle_sel_button(TRUE);
			sel_create_from_pixbuf(pixbuf);
			g_object_unref(pixbuf);
		}
		g_free(filename);
	}
}


void menu_edit_activate(GtkWidget *widget, gpointer data)
{
	GtkWidget *menu_item;
	GtkClipboard *clipboard;
	gboolean ret;

	// TODO: These should be set by undo.c
	menu_item = get_widget("edit_undo");
	if(undo_can_undo(SIMPLE_CANVAS(dibuja_get_canvas()))){
		gtk_widget_set_sensitive(menu_item, TRUE);
	}
	else{
		gtk_widget_set_sensitive(menu_item, FALSE);
	}

	menu_item = get_widget("edit_redo");
	if(undo_can_redo(SIMPLE_CANVAS(dibuja_get_canvas()))){
		gtk_widget_set_sensitive(menu_item, TRUE);
	}
	else{
		gtk_widget_set_sensitive(menu_item, FALSE);
	}

	clipboard = gtk_clipboard_get_for_display(gdk_display_get_default(),
											  GDK_SELECTION_CLIPBOARD);

	ret = gtk_clipboard_wait_is_image_available(clipboard);
	menu_item = get_widget("edit_paste");
	gtk_widget_set_sensitive(menu_item, ret);

	ret = sel_have();
	menu_item = get_widget("edit_copy");
	gtk_menu_item_set_label(GTK_MENU_ITEM(menu_item), TRUE == ret ?
											_("_Copy Selection") : _("_Copy Image"));

	menu_item = get_widget("menu_edit_cut");
	gtk_widget_set_sensitive(menu_item, ret);
}

void menu_edit_clear(GtkWidget *widget, gpointer data)
{
	dibuja_clear_image();
}

/* Toolbar zoom ***************************************************************/
static int *combo_get_index(GtkWidget *combo)
{
	return (int *)(g_object_get_data(G_OBJECT(combo), "index"));
}

static void combo_set_index(GtkWidget *combo, int index)
{
	if(GTK_IS_WIDGET(combo)){
		int *i = combo_get_index(combo);
		if(i){
			*i = index;
		}
	}
}

static int get_n_zooms(void)
{
	int n = 0;
	GtkTreeModel *tree;

	tree = gtk_combo_box_get_model(GTK_COMBO_BOX(get_widget("combo")));

	n = gtk_tree_model_iter_n_children(tree, NULL);

	return n;
}

static char *combo_get_text_from_index(GtkComboBox *combo, int index)
{
	GtkTreeIter iter;
	GtkTreeModel *tree;
	char *text;

	tree = gtk_combo_box_get_model(combo);
	gtk_tree_model_iter_nth_child(tree, &iter, NULL, index);
	gtk_tree_model_get(tree, &iter, 0, &text, -1);
	return text;
}

int zoom_combo_default_index(GtkComboBox *combo)
{
	int i, z, max = get_n_zooms();

	for(i = 0; i < max; i++){
		char *str;

		str = combo_get_text_from_index(combo, i);
		z = g_ascii_strtod(str, NULL);
		g_free(str);

		if(100 == z){
			return i;
		}
	}

	return -1;
}

static void insert_zoom(char *text)
{
	int new_zoom = g_ascii_strtod(text, NULL);
	int i, max = get_n_zooms();
	int cur_zoom;
	GtkComboBox *combo = GTK_COMBO_BOX(get_widget("combo"));

	if(new_zoom < MIN_ZOOM_INT || new_zoom > MAX_ZOOM_INT){
		return;
	}

	for(i = 0; i < max; i++){
		char *str;

		str = combo_get_text_from_index(combo, i);
		cur_zoom = g_ascii_strtod(str, NULL);
		g_free(str);
		// Don't need copy of existing item...
		if(new_zoom == cur_zoom){
			// ...but set it active anyway
			gtk_combo_box_set_active(combo, i);
			return;
		}

		if(new_zoom < cur_zoom){
			char *new_str = g_strdup_printf("%d%%", new_zoom);

			gtk_combo_box_insert_text(combo, i, new_str);
			gtk_combo_box_set_active(combo, i);
			g_free(new_str);
			return;
		}
	}

}

void zoom_combo_changed(GtkComboBox *combo, gpointer data)
{
	char *p, *text = NULL;
	int zoom;
	int active;

	active = gtk_combo_box_get_active(combo);
	if(active < 0)return;

	text = gtk_combo_box_get_active_text(combo);
	combo_set_index(GTK_WIDGET(combo), active);

	if(text){
		if((p = strchr(text, ' ')) != NULL){
			*p = '\0';
		}
		zoom = (int)g_ascii_strtod(text, NULL);
		dibuja_set_zoom(zoom);
		g_free(text);
	}
}

void zoom_combo_entry_activate(GtkEntry *entry)
{

	char *p;
	char *text = g_strdup(gtk_entry_get_text(entry));
	int zoom;

	if(text){
		if((p = strchr(text, '%')) != NULL){
			*p = '\0';
		}

		zoom = (int)g_ascii_strtod(text, NULL);
		// Zoom is not valid, reset entry's text
		if(!(zoom >= MIN_ZOOM_INT && zoom <= MAX_ZOOM_INT)){
			GtkComboBox *combo = GTK_COMBO_BOX(get_widget("combo"));
			int index = *combo_get_index(GTK_WIDGET(combo));
			char *s;

			g_free(text);
			if(index > -1){
				s = combo_get_text_from_index(combo, index);
				gtk_entry_set_text(entry, s);
				/* GtkComboBoxEntry likes to set index to -1 for some reason,
				   we set it here
				 */
				gtk_combo_box_set_active(combo, index);
				g_free(s);
			}
			return;
		}

		insert_zoom(text);
		dibuja_set_zoom(zoom);

		g_free(text);
		text = g_strdup_printf("%d%%", zoom);
		gtk_entry_set_text(entry, text);

		g_free(text);
	}
}

void tb_zoom_in_clicked(GtkWidget *widget, gpointer data)
{
	GtkComboBox *combo = GTK_COMBO_BOX(get_widget( "combo"));
	int index = gtk_combo_box_get_active(combo);

	index++;
	if(!(index >= 0 && index < 15)){ return; }
	gtk_combo_box_set_active(combo, index);
}

void tb_zoom_out_clicked(GtkWidget *widget, gpointer data)
{
	GtkComboBox *combo = GTK_COMBO_BOX(get_widget("combo"));
	int index = gtk_combo_box_get_active(combo);
	index--;
	if(index < 0){ return; }
	gtk_combo_box_set_active(combo, index);
}

/* Toolbar *//*****************************************************/

void tb_undo_clicked(GtkWidget *widget, gpointer data)
{
	dibuja_do_undo();
}

void tb_redo_clicked(GtkWidget *widget, gpointer data)
{
	dibuja_do_redo();
}

/* Image Menu */
void menu_image_flip_horz(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		sel_flip(FLIP_HORZ);
	}
	else
		dibuja_flip(FLIP_HORZ);
}

void menu_image_flip_vert(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		sel_flip(FLIP_VERT);
	}
	else
		dibuja_flip(FLIP_VERT);
}

void menu_image_90_clockwise(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		sel_rotate(90.0);
	}
	else
		dibuja_rotate_image(90.);
}

void menu_image_90_counter_clockwise(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		sel_rotate(270.0);
	}
	else
		dibuja_rotate_image(270.);
}

void menu_image_rotate_180(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		sel_rotate(180.0);
	}
	else
		dibuja_rotate_image(180.);
}

void menu_image_rotate_custom(GtkWidget *widget, gpointer data)
{
	dialog_rotate_run();
}

void menu_image_resize_activate(GtkWidget *widget, gpointer data)
{
	dialog_resize_run();
}

void menu_image_scale_activate(GtkWidget *widget, gpointer data)
{
	dialog_scale_run();
}

void menu_image_crop(GtkWidget *widget, gpointer data)
{
	sel_crop_to();
}

void menu_image_selection_as_brush_activate(GtkWidget *widget, gpointer data)
{
	if(sel_have()){
		GdkPixbuf *pixbuf = sel_copy();
		if(GDK_IS_PIXBUF(pixbuf)){
			if(brush_set_brush_image_from_selection(pixbuf)){
				brush_image_brush_activate();
				tools_toggle_brush_button(TRUE);
			}
			g_object_unref(pixbuf);
		}
	}
}

void menu_image_activate(GtkWidget *widget)
{
	GtkWidget *menu_item;

	menu_item = get_widget("menu_image_crop");
	gtk_widget_set_sensitive(menu_item, sel_have());

	menu_item = get_widget("menu_image_selection_as_brush");
	gtk_widget_set_sensitive(menu_item, sel_have());
}

/* View Menu ******************************************************************/
void menu_view_normal_size_activate(GtkWidget *widget)
{
	GtkComboBox *combo = GTK_COMBO_BOX(get_widget("combo"));
	int i;

	i = zoom_combo_default_index(combo);
	gtk_combo_box_set_active(combo, i);
}

void menu_view_toolbar_toggled(GtkCheckMenuItem *toggle, gpointer data)
{
	gtk_widget_set_visible(get_widget("toolbar1"), gtk_check_menu_item_get_active(toggle));
}

void menu_view_tools_toggled(GtkCheckMenuItem *toggle, gpointer data)
{
	gtk_widget_set_visible(get_widget("table1"), gtk_check_menu_item_get_active(toggle));
}

void menu_view_colors_toggled(GtkCheckMenuItem *toggle, gpointer data)
{
	gtk_widget_set_visible(get_widget("table_col_pal"), gtk_check_menu_item_get_active(toggle));
}

void menu_view_statusbar(GtkCheckMenuItem *toggle, gpointer data)
{
	gtk_widget_set_visible(get_widget("statusbar1"), gtk_check_menu_item_get_active(toggle));
}

/* Effects Menu ***************************************************************/
void menu_effects_grayscale_activate(GtkWidget *widget, gpointer data)
{
	dibuja_do_grayscale();
}

void menu_effects_invert_activate(GtkWidget *widget, gpointer data)
{
	dibuja_do_invert();
}

void menu_effects_insta_brannan_activate(GtkWidget *widget, gpointer data)
{
	char *op = g_object_get_data(G_OBJECT(widget), "op");
	dibuja_do_insta_brannan(op);
}

void menu_effects_insta_1977_activate(GtkWidget *widget, gpointer data)
{
	char *op = g_object_get_data(G_OBJECT(widget), "op");
	dibuja_do_insta_1977(op);
}

void menu_effects_pixelate_activate(GtkWidget *widget, gpointer data)
{
	dibuja_do_pixelize();
}

void menu_effects_emboss_activate(GtkWidget *widget, gpointer data)
{
	dibuja_do_emboss();
}

static void menu_run_fx_dialog(char *op, char *undo_name)
{
#if GEGL_MINOR_VERSION >= 4
	char *title;

	title = (char *)gegl_operation_get_key(op, "title");
	if((title) && '\0' != title[0]){
		dibuja_run_fx_dilalog(op, title);
	}
	else
#endif
		dibuja_run_fx_dilalog(op, undo_name);
}

void menu_effects_brightness_contrast_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:brightness-contrast", _("Brightness - Contrast"));
}

void menu_effects_gaussian_blur_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:gaussian-blur", _("Gaussian Blur"));
}

void menu_effects_threshold_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:threshold", _("Threshold"));
}

void menu_effects_posterize_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:posterize", _("Posterize"));
}

void menu_effects_vignette_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:vignette", _("Vignette"));
}

void menu_effects_sharpen_activate(GtkWidget *menu, gpointer data)
{
	menu_run_fx_dialog("gegl:unsharp-mask", _("Sharpen"));
}

void menu_effects_activate(GtkWidget *widget, gpointer data)
{
	GtkWidget *menu_item;

	menu_item = get_widget("menu_effects_pixelate");
	gtk_widget_set_sensitive(menu_item, sel_have());

}

/******************************************************************************/

gboolean menu_default_popup(GtkMenu *menu, GdkEventButton  *event )
{
	/* Update the edit menu */
	menu_edit_activate(get_widget("editmenu"), NULL);
	/* Show the popup */
	gtk_menu_popup (menu, NULL, NULL, NULL, NULL,
					event->button, event->time);
	return TRUE;
}

/* Help menu see help.c */




