/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "tools.h"
#include "polyline.h"
#include "dibuja.h"
#include "util.h"
#include "draw-funcs.h"
#include "undo.h"
#include "file.h"


gboolean polyline_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
gboolean polyline_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data);
gboolean polyline_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);
gboolean polyline_expose(GtkWidget *widget, cairo_t *cr, gpointer data);
void polyline_finish(void);

static DrawTool tool = {
				polyline_button_press,
				polyline_motion,
				polyline_button_release,
				polyline_expose,
				polyline_finish,
				polyline_button_press
			};

static double sx = 0;
static double sy = 0;
static double ex = 0;
static double ey = 0;
static GList *lines = NULL;

static GdkRectangle old, cur;



DrawTool *polyline_get_tool(void)
{
	return &tool;
}

void set_bg_color(cairo_t *cr)
{
	SCColor bg;
	dibuja_get_bg_sccolor(&bg);
	cairo_set_source_rgba(cr, bg.r, bg.g, bg.b, bg.a);
}

void set_fg_color(cairo_t *cr)
{
	SCColor fg;
	dibuja_get_fg_sccolor(&fg);
	cairo_set_source_rgba(cr, fg.r, fg.g, fg.b, fg.a);
}

void polyline_calc_update_area(GdkRectangle *r)
{
	GList *tmp = lines;
	GdkPoint *pt;
	int maxx, maxy, minx, miny;
	double z = dibuja_get_zoom();
	int line_width = dibuja_get_line_width();

	if(!tmp || !r)
		return;

	pt = (GdkPoint *)tmp->data;
	if(!pt)
		return;

	maxx = minx = pt->x;
	maxy = miny = pt->y;
	while(tmp){
		pt = (GdkPoint *)tmp->data;
		if(pt){
			minx = MIN(minx, pt->x);
			miny = MIN(miny, pt->y);
			maxx = MAX(maxx, pt->x);
			maxy = MAX(maxy, pt->y);
		}
		tmp = tmp->next;
	}

	set_update_rect(r, minx, miny, maxx, maxy, z);
	resize_rect(r, (line_width * 2) * z, (line_width * 2) * z);
}

void polyline_finish(void)
{
	GList *tmp = lines;
	GdkPoint *pt;

	while(tmp){
		pt = tmp->data;
		if(pt){
			g_free(pt);
			pt = NULL;
		}
		tmp = tmp->next;
	}
	if(lines)
		g_list_free(lines);
	lines = NULL;

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void polyline_add(int x, int y)
{
	GdkPoint *pt = NULL;

	pt = g_malloc(sizeof(GdkPoint));
	if(pt){
		pt->x = x;
		pt->y = y;

		lines = g_list_prepend(lines, pt);
	}
}

void polyline_draw(cairo_t *cr)
{
	GList *tmp = lines;

	draw_line(cr, ex, ey, sx, sy);

	if(tmp){
		GdkPoint p1, p2;

		p1 = *((GdkPoint *)tmp->data);
		tmp = tmp->next;

		while(tmp){
			p2 = *((GdkPoint *)tmp->data);

			cairo_move_to(cr, (double)p1.x, (double)p1.y);
			cairo_line_to(cr, (double)p2.x, (double)p2.y);

			p1 = p2;
			tmp = tmp->next;
		}

		set_fg_color(cr);
		cairo_stroke(cr);
	}
}


int filled_polygon_draw(cairo_t *cr)
{
	GList *tmp = lines;
	double xa;
	double ya;
	double xb;
	double yb;
	ShapeFillType fill_mode = dibuja_get_fill_mode();

	if(g_list_length(lines) < 3){
		return 0;
	}

	if(tmp){
		GdkPoint p1;

		cairo_new_sub_path (cr);

		while(tmp){
			p1 = *((GdkPoint *)tmp->data);
			cairo_line_to(cr, (double)p1.x, (double)p1.y);
			tmp = tmp->next;
		}
		p1 = *((GdkPoint *)lines->data);
		cairo_line_to(cr, (double)p1.x, (double)p1.y);
		cairo_close_path(cr);
		cairo_stroke_extents(cr, &xa, &ya, &xb, &yb);
{
GdkRectangle area;
rect_from_pts(&area, (int)xa, (int)ya, (int)xb, (int)yb);
resize_rect(&area, dibuja_get_line_width(), dibuja_get_line_width());
undo_add(SIMPLE_CANVAS(dibuja_get_canvas()), area, _("Polygon"));
}
		if(SHAPE_FILL_NONE != fill_mode){
			if(SHAPE_FILL_BACK == fill_mode){
				set_bg_color(cr);
			}
			else{
				set_fg_color(cr);
			}
			cairo_set_fill_rule (cr, CAIRO_FILL_RULE_EVEN_ODD);
			cairo_fill_preserve(cr);
		}
		set_fg_color(cr);
		cairo_stroke(cr);

		return 1;
	}
	return 0;
}

gboolean polyline_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(!lines){
		sx = ex = event->x;
		sy = ey = event->y;

		polyline_add((int)sx, (int)sy);
		polyline_add((int)ex, (int)ey);
	}

	set_update_rect(&old, sx, sy, ex, ey, simple_canvas_get_zoom(SIMPLE_CANVAS(widget)));
	cur = old;

	return TRUE;
}

gboolean polyline_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	GdkRectangle area = {0};

	if (event->state & GDK_BUTTON1_MASK){
		if(lines){
			ex = (double)event->x;
			ey = (double)event->y;

			if(event->state & GDK_CONTROL_MASK){
				draw_constrain_pt(sx, sy, ex, ey, &ex, &ey);
			}

			((GdkPoint *)lines->data)->x = ex;
			((GdkPoint *)lines->data)->y = ey;

			polyline_calc_update_area(&cur);
			gdk_rectangle_union(&old, &cur, &area);
			old = cur;

			gdk_window_invalidate_rect(widget->window, &area, 0);
		}
	}
	return TRUE;
}

gboolean polyline_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(1 == event->button){
		GdkRectangle area = {0};

		ex = event->x;
		ey = event->y;

		if(event->state & GDK_CONTROL_MASK){
			draw_constrain_pt(sx, sy, ex, ey, &ex, &ey);
		}

		if(ex == sx && sy == ey) return TRUE;

		polyline_add((int)ex, (int)ey);

		polyline_calc_update_area(&area);

		sx = ex;
		sy = ey;
		gdk_window_invalidate_rect(widget->window, &area, 0);

		//old = area;

	}
	else if(3 == event->button){
		cairo_t *cr;
		GList *tmp = NULL;

		tmp = g_list_last(lines);
		polyline_add(((GdkPoint *)tmp->data)->x,
					 ((GdkPoint *)tmp->data)->y
					 );

		cr = dibuja_get_cr(CR_SURFACE);

		if(!filled_polygon_draw(cr)){
			polyline_draw(cr);
		}

		cairo_destroy(cr);

		polyline_finish();
		gtk_widget_queue_draw(widget);
	}

	return TRUE;
}

gboolean polyline_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	polyline_draw(cr);

	return TRUE;
}


