/*
 * Copyright (C) Juan Balderas 2015 <juan.balderas.0@gmail.com>
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sel-obj.h"



struct _SelObjPrivate{
	GdkRectangle rect;
	cairo_surface_t *surf;
	int xoff, yoff;
	GdkRectangle orig_pos;
	gboolean modified;
	gchar *description;
};

#define SEL_OBJ_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), SEL_TYPE_OBJ, SelObjPrivate))



G_DEFINE_TYPE (SelObj, sel_obj, G_TYPE_OBJECT);

static void
sel_obj_init (SelObj *object)
{
	GdkRectangle tmp = {0};
	SelObj *sel;

	object->priv = G_TYPE_INSTANCE_GET_PRIVATE (object,
												SEL_TYPE_OBJ,
												SelObjPrivate);
	sel = SEL_OBJ(object);

	sel->priv->rect = tmp;
	sel->priv->surf = NULL;
	sel->priv->xoff = 0;
	sel->priv->yoff = 0;
	sel->priv->orig_pos = tmp;
	sel->priv->modified = FALSE;
	sel->priv->description = NULL;
}

static void
sel_obj_finalize (GObject *object)
{
	SelObj *sel = SEL_OBJ(object);


	if(sel->priv->surf){
		cairo_surface_destroy(sel->priv->surf);
	}
	if(sel->priv->description){
		g_free(sel->priv->description);
		sel->priv->description = NULL;
	}

	G_OBJECT_CLASS (sel_obj_parent_class)->finalize (object);
}

static void
sel_obj_class_init (SelObjClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);

	g_type_class_add_private (klass, sizeof (SelObjPrivate));

	object_class->finalize = sel_obj_finalize;
}

SelObj *
sel_obj_new (void)
{

	return SEL_OBJ(g_object_new (SEL_TYPE_OBJ, NULL));
}

void sel_obj_set_image(SelObj *sel, cairo_surface_t *new_img)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	g_return_if_fail(new_img);

	if(sel->priv->surf){
		cairo_surface_destroy(sel->priv->surf);
	}
	sel->priv->surf = new_img;
	sel->priv->rect.width = cairo_image_surface_get_width(new_img);
	sel->priv->rect.height = cairo_image_surface_get_height(new_img);
}

cairo_surface_t *sel_obj_swap_image(SelObj *sel, cairo_surface_t *new_img)
{
	cairo_surface_t *old_img = NULL;

	g_return_val_if_fail(SEL_IS_OBJ(sel), NULL);
	g_return_val_if_fail(new_img, NULL);

	old_img = sel->priv->surf;

	sel->priv->surf = new_img;
	sel->priv->rect.width = cairo_image_surface_get_width(new_img);
	sel->priv->rect.height = cairo_image_surface_get_height(new_img);

	return old_img;
}
/*
int sel_obj_create_from_rect(SelObj *sel, GdkRectangle *rect)
{
	g_return_val_if_fail(SEL_IS_OBJ(sel), 0);
	g_return_val_if_fail(rect, 0);
	g_return_val_if_fail((rect->width > 0 && rect->height > 0), 0);

	...

	return 1;
}
*/
SelObj *sel_obj_create_from_surface(cairo_surface_t *source)
{
	SelObj *sel = NULL;

	g_return_val_if_fail(source != NULL, NULL);

	sel = sel_obj_new();
	g_return_val_if_fail(SEL_IS_OBJ(sel), NULL);

	sel->priv->rect.width = cairo_image_surface_get_width(source);
	sel->priv->rect.height = cairo_image_surface_get_height(source);
	sel->priv->surf = source;

	return sel;
}

void sel_obj_move(SelObj *sel, int x, int y, int use_offset)
{
	g_return_if_fail(SEL_IS_OBJ(sel));

	if(use_offset){
		sel->priv->rect.x = x - sel->priv->xoff;
		sel->priv->rect.y = y - sel->priv->yoff;
	}
	else{
		sel->priv->rect.x = x;
		sel->priv->rect.y = y;
	}
}

void sel_obj_resize(SelObj *sel, int width, int height)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	if(!(width > 0 && height > 0)) return;

	sel->priv->rect.width = width;
	sel->priv->rect.height = height;
}

cairo_surface_t *sel_obj_get_image(SelObj *sel)
{
	g_return_val_if_fail(SEL_IS_OBJ(sel), NULL);

	return sel->priv->surf;
}

void sel_obj_get_rectangle(SelObj *sel, GdkRectangle *rect)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	g_return_if_fail(rect);

	*rect = sel->priv->rect;
}

int sel_obj_point_in(SelObj *sel, int x, int y)
{
	g_return_val_if_fail(SEL_IS_OBJ(sel), 0);

	if(x >= sel->priv->rect.x && x < sel->priv->rect.x + sel->priv->rect.width){
		if(y >= sel->priv->rect.y && y < sel->priv->rect.y + sel->priv->rect.height){
			return 1;
		}
	}

	return 0;
}

void sel_obj_set_offset(SelObj *sel, int x, int y)
{
	g_return_if_fail(SEL_IS_OBJ(sel));

	sel->priv->xoff = x - sel->priv->rect.x;
	sel->priv->yoff = y - sel->priv->rect.y;
}

void sel_obj_set_orig_pos(SelObj *sel, GdkRectangle *rect)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	g_return_if_fail(rect);

	sel->priv->orig_pos = *rect;
}

void sel_obj_get_orig_pos(SelObj *sel, GdkRectangle *rect)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	g_return_if_fail(rect);

	*rect = sel->priv->orig_pos;
}

cairo_surface_t *sel_obj_remove_image(SelObj *sel)
{
	cairo_surface_t *surf;

	g_return_val_if_fail(SEL_IS_OBJ(sel), NULL);

	surf = sel->priv->surf;
	sel->priv->surf = NULL;

	return surf;
}

gboolean sel_obj_get_modified(SelObj *sel)
{
	g_return_val_if_fail(SEL_IS_OBJ(sel), FALSE);

	return sel->priv->modified;
}

void sel_obj_set_modified(SelObj *sel, gboolean modified)
{
	g_return_if_fail(SEL_IS_OBJ(sel));

	sel->priv->modified = modified;
}

char *sel_obj_get_description(SelObj *sel)
{
	g_return_val_if_fail(SEL_IS_OBJ(sel), NULL);

	return sel->priv->description;
}

void sel_obj_set_description(SelObj *sel, char *desc)
{
	g_return_if_fail(SEL_IS_OBJ(sel));
	g_return_if_fail(desc != NULL);

	if(sel->priv->description){
		g_free(sel->priv->description);
	}
	sel->priv->description = g_strdup(desc);
}

