/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include "selection.h"
#include "sel-obj.h"
#include "draw-funcs.h"
#include "dibuja.h"
#include "util.h"
#include "gdkpixbuf-drawable.h"
#include "undo.h"

static gboolean sel_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean sel_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data);
static gboolean sel_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);
static gboolean sel_expose(GtkWidget *widget, cairo_t *cr, gpointer data);
static void sel_finish(void);
static gboolean sel_cancel(void);
static gboolean sel_set_image(SelObj *so, cairo_surface_t *surf);

static void
set_alpha_pixel(GdkPixbuf *pixbuf, guchar r, guchar g, guchar b, guchar a);
static void sel_make_trans(SelObj *so, gboolean trans);

extern void menu_edit_copy(GtkWidget *widget, gpointer data);

static DrawTool draw_tool = {
						sel_button_press,
						sel_motion,
						sel_button_release,
						sel_expose,
						sel_finish,
						NULL,
						sel_cancel
					};

static gboolean move_sel = FALSE;
static SelObj *sel = NULL;
static int startx, starty, endx, endy;

DrawTool *sel_get_tool(void)
{
	return &draw_tool;
}

void sel_destroy_pixbuf(GdkPixbuf *pixbuf)
{
	if(GDK_IS_PIXBUF(pixbuf)){
		g_object_unref(pixbuf);
	}
}

gboolean sel_have(void)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!SEL_IS_OBJ(sel)) return FALSE;

	tmpsurf = sel_obj_get_image(sel);

	return (((tmpsurf != NULL) ? TRUE : FALSE));
}

GdkPixbuf *sel_copy(void)
{
	cairo_surface_t *tmpsurf = NULL;
	GdkPixbuf *pixbuf = NULL;

	if(!sel_have()) return NULL;

	pixbuf = GDK_PIXBUF(g_object_get_data(G_OBJECT(sel), "pixbuf"));
	if(GDK_IS_PIXBUF(pixbuf)){
		return gdk_pixbuf_copy(pixbuf);
	}
	tmpsurf = sel_obj_get_image(sel);

	return gdk_pixbuf_get_from_surface (tmpsurf, 0, 0,
									cairo_image_surface_get_width(tmpsurf),
									cairo_image_surface_get_height(tmpsurf));
}

static void sel_get_crtl_rect(GdkRectangle *r,
							  double sx, double sy,
							  double ex, double ey)
{
	double len = sqrt(((sx - ex) * (sx - ex)) + ((sy - ey) * (sy - ey)));

	rect_from_pts(r, (int)roundf(sx - len), (int)roundf(sy - len),
				  (int)roundf(sx + len), (int)roundf(sy + len) );
}

static void sel_update_tb_and_menu(gboolean sens)
{
	gtk_widget_set_sensitive(get_widget("menu_edit_cut"), sens);
	gtk_widget_set_sensitive(get_widget("tb_cut"), sens);
	gtk_widget_set_sensitive(get_widget("tb_apply_selection"), sens);
	gtk_widget_set_sensitive(get_widget("tb_clear_selection"), sens);
}

static void sel_update_tb_tools(gboolean sens)
{
	gtk_widget_set_sensitive(get_widget("tb_apply_selection"), sens);
	gtk_widget_set_sensitive(get_widget("tb_clear_selection"), sens);
}

static void sel_finish(void)
{
	cairo_t *cr;
	cairo_surface_t *tmpsurf = NULL;
	GdkRectangle curpos = {0};

	if(SEL_IS_OBJ(sel)){
		if((tmpsurf = sel_obj_get_image(sel)) != NULL){
			if(sel_obj_get_modified(sel)){
				char *desc = NULL;
				sel_obj_get_rectangle(sel, &curpos);
				desc = sel_obj_get_description(sel);
				undo_add(SIMPLE_CANVAS(dibuja_get_canvas()),
							curpos, desc != NULL ? desc : _("Selection"));

				cr = dibuja_get_cr(CR_SURFACE);
				cairo_set_source_surface(cr, tmpsurf, curpos.x, curpos.y);
				cairo_paint(cr);
				cairo_destroy(cr);
			}
		}
	}

	move_sel = FALSE;
	if(SEL_IS_OBJ(sel)){
		g_object_unref(sel);
		sel = NULL;
	}
	startx = endx = starty = endy = 0;
	gtk_widget_queue_draw(dibuja_get_canvas());
	sel_update_tb_and_menu(FALSE);
	sel_update_tb_tools(FALSE);
	// Use to update image size
	canvas_image_changed(SIMPLE_CANVAS(dibuja_get_canvas()), NULL);
}

static gboolean sel_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	int x, y;
	GdkRectangle tmp;

	x = (int)event->x;
	y = (int)event->y;

	if((SEL_IS_OBJ(sel)) && sel_obj_point_in(sel, x, y)){
		sel_obj_set_offset(sel, x, y);
		move_sel = TRUE;

		return TRUE;
	}

	sel_finish();

	sel = sel_obj_new();

	startx = endx = x;
	starty = endy = y;
	move_sel = FALSE;
	sel_obj_move(sel, (int)event->x, (int)event->y, FALSE);
	sel_obj_get_rectangle(sel, &tmp);
	sel_obj_set_orig_pos(sel, &tmp);


	return TRUE;
}

static gboolean sel_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	int line_width = 1;
	GdkRectangle cur, old = {0};

	if(event->state & GDK_BUTTON1_MASK){
		endx = (int)event->x;
		endy = (int)event->y;

		if(SEL_IS_OBJ(sel) && move_sel){

			sel_obj_get_rectangle(sel, &old);

			sel_obj_move(sel, (int)event->x, (int)event->y, TRUE);

			sel_obj_get_rectangle(sel, &cur);
			
			gdk_rectangle_union(&old, &cur, &cur);

			resize_rect(&cur, line_width * 2, line_width * 2);
			simple_canvas_update_rect(SIMPLE_CANVAS(widget), &cur);
		}
		else{
			// Draw just the frame
			if(SEL_IS_OBJ(sel)){
				sel_obj_get_rectangle(sel, &old);

				if(event->state & GDK_CONTROL_MASK){
					sel_get_crtl_rect(&cur, startx, starty, endx, endy);
/*					double len = sqrt(((startx - endx) * (startx - endx)) +
									  ((starty - endy) * (starty - endy)));

					printf("len: %f                           \r", len);
					fflush(stdout);
					rect_from_pts(&cur, (int)roundf(startx - len), (int)roundf(starty - len),
						  (int)roundf(startx + len), (int)roundf(starty + len) );
*/
				}
				else
					rect_from_pts(&cur, startx, starty, endx, endy);

				sel_obj_move(sel, cur.x, cur.y, FALSE);
				sel_obj_resize(sel, cur.width, cur.height);

				dibuja_show_sel_dims((double)cur.width, (double)cur.height);

				gdk_rectangle_union(&old, &cur, &cur);
				resize_rect(&cur, line_width * 2, line_width * 2);
				simple_canvas_update_rect(SIMPLE_CANVAS(widget), &cur);
			}
		}
	}

	return TRUE;
}

static gboolean sel_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(SEL_IS_OBJ(sel) && move_sel){
		GdkRectangle op = {0};
		sel_obj_move(sel, (int)event->x, (int)event->y, TRUE);
		sel_obj_get_orig_pos(sel, &op);
		if(op.x != (int)event->x && op.y != (int)event->y){
			sel_obj_set_modified(sel, TRUE);
		}
	}
	else{
		if(SEL_IS_OBJ(sel)){
			// Create the image
			GdkRectangle tmp = {0};

			endx = (int)event->x;
			endy = (int)event->y;

			if(event->state & GDK_CONTROL_MASK){
				sel_get_crtl_rect(&tmp, startx, starty, endx, endy);
/*				double len = sqrt(((startx - endx) * (startx - endx)) +
								  ((starty - endy) * (starty - endy)));

				rect_from_pts(&tmp, (int)(startx - len), (int)(starty - len),
							  (int)(startx + len), (int)(starty + len) );
*/
			}
			else
				rect_from_pts(&tmp, startx, starty, endx, endy);

			sel_obj_move(sel, tmp.x, tmp.y, FALSE);
			sel_obj_resize(sel, tmp.width, tmp.height);
			sel_obj_get_rectangle(sel, &tmp);

			if((tmp.width - 1 > 0) && (tmp.height - 1 > 0)){// -1 offset for sel line width
				cairo_surface_t *tmpsurf = NULL;

				tmpsurf = draw_copy_area(dibuja_get_surface(),
							  tmp.x, tmp.y,
							  tmp.width ,//- 1,
							  tmp.height );//- 1);

				undo_add_temp(SIMPLE_CANVAS(widget), _("Selection"));
				sel_set_image(sel, tmpsurf);
				sel_obj_set_orig_pos(sel, &tmp);
				sel_update_tb_and_menu(TRUE);
				sel_update_tb_tools(TRUE);

				sel_obj_get_rectangle(sel, &tmp);
				dibuja_show_sel_dims((double)tmp.width, (double)tmp.height);
			}
		}
	}
	
	move_sel = FALSE;

	return TRUE;
}

static void sel_draw_sel_frame(cairo_t *cr, GdkRectangle r)
{
	double dashes[2] = {2., 2.};
	int num_dashes = (int)(sizeof(dashes) / sizeof(dashes[0]));

	if(r.width < 1 || r.height < 1) return;

	cairo_set_line_width(cr, 1.0);
	cairo_set_source_rgba(cr, 1., 1., 1., 1.);
	cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);

	gdk_cairo_rectangle(cr, &r);
	cairo_stroke(cr);

	cairo_set_source_rgba(cr, 0., 0., 0., 1.);
	cairo_set_dash(cr, dashes, num_dashes, 0.);

	gdk_cairo_rectangle(cr,  &r);
	cairo_stroke(cr);
}

static void sel_draw(cairo_t *cr)
{
	GdkRectangle tmp;
	cairo_surface_t *tmpsurf = NULL;
	
	if(SEL_IS_OBJ(sel)){
		sel_obj_get_rectangle(sel, &tmp);

		if((tmpsurf = sel_obj_get_image(sel)) != NULL){
			cairo_set_source_surface(cr, tmpsurf,
									 (double)tmp.x,
									 (double)tmp.y);
			cairo_paint(cr);
		}

		sel_draw_sel_frame(cr, tmp);
	}
}

static gboolean sel_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	sel_draw(cr);
	return TRUE;
}

// For the clipboard
void sel_create_from_pixbuf(GdkPixbuf *pixbuf)
{
	int w, h;
	cairo_surface_t *tmp = NULL;
	GdkPixbuf *copy;
	SelObj *tmp_sel = NULL;

	g_return_if_fail(GDK_IS_PIXBUF(pixbuf));

	w = gdk_pixbuf_get_width(pixbuf);
	h = gdk_pixbuf_get_height(pixbuf);

	tmp = cairo_image_surface_create(DEF_SURF_FORMAT, w, h);
	if(tmp){
		int imgw, imgh;
		cairo_t *cr;

		copy = gdk_pixbuf_copy(pixbuf);
		g_return_if_fail(GDK_IS_PIXBUF(copy));
		if(!gdk_pixbuf_get_has_alpha(copy)){
			GdkPixbuf *pb = gdk_pixbuf_add_alpha(copy, FALSE, 0, 0, 0);
			if(!GDK_IS_PIXBUF(pb)){
				cairo_surface_destroy(tmp);
				message_box_error(_("Could not paste image from clipboard"));
				return;
			}
		}

		tmp_sel = sel_obj_new();
		if(!SEL_IS_OBJ(tmp_sel)){
			cairo_surface_destroy(tmp);
			g_object_unref(copy);
			copy = NULL;
			tmp = NULL;
			message_box_error(_("Could not paste image from clipboard"));
		}
		else{
			int x = 0;
			int y = 0;

			// Finish any drawing
			sel_finish();

			sel = tmp_sel;

			cr = cairo_create(tmp);
			gdk_cairo_set_source_pixbuf(cr, copy, 0., 0.);
			cairo_paint(cr);
			cairo_destroy(cr);

			imgw = cairo_image_surface_get_width(dibuja_get_surface());
			imgh = cairo_image_surface_get_height(dibuja_get_surface());
			if(w > imgw || h > imgh){
				int ret = message_box_yes_no(
						_("The size of the image in the clipboard is larger\n"
						  "than the size of the drawing area.\n"
						  "Resize the drawing area to fit the new image?"));
				if(GTK_RESPONSE_YES == ret){
					dibuja_resize_image(MAX(w, imgw), MAX(h, imgh), TRUE);
				}
			}

			sel_obj_set_image(sel, tmp);
			g_object_set_data_full(G_OBJECT(sel), "pixbuf", (gpointer)copy, 
									(GDestroyNotify)sel_destroy_pixbuf);

			sel_obj_move(sel, 0, 0, FALSE);
			dibuja_get_visible_pos(GTK_SCROLLED_WINDOW(get_widget("scroll")),
									&x, &y);
			x = (int)win_to_img((double)x, dibuja_get_zoom());
			y = (int)win_to_img((double)y, dibuja_get_zoom());
			undo_add_temp(SIMPLE_CANVAS(dibuja_get_canvas()), _("Selection"));
			sel_obj_move(sel, x, y, FALSE);
			move_sel = TRUE;
			sel_obj_set_modified(sel, TRUE);
			sel_update_tb_tools(TRUE);

			sel_make_trans(sel, dibuja_get_use_trans_sel());
		}

	}
}

static void sel_adjust(SelObj *s, cairo_surface_t *tmpsurf)
{
	GdkRectangle old, new;
	int cx, cy, ncx, ncy;

	if(!s) return;

	sel_obj_get_rectangle(s, &old);

	sel_set_image(s, tmpsurf);
	sel_obj_get_rectangle(s, &new);

	cx = old.x + old.width / 2;
	cy = old.y + old.height / 2;

	ncx = cx - new.width / 2;
	ncy = cy - new.height / 2;

	sel_obj_move(s, ncx, ncy, 0);
	sel_obj_get_rectangle(s, &new);
	sel_obj_set_orig_pos(s, &new);

	gtk_widget_queue_draw(dibuja_get_canvas());
}

void sel_flip(int dir)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!sel_have()) return;
	if(FLIP_HORZ != dir && FLIP_VERT != dir) return;

	tmpsurf = draw_surface_flip(sel_obj_get_image(sel), dir);
	if(tmpsurf){
		sel_adjust(sel, tmpsurf);
		sel_obj_set_modified(sel, TRUE);
	}
}

void sel_rotate(double a)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!sel_have()) return;

	tmpsurf = draw_rotate_image(sel_get_cur_image(), a);
	if(tmpsurf){
		sel_adjust(sel, tmpsurf);
		sel_obj_set_modified(sel, TRUE);
	}
}

void sel_scale(int neww, int newh)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!sel_have()) return;
	if(!(neww > 0 && newh > 0)) return;

	tmpsurf = draw_scale_image(sel_obj_get_image(sel), neww, newh);
	if(tmpsurf){
		sel_adjust(sel, tmpsurf);
		sel_obj_set_modified(sel, TRUE);
	}
}

void sel_resize_cur(int neww, int newh, gboolean centered)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!sel_have()) return;
	if(!(neww > 0 && newh > 0)) return;

	tmpsurf = draw_resize_image(sel_obj_get_image(sel), neww, newh, centered);
	if(tmpsurf){
		sel_adjust(sel, tmpsurf);
		sel_obj_set_modified(sel, TRUE);
	}
}

int sel_apply_effect(cairo_surface_t *surf)
{
	if(!SEL_IS_OBJ(sel)) return 0;
	if(!surf) return 0;

	sel_set_image(sel, surf);
	sel_obj_set_modified(sel, TRUE);
	gtk_widget_queue_draw(dibuja_get_canvas());

	return 1;
}

void sel_cut(void)
{
	cairo_surface_t *tmpsurf = NULL;
	cairo_t *cr;
	SCColor c;

	if(!sel_have()) return;

	menu_edit_copy(dibuja_get_canvas(), NULL);

	tmpsurf = sel_obj_get_image(sel);
	cr = cairo_create(tmpsurf);
	dibuja_get_bg_sccolor(&c);
	cairo_set_source_rgba(cr, c.r, c.g, c.b, 1.);
	cairo_paint(cr);
	cairo_destroy(cr);

	sel_obj_set_description(sel, _("Cut"));
	sel_obj_set_modified(sel, TRUE);
	sel_finish();
}

int sel_crop_to(void)
{
	cairo_surface_t *tmpsurf = NULL;

	if(!sel_have()) return 0;

	/* canvas assumes image */
	tmpsurf = sel_obj_remove_image(sel);
	dibuja_resize(tmpsurf, _("Crop"));

	sel_finish();

	return 1;
}

static gboolean sel_cancel(void)
{
	if(sel_have()){
		move_sel = FALSE;

		sel_obj_set_modified(sel, FALSE);
		g_object_unref(sel);
		sel = NULL;

		startx = endx = starty = endy = 0;
		sel_finish();
		return TRUE;
	}
	return FALSE;
}

void tb_clear_selection_clicked(GtkWidget *widget)
{
	sel_cancel();
	sel_update_tb_tools(FALSE);
	sel_update_tb_and_menu(FALSE);
	gtk_widget_queue_draw(dibuja_get_canvas());
}

void tb_apply_selection_clicked(GtkWidget *widget)
{
	sel_obj_set_modified(sel, TRUE);
	sel_finish();
}

// FIXME: Just use pixbufs for selection
cairo_surface_t *sel_get_cur_image(void)
{
	if(!sel_have()) return NULL;

	// Make not transparent to keep current alpha from being transformed
	sel_make_trans(sel, FALSE);

	return sel_obj_get_image(sel);
}

static void sel_make_trans(SelObj *so, gboolean trans)
{
	cairo_t *cr;
	cairo_surface_t *surf;
	GdkPixbuf *pixbuf = NULL;
	GdkColor color;

	if(!SEL_IS_OBJ(so)) return;

	pixbuf = GDK_PIXBUF(g_object_get_data(G_OBJECT(so), "pixbuf"));
	if(!GDK_IS_PIXBUF(pixbuf)) return;

	surf = sel_obj_get_image(so);
	if(!surf) return;

	dibuja_get_bg_color(&color);
	set_alpha_pixel(pixbuf, (guchar)(color.red / 256),
							(guchar)(color.green / 256),
							(guchar)(color.blue / 256),
							TRUE == trans ? 0 : 0xFF);
	cr = cairo_create(surf);
	cairo_set_source_rgba(cr, 0., 0., 0., 0.);
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_paint(cr);
	gdk_cairo_set_source_pixbuf(cr, pixbuf, 0., 0.);
	cairo_paint(cr);
	gtk_widget_queue_draw(dibuja_get_canvas());
	cairo_destroy(cr);
}

void check_button_transparent_selection_toggled(GtkToggleButton *toggle)
{
	gboolean trans = gtk_toggle_button_get_active(toggle);

	sel_make_trans(sel, trans);
}

static void
set_alpha_pixel(GdkPixbuf *pixbuf, guchar r, guchar g, guchar b, guchar a)
{
	int width, height, rowstride, n_channels = 4;
	guchar *pixels, *p;
	int x, y;

	width = gdk_pixbuf_get_width (pixbuf);
	height = gdk_pixbuf_get_height (pixbuf);

	rowstride = gdk_pixbuf_get_rowstride (pixbuf);
	pixels = gdk_pixbuf_get_pixels (pixbuf);

	for(y = 0; y < height; y++){
		for(x = 0; x < width; x++){
			p = pixels + y * rowstride + x * n_channels;
			if((p[0] == r) && (p[1] == g) && (p[2] == b)){
				p[3] = a;
			}
		}
	}
}

static gboolean sel_set_image(SelObj *so, cairo_surface_t *surf)
{
	GdkPixbuf *pixbuf = NULL;
	GdkPixbuf *old = NULL;
	int w, h, trans;

	if(!SEL_IS_OBJ(so)) return FALSE;
	if(!surf) return FALSE;

	w = cairo_image_surface_get_width(surf);
	h = cairo_image_surface_get_height(surf);
	pixbuf = gdk_pixbuf_get_from_surface(surf, 0, 0, w, h);

	if(!GDK_IS_PIXBUF(pixbuf)) return FALSE;

	old = GDK_PIXBUF(g_object_get_data(G_OBJECT(so), "pixbuf"));
	if(GDK_IS_PIXBUF(old))
		g_object_unref(old);

	sel_obj_set_image(so, surf);
	g_object_set_data_full(G_OBJECT(so), "pixbuf",
						   (gpointer)pixbuf,
						   (GDestroyNotify)sel_destroy_pixbuf);

	trans = dibuja_get_use_trans_sel();
	if(trans){
		sel_make_trans(so, trans);
	}
	return TRUE;
}











