/* 
 * (c) 2015 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib/gi18n.h>
#include <math.h>
#include "shapes.h"
#include "util.h"
#include "draw-funcs.h"
#include "dibuja.h"
#include "undo.h"
#include "file.h"



gboolean shapes_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);
gboolean shapes_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data);
gboolean shapes_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data);
gboolean shapes_expose(GtkWidget *widget, cairo_t *cr, gpointer data);

static DrawTool tool = {
				shapes_button_press,
				shapes_motion,
				shapes_button_release,
				shapes_expose,
				NULL,
				NULL
			};

static double sx = 0;
static double sy = 0;
static double ex = 0;
static double ey = 0;

static GdkRectangle shaperect = {0};

static ShapeType shape_type = SHAPE_RECTANGLE;
static ShapeFillType fill_type = SHAPE_FILL_NONE;
static GdkRectangle old, cur;

DrawTool *shapes_get_tool(void)
{
	return &tool;
}

void draw_shape(cairo_t *cr, double x1, double y1, double x2, double y2)
{
	SCColor fg, fill;

	dibuja_get_fg_sccolor(&fg);
	if(SHAPE_FILL_FORE == fill_type)
		dibuja_get_fg_sccolor(&fill);
	else if(SHAPE_FILL_BACK == fill_type)
		dibuja_get_bg_sccolor(&fill);

	switch(shape_type){
		case SHAPE_RECTANGLE:
			draw_rectangle(cr, shaperect, &fg,
						   (SHAPE_FILL_NONE == fill_type) ? NULL : &fill);
			break;
		case SHAPE_ELLIPSE:
			draw_ellipse(cr, shaperect, &fg,
						   (SHAPE_FILL_NONE == fill_type) ? NULL : &fill);
			break;
		case SHAPE_ROUNDED_RECTANGLE:
			draw_rounded_rectangle(cr, shaperect, &fg,
						   (SHAPE_FILL_NONE == fill_type) ? NULL : &fill);
			break;
		default:
			break;
	}
}

void shape_set_type(ShapeType type)
{
	if(type >= SHAPE_RECTANGLE && type < SHAPE_LAST){
		shape_type = type;
	}
}

void shape_set_fill_type(ShapeFillType type)
{
	if(type >= SHAPE_FILL_NONE && type < SHAPE_FILL_LAST){
		dibuja_set_fill_mode(type);
		fill_type = type;
	}
}

gboolean shapes_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	sx = ex = event->x;
	sy = ey = event->y;

	rect_from_pts(&shaperect, sx, sy, ex, ey);
	old = cur = shaperect;

	return TRUE;
}

gboolean shapes_motion(GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
	int line_width = dibuja_get_line_width();//2;
	GdkRectangle area = {0};

	if (event->state & GDK_BUTTON1_MASK){
		ex = event->x;
		ey = event->y;

		if(event->state & GDK_CONTROL_MASK){
			double len = sqrt(((sx - ex) * (sx - ex)) + ((sy - ey) * (sy - ey)));
			rect_from_pts(&shaperect, (int)(sx - len), (int)(sy - len),
						  (int)(sx + len), (int)(sy + len) );
		}
		else
			rect_from_pts(&shaperect, (int)sx, (int)sy, (int)ex, (int)ey);

		resize_rect(&shaperect, -(line_width / 2), -(line_width / 2));

		cur = shaperect;
		resize_rect(&cur, (int)(line_width * 2.), (int)(line_width * 2.));
		gdk_rectangle_union(&old, &cur, &area);
		old = cur;
		
		simple_canvas_update_rect(SIMPLE_CANVAS(widget), &area);
	}

	return TRUE;
}

gboolean shapes_button_release(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	if(1 == event->button){
		int line_width = dibuja_get_line_width();//2;
		GdkRectangle area = {0};
		cairo_t *cr;
		char *sh[SHAPE_LAST] = {
								N_("Rectangle"), N_("Ellipse"), N_("Rounded Rectangle")
							};

		ex = event->x;
		ey = event->y;

		if(event->state & GDK_CONTROL_MASK){
			double len = sqrt(((sx - ex) * (sx - ex)) + ((sy - ey) * (sy - ey)));
			rect_from_pts(&shaperect, (int)(sx - len), (int)(sy - len),
						  (int)(sx + len), (int)(sy + len) );
		}
		else
			rect_from_pts(&shaperect, (int)sx, (int)sy, (int)ex, (int)ey);

		resize_rect(&shaperect, -(line_width / 2), -(line_width / 2));
		area = shaperect;

		if(!(area.width > 0 && area.height > 0))
			return TRUE;

		resize_rect(&area, line_width * 2, line_width * 2);

		undo_add(SIMPLE_CANVAS(widget), area, _(sh[shape_type]));

		cr = dibuja_get_cr(CR_SURFACE);
		draw_shape(cr, sx, sy, ex, ey);
		cairo_destroy(cr);

		set_update_rect(&cur, sx, sy, ex, ey, simple_canvas_get_zoom(SIMPLE_CANVAS(widget)));
		gdk_rectangle_union(&old, &cur, &area);
		resize_rect(&area, line_width * 2, line_width * 2);

		sx = ex = sy = ey = 0;
		rect_from_pts(&shaperect, 0, 0, 0, 0);
		gdk_window_invalidate_rect(widget->window, &area, 0);
		old = area;
		
	}

	return TRUE;
}

gboolean shapes_expose(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	draw_shape(cr, sx, sy, ex, ey);

	return TRUE;
}


