/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
	TODO: Option to draw opaque box under text
*/
#include <glib/gi18n.h>
#include "tools.h"
#include "text.h"
#include "dibuja.h"
#include "util.h"
#include "draw-funcs.h"
#include "file.h"
#include "undo.h"

static gboolean text_button_press(GtkWidget *widget, GdkEventButton *event, gpointer data);

static GtkWidget *text_dialog = NULL;
static int startx, starty;

static DrawTool tool = {
				text_button_press,
				NULL,
				NULL,
				NULL,
				NULL,
				NULL,
				NULL
			};

DrawTool *text_get_tool(void)
{
	return &tool;
}

void check_button_meme_text_toggled(GtkToggleButton *toggle, gpointer data)
{
	int active = gtk_toggle_button_get_active(toggle);
	GtkWidget *fill = get_widget("use_fill_prop");

	if(active){
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fill), FALSE);
		gtk_widget_set_sensitive(fill, FALSE);
	}
	else {
		gtk_widget_set_sensitive(fill, TRUE);
	}
}

static void text_set_textview_font(GtkFontButton *widget)
{
	GtkWidget *textview = get_widget("textview");
	PangoFontDescription *desc;
	char *fontname = NULL;

	fontname = (gchar *)gtk_font_button_get_font_name( GTK_FONT_BUTTON(widget) );
	if(fontname){
		desc = pango_font_description_from_string (fontname);
		gtk_widget_modify_font(textview, desc);
		pango_font_description_free (desc);
	}
}

void on_fontbutton1_font_set(GtkFontButton *widget, gpointer user_data)
{
	text_set_textview_font(widget);
}

void textview1_realize(GtkWidget *widget)
{
	GtkWidget *fontbutton = get_widget("fontbutton");
	gtk_widget_realize(fontbutton);
	text_set_textview_font(GTK_FONT_BUTTON(fontbutton));
}

static void text_do_effects(cairo_t *cr, PangoLayout *layout, int x, int y,
							ShapeFillType mode, double line_width)
{
	guint16 alpha = alpha;
	SCColor color;

	if(mode != SHAPE_FILL_NONE){
		cairo_move_to(cr, x, y);
		if( SHAPE_FILL_BACK == mode ){
			dibuja_get_bg_sccolor(&color);
		}
		else if(SHAPE_FILL_FORE == mode ){
			dibuja_get_fg_sccolor(&color);
		}

		cairo_set_source_rgba(cr, color.r, color.g, color.b, color.a);

		pango_cairo_update_layout(cr, layout);
		pango_cairo_show_layout (cr, layout);
	}
	cairo_new_path(cr);
	cairo_move_to(cr, x, y);
	cairo_set_line_width(cr, line_width);

	//cairo_set_line_cap(cr, CAIRO_LINE_CAP_SQUARE);
	cairo_set_line_join(cr, CAIRO_LINE_JOIN_BEVEL);

	dibuja_get_fg_sccolor(&color);
	cairo_set_source_rgba(cr, color.r, color.g, color.b, color.a);
	pango_cairo_update_layout(cr, layout);
	pango_cairo_layout_path(cr, layout);

	cairo_stroke_preserve(cr);
}

static void text_meme_text(cairo_t *cr, PangoLayout *layout, int x, int y)
{
	text_do_effects(cr, layout, x, y, SHAPE_FILL_NONE, (double)dibuja_get_line_width());
}

static void text_show(cairo_t *cr, PangoLayout *layout, int x, int y)
{
	cairo_move_to(cr, x, y);

	pango_cairo_show_layout (cr, layout);
}

static void run_text_dialog(GtkWidget *canvas)
{
	gint result;
	GtkWidget *dialog, *fontbutton, *textview;
	cairo_t *cr;
	gchar *text;
	gboolean use_effects = 0;
	gboolean vcenter, hcenter;

	dialog = get_widget("text-dialog");
	result = gtk_dialog_run (GTK_DIALOG (dialog));
	if(GTK_RESPONSE_OK == result){
		fontbutton = get_widget("fontbutton");
		textview = get_widget("textview");
		text = text_view_get_text(GTK_TEXT_VIEW(textview), 0);

		if(text){
			GdkRectangle area;
			GdkWindow *window;
			gchar *fontname = NULL;
			PangoLayout *layout;
			PangoFontDescription *desc;

			window = gtk_text_view_get_window(GTK_TEXT_VIEW(textview), GTK_TEXT_WINDOW_TEXT);
			gdk_drawable_get_size(window, &area.width, &area.height);

			cr = dibuja_get_cr(CR_SURFACE);
			fontname = (char *)gtk_font_button_get_font_name(GTK_FONT_BUTTON(fontbutton));

			layout = pango_cairo_create_layout (cr);
			pango_layout_set_wrap(layout, PANGO_WRAP_CHAR);
			pango_layout_set_text (layout, text, -1);

			desc = pango_font_description_from_string (fontname);
			pango_layout_set_font_description (layout, desc);
			pango_font_description_free (desc);

			pango_layout_get_size(layout, &area.width, &area.height);
			area.width = PANGO_PIXELS(area.width);
			area.height = PANGO_PIXELS(area.height);

			vcenter = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("center_vert")));
			if(vcenter){
				starty = (dibuja_get_height() / 2) - area.height / 2;
			}

			hcenter = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("center_horz")));
			if(hcenter){
				startx = (dibuja_get_width() / 2) - area.width / 2;
			}

			area.x = startx;
			area.y = starty;

			resize_rect(&area, 5 + dibuja_get_line_width(), 10 + dibuja_get_line_width());
			undo_add(SIMPLE_CANVAS(canvas), area, _("Text"));

			use_effects = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("use_fill_prop")));
			if(use_effects){
				text_do_effects(cr, layout, startx, starty, dibuja_get_fill_mode(), (double)dibuja_get_line_width());
			}
			else if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("meme-text")))){
				SCColor fg, bg;

				cairo_save(cr);
				text_meme_text(cr, layout, startx, starty);
				cairo_restore(cr);

				dibuja_get_bg_sccolor(&bg);
				dibuja_get_fg_sccolor(&fg);

				cairo_save(cr);
				cairo_set_source_rgb(cr, bg.r, bg.g, bg.b);
				pango_cairo_update_layout(cr, layout);
				text_show(cr, layout, startx, starty);
				cairo_set_source_rgb(cr, fg.r, fg.g, fg.b);
				cairo_restore(cr);
			}
			else{
				text_show(cr, layout, startx, starty);
			}

			g_free(text);
			g_object_unref (layout);
			cairo_destroy(cr);

		}
	}

	gtk_widget_hide (dialog);
}

static gboolean text_button_press(GtkWidget *canvas, GdkEventButton *event, gpointer data)
{
	//GtkWidget *dialog = get_widget("text-dialog");
	startx = (int)event->x;
	starty = (int)event->y;

	if(1 == event->button){
		if(!GTK_IS_WIDGET(text_dialog)){
			text_dialog = get_widget("text-dialog");
			gtk_window_set_default_size(GTK_WINDOW(text_dialog), 350, 350);
		}
		if(!gtk_widget_get_visible(text_dialog))
			run_text_dialog(GTK_WIDGET(canvas));
	}
	return TRUE;
}

