/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "tools.h"
#include "lines.h"
#include "brushes.h"
#include "shapes.h"
#include "text.h"
#include "selection.h"
#include "flood-fill.h"
#include "polyline.h"
#include "curve.h"
#include "dibuja.h"
#include "util.h"
#include "pencil.h"


void tool_change(GtkToggleButton *togglebutton, gpointer user_data);

static DrawTool draw_tool = {0};
static GtkWidget *sel_togglebutton; // Used for pasting from clipboard
static GtkWidget *line_togglebutton; // Used for resetting
static GtkWidget *brush_togglebutton; // 

DrawTool *tool_get(void)
{
	return &draw_tool;
}
// Give each tool an id
void tool_init(GtkBuilder *builder)
{
	GtkWidget *button;
	char name[40];
	gint i, *id = NULL;
	int ntools = TOOL_TYPE_LAST;

	for(i = 0; i < ntools; i++){
		sprintf(name, "drawtool%d", i);
		button = GTK_WIDGET (gtk_builder_get_object (builder, name));
		id = g_malloc(sizeof(gint));
		*id = i;
		g_object_set_data_full(G_OBJECT(button), "id", (gpointer)id, (GDestroyNotify)g_free);
		g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(tool_change), NULL);
		if(TOOL_TYPE_SELECTION == (ToolType)i){
			sel_togglebutton = button;
		}
		if(TOOL_TYPE_LINES == (ToolType)i){
			line_togglebutton = button;
		}
		if(TOOL_TYPE_BRUSH == (ToolType)i){
			brush_togglebutton = button;
		}
	}

	draw_tool = *line_get_tool();

	brush_init(builder);
	curve_init(dibuja_get_canvas());
}

void tool_change(GtkToggleButton *togglebutton, gpointer user_data)
{
	int *id = NULL;
	DrawTool prev_tool;

	if(gtk_toggle_button_get_active(togglebutton)){
		id = (int *)g_object_get_data(G_OBJECT(togglebutton), "id");

		if(!id) return;

		prev_tool = draw_tool;
		switch(*id){
			case TOOL_TYPE_LINES:
				draw_tool = *line_get_tool();
				util_set_tool_page(*id, 0);
				break;
			case TOOL_TYPE_RECTANGLE:
				shape_set_type(SHAPE_RECTANGLE);
				draw_tool = *shapes_get_tool();
				util_set_tool_page(*id, 2);
				break;
			case TOOL_TYPE_ELLIPSE:
				shape_set_type(SHAPE_ELLIPSE);
				draw_tool = *shapes_get_tool();
				util_set_tool_page(*id, 2);
				break;
			case TOOL_TYPE_ROUNDED_RECTANGLE:
				shape_set_type(SHAPE_ROUNDED_RECTANGLE);
				draw_tool = *shapes_get_tool();
				util_set_tool_page(*id, 2);
				break;
			case TOOL_TYPE_PENCIL:
				//brush_set_type(BRUSH_PENCIL);
				//brush_pencil_activate();
				//draw_tool = *brush_get_tool();
				draw_tool = *pencil_get_tool();
				util_set_tool_page(*id, 3);
				break;
			case TOOL_TYPE_TEXT:
				draw_tool = *text_get_tool();
				util_set_tool_page(*id, 2);
				break;
			case TOOL_TYPE_BRUSH:
				//brush_set_type(BRUSH_ELLIPSE);
				//brush_activate();
				draw_tool = *brush_get_tool();
				util_set_tool_page(*id, 1);
				break;
			case TOOL_TYPE_SELECTION:
				draw_tool = *sel_get_tool();
				util_set_tool_page(*id, 4);
				break;
			case TOOL_TYPE_FLOODFILL:
				draw_tool = *flood_fill_get_tool();
				util_set_tool_page(*id, 5);
				break;
			case TOOL_TYPE_POLYLINE:
				draw_tool = *polyline_get_tool();
				util_set_tool_page(*id, 2);
				break;
			case TOOL_TYPE_CURVE:
				draw_tool = *curve_get_tool();
				util_set_tool_page(*id, 0);
				break;
			default:
				return;
		}

		if(prev_tool.finish){
			prev_tool.finish();
		}
	}
}
// Called to activate for pasting text from clipboard 
void tools_toggle_sel_button(gboolean is_active)
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(sel_togglebutton), is_active);
}

void tools_toggle_brush_button(gboolean is_active)
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(brush_togglebutton),
								 is_active);
}

void tools_toggle_line_button(gboolean is_active)
{
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(line_togglebutton), is_active);
}

void adj_line_width_value_changed(GtkAdjustment *adjustment)
{
	dibuja_set_line_width((int)gtk_adjustment_get_value(adjustment));
}

void adj_brush_width_value_changed(GtkAdjustment *adjustment)
{
	brush_set_width(gtk_adjustment_get_value(adjustment));
}

void adj_brush_spacing_value_changed(GtkAdjustment *adjustment)
{
	brush_set_spacing(gtk_adjustment_get_value(adjustment));
}

void on_rb_no_fill_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
	if(gtk_toggle_button_get_active(togglebutton)){
		shape_set_fill_type(SHAPE_FILL_NONE);
	}
}

void on_rb_fore_fill_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
	if(gtk_toggle_button_get_active(togglebutton)){
		shape_set_fill_type(SHAPE_FILL_FORE);
	}
}

void on_rb_back_fill_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
	if(gtk_toggle_button_get_active(togglebutton)){
		shape_set_fill_type(SHAPE_FILL_BACK);
	}
}

void on_check_button_antialias_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
	if(gtk_toggle_button_get_active(togglebutton)){
		dibuja_set_antialias(TRUE);
	}
	else{
		dibuja_set_antialias(FALSE);
	}
}

gboolean tools_get_draw_arrow(void)
{
	return gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_widget("check_button_use_arrow")));
}

