/*
 * undo-linear.c
 * Copyright (C)  2012 - 2018 Juan Balderas
 * 
 * undo-linear.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * undo-linear.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Single user linear undo
 */

#include "undo-linear.h"

enum {
	UNDOING,
	REDOING,
	UNDO_DELETE,
	REDO_DELETE,
	UNDO_MAX_DELETE,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

struct _UndoLinearPrivate
{
	GList *undo;
	GList *redo;
	guint undo_max;
	guint undo_cnt;
};

#define UNDO_LINEAR_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), UNDO_TYPE_LINEAR, UndoLinearPrivate))


G_DEFINE_TYPE (UndoLinear, undol, G_TYPE_OBJECT);

static
void undol_delete_last(UndoLinear *ul)
{
	GList *last = NULL;

	g_return_if_fail(UNDO_IS_LINEAR(ul));

	last = g_list_last(ul->priv->undo);
	g_signal_emit (ul, signals[UNDO_MAX_DELETE], 0, last->data);
	ul->priv->undo = g_list_remove_link(ul->priv->undo, last);

	g_list_free(last);
}

static
void undol_destroy(UndoLinear *ul)
{
	GList *p;
	int n = 0;

	g_return_if_fail(UNDO_IS_LINEAR(ul));

	p = ul->priv->undo;

	while(p){
		if(p->data){
			g_signal_emit (ul, signals[UNDO_DELETE], 0, p->data);
			n++;
		}
		p = p->next;
	}
	g_list_free(ul->priv->undo);
	ul->priv->undo = NULL;
}

static
void redo_destroy(UndoLinear *ul)
{
	GList *p;
	int n = 0;

	g_return_if_fail(UNDO_IS_LINEAR(ul));

	p = ul->priv->redo;

	while(p){
		if(p->data){
			g_signal_emit (ul, signals[REDO_DELETE], 0, p->data);
			n++;
		}
		p = p->next;
	}
	g_list_free(ul->priv->redo);
	ul->priv->redo = NULL;

}

static void
undol_init (UndoLinear *ul)
{
	ul->priv = G_TYPE_INSTANCE_GET_PRIVATE (ul, UNDO_TYPE_LINEAR, UndoLinearPrivate);

	ul->priv->undo_max = UNDO_MAX_DEFAULT;
	ul->priv->undo_cnt = 0;
	ul->priv->undo = NULL;
	ul->priv->redo = NULL;

}

static void
undol_dispose (GObject *object)
{
	UndoLinear *ul = UNDO_LINEAR(object);
g_print("undol_dispose\n");
	if(UNDO_IS_LINEAR(ul)){
		undol_destroy(ul);
		redo_destroy(ul);
	}

	G_OBJECT_CLASS (undol_parent_class)->dispose (object);
}

static void
undol_finalize (GObject *object)
{
	G_OBJECT_CLASS (undol_parent_class)->finalize (object);
}

static void
undol_class_init (UndoLinearClass *klass)
{
	GObjectClass* object_class;

	object_class = G_OBJECT_CLASS (klass);

	g_type_class_add_private (object_class, sizeof (UndoLinearPrivate));

	object_class->finalize = undol_finalize;
	object_class->dispose = undol_dispose;

	signals[UNDOING] =
		g_signal_new ("undoing",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (UndoLinearClass, undol_undoing),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

	signals[REDOING] =
		g_signal_new ("redoing",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (UndoLinearClass, undol_redoing),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

	signals[UNDO_DELETE] =
		g_signal_new ("undo-delete",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (UndoLinearClass, undol_undo_delete),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

	signals[REDO_DELETE] =
		g_signal_new ("redo-delete",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (UndoLinearClass, undol_redo_delete),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

	signals[UNDO_MAX_DELETE] =
		g_signal_new ("undo-max-delete",
					  G_OBJECT_CLASS_TYPE (object_class),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (UndoLinearClass, undol_undo_max_delete),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__POINTER,
					  G_TYPE_NONE,
					  1,
					  G_TYPE_POINTER );

}

UndoLinear *
undol_new (void)
{
	return UNDO_LINEAR(g_object_new (UNDO_TYPE_LINEAR, NULL));
}

gboolean
undol_can_undo(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), FALSE);
	return (ul->priv->undo != NULL);
}

gboolean
undol_can_redo(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), FALSE);
	return (ul->priv->redo != NULL);
}

gboolean
undol_add(UndoLinear *ul, gpointer data)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), FALSE);
	g_return_val_if_fail(data, FALSE);

	if(ul->priv->redo){
		redo_destroy(ul);
		ul->priv->redo = NULL;
	}

	ul->priv->undo = g_list_prepend(ul->priv->undo, data);

	ul->priv->undo_cnt++;
	if(ul->priv->undo_cnt > ul->priv->undo_max){
		undol_delete_last(ul);
		ul->priv->undo_cnt--;
	}

	return TRUE;
}

gboolean
undol_do_undo(UndoLinear *ul)
{
	GList *first = NULL;

	g_return_val_if_fail(UNDO_IS_LINEAR(ul), FALSE);
	if(!undol_can_undo(ul)) { 
		return FALSE;
	}

	first = g_list_first(ul->priv->undo);
	ul->priv->undo = g_list_remove_link(ul->priv->undo, first);

	g_signal_emit (ul, signals[UNDOING], 0, first->data);

	ul->priv->redo = g_list_prepend(ul->priv->redo, first->data);
	g_list_free(first);
	first = NULL;

	ul->priv->undo_cnt--;

	return TRUE;
}

gboolean
undol_do_redo(UndoLinear *ul)
{
	GList *first = NULL;

	g_return_val_if_fail(UNDO_IS_LINEAR(ul), FALSE);

	if(!undol_can_redo(ul)){
		return FALSE;
	}

	first = g_list_first(ul->priv->redo);
	ul->priv->redo = g_list_remove_link(ul->priv->redo, first);

	g_signal_emit (ul, signals[REDOING], 0, first->data);

	ul->priv->undo = g_list_prepend(ul->priv->undo, first->data);

	g_list_free(first);
	first = NULL;

	ul->priv->undo_cnt++;
	if(ul->priv->undo_cnt > ul->priv->undo_max){
		undol_delete_last(ul);
		ul->priv->undo_cnt--;
	}

	return TRUE;
}

guint
undol_get_undo_cnt(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), 0);
	return ul->priv->undo_cnt;
}

guint
undol_get_undo_max(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), 0);
	return ul->priv->undo_max;
}

void
undol_set_undo_max(UndoLinear *ul, guint max)
{
	g_return_if_fail(UNDO_IS_LINEAR(ul));
	if(max > 1){
		ul->priv->undo_max = max;
	}
}

void
undol_clear(UndoLinear *ul)
{
	if(UNDO_IS_LINEAR(ul)){
		undol_destroy(ul);
		redo_destroy(ul);
		ul->priv->undo_cnt = 0;
	}
}

/* Get data from top of undo stack */
gpointer
undol_get_undo_data(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), NULL);
	if(ul->priv->undo){
		return ul->priv->undo->data;
	}
	return NULL;
}

/* Get data from top of redo stack */
gpointer
undol_get_redo_data(UndoLinear *ul)
{
	g_return_val_if_fail(UNDO_IS_LINEAR(ul), NULL);
	if(ul->priv->redo){
		return ul->priv->redo->data;
	}
	return NULL;
}




















