/* 
 * (c) 2014 - 2018 Juan Balderas
 * This file is part of Dibuja.

 * Dibuja is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Dibuja is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Dibuja.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <glib/gi18n.h>
#include <string.h>
#include "undo.h"
#include "dibuja.h"
#include "util.h"
#include "menu.h"
#include "file.h"


typedef struct{
	cairo_surface_t *surface;
	GdkRectangle rect;
	gchar *desc;
}UndoData;

static const gchar *swapable[] = {
						N_("Crop"), N_("Resize"), N_("Flip"),
						N_("Flip Horizontally"), N_("Flip Vertically"),
						N_("Rotate"), N_("Scale"), NULL
					};

void undo_undoing(UndoLinear *ul, gpointer undo_data);
void undo_redoing(UndoLinear *ul, gpointer undo_data);
void undo_delete(UndoLinear *ul, gpointer undo_data);
void redo_delete(UndoLinear *ul, gpointer undo_data);
void undo_destroy(UndoLinear *undo);

static gboolean undo_surf_is_swappable(char *desc)
{
	int i = 0;

	if(!desc) return 0;
	while(swapable[i]){
		if(strstr(_(swapable[i]), desc))
			return TRUE;
		i++;
	}
	return 0;
}

void undo_update_tb_buttons(SimpleCanvas *canvas)
{
	GtkWidget *button;
	UndoLinear *undo = NULL;
	gboolean ret;

	if(!SIMPLE_IS_CANVAS(canvas)) return;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return;

	button = get_widget("tb_undo");
	ret = undol_can_undo(undo);

	gtk_widget_set_sensitive(button, ret);

	button = get_widget("tb_redo");
	ret = undol_can_redo(undo);

	gtk_widget_set_sensitive(button, ret);
}

int undo_init(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return 0;

	undo = undol_new();
	if(!UNDO_IS_LINEAR(undo)) return 0;

	g_object_set_data(G_OBJECT(undo), "canvas", (gpointer)canvas);
	g_object_set_data_full(G_OBJECT(canvas), "undo-obj",
										(gpointer)undo, (GDestroyNotify)undo_destroy);
	g_signal_connect(G_OBJECT(undo), "undoing", G_CALLBACK(undo_undoing), NULL);
	g_signal_connect(G_OBJECT(undo), "redoing", G_CALLBACK(undo_redoing), NULL);
	g_signal_connect(G_OBJECT(undo), "undo-delete", G_CALLBACK(undo_delete), NULL);
	g_signal_connect(G_OBJECT(undo), "redo-delete", G_CALLBACK(redo_delete), NULL);
	// Use same as "undo-delete"
	g_signal_connect(G_OBJECT(undo), "undo-max-delete", G_CALLBACK(undo_delete), NULL);

	undol_set_undo_max(undo, UNDO_DEF_MAX);

	return 1;
}

void undo_destroy(UndoLinear *undo)
{
	if(UNDO_IS_LINEAR(undo)){
		g_object_unref(undo);
		undo = NULL;
	}
}

void undo_update_undo_tb_tooltip(char *desc)
{
	GtkWidget *button = get_widget("tb_undo");
	char *tooltip = NULL;

	if(!button) return;

	if(desc)
		tooltip = g_strdup_printf("%s%s", _("Undo "), desc);

	gtk_widget_set_tooltip_text(button, tooltip);

	if(tooltip) g_free(tooltip);
}

void undo_update_redo_tb_tooltip(char *desc)
{
	GtkWidget *button = get_widget("tb_redo");
	char *tooltip = NULL;

	if(!button) return;

	if(desc)
		tooltip = g_strdup_printf("%s%s", _("Redo "), desc);

	gtk_widget_set_tooltip_text(button, tooltip);

	if(tooltip) g_free(tooltip);
}

void undo_update_undo_menu_text(char *desc)
{
	GtkWidget *menuitem;
	gchar *label = NULL;

	menuitem = get_widget("edit_undo");

	if(GTK_IS_MENU_ITEM(menuitem)){
		if(desc){
			/* Undo plus undo action
			   ex Undo %s -> Undo Rectangle */
			label = g_strdup_printf(_("_Undo %s"), desc);
			gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), label);
			g_free(label);
		}
		else{
			gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), _("_Undo"));
		}
	}
}

void undo_update_redo_menu_text(char *desc)
{
	GtkWidget *menuitem;
	gchar *label = NULL;

	menuitem = get_widget("edit_redo");

	if(GTK_IS_MENU_ITEM(menuitem)){
		if(desc){
			/* Undo plus undo action
			   ex Undo %s -> Undo Rectangle */
			label = g_strdup_printf(_("_Redo %s"), desc);
			gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), label);
			g_free(label);
		}
		else{
			gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), _("_Redo"));
		}
	}
}

void undo_update_text(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;
	UndoData *data = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	if(undol_can_undo(undo)){
		data = (UndoData *)undol_get_undo_data(undo);
		if(data){
			undo_update_undo_menu_text(data->desc);
			undo_update_undo_tb_tooltip(data->desc);
		}
	}
	else{
		undo_update_undo_menu_text(NULL);
		undo_update_undo_tb_tooltip(NULL);
	}


	if(undol_can_redo(undo)){
		data = (UndoData *)undol_get_redo_data(undo);
		if(data){
			undo_update_redo_menu_text(data->desc);
			undo_update_redo_tb_tooltip(data->desc);
		}
	}
	else{
		undo_update_redo_menu_text(NULL);
		undo_update_redo_tb_tooltip(NULL);
	}
}

void undo_add(SimpleCanvas *canvas, GdkRectangle rect, gchar *desc)
{
	UndoData *data = NULL;
	UndoLinear *undo = NULL;


	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	data = g_malloc0(sizeof(UndoData));
	if(NULL == data) return;

	data->surface = dibuja_copy_from_surface(&rect);
	if(!data->surface){
		undo_delete(undo, (gpointer)data);
		g_warning("[%s: %d] Could not create image for undo \"%s\"\n",
					__FILE__, __LINE__, desc);
		return;
	}

	data->desc = g_strdup(desc);
	data->rect = rect;

	undol_add(undo, (gpointer)data);

	undo_update_tb_buttons(canvas);
	undo_update_text(canvas);
	file_set_status(file_get_status() | FILE_NEED_SAVE);
}

void undo_add_from_surface(SimpleCanvas *canvas, cairo_surface_t *surface, GdkRectangle rect, gchar *desc)
{
	UndoData *data = NULL;
	UndoLinear *undo = NULL;

	g_return_if_fail(surface != NULL);

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	data = g_malloc0(sizeof(UndoData));
	if(NULL == data) return;

	data->surface = surface;

	data->desc = g_strdup(desc);
	data->rect = rect;
	undol_add(undo, (gpointer)data);

	undo_update_tb_buttons(canvas);
	undo_update_text(canvas);
	file_set_status(file_get_status() | FILE_NEED_SAVE);
}
/* Floating items like selection & curve
   Don't actually add anything, just update the menu & toolbar */
void undo_add_temp(SimpleCanvas *canvas, gchar *desc)
{
	GtkWidget *menuitem;
	char *label = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	menuitem = get_widget("edit_undo");

	label = g_strdup_printf(_("_Undo %s Create"), desc);
	gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), label);
	g_free(label);

	label = g_strdup_printf(_("%s Create"), desc);
	undo_update_undo_tb_tooltip(label);
	g_free(label);

	gtk_widget_set_sensitive(get_widget("tb_undo"), TRUE);
}
/****************************************************************/
void undo_undoing(UndoLinear *ul, gpointer undo_data)
{
	UndoData *data = NULL;
	cairo_surface_t *tmp = NULL;
	GtkWidget *widget;
	cairo_t *cr;

	data = (UndoData *)undo_data;
	if(!data){
		return;
	}

	widget = g_object_get_data(G_OBJECT(ul), "canvas");

	if(undo_surf_is_swappable(data->desc)){
		tmp = simple_canvas_swap_image(SIMPLE_CANVAS(widget), data->surface);
		data->surface = tmp;
	}
	else{
		tmp = data->surface;
		data->surface = dibuja_copy_from_surface(&(data->rect));
		cr = cairo_create(simple_canvas_get_image(SIMPLE_CANVAS(widget)));
		cairo_set_source_surface(cr, tmp, data->rect.x, data->rect.y);
		cairo_paint(cr);
		cairo_destroy(cr);
		cairo_surface_destroy(tmp);
		gdk_window_invalidate_rect(widget->window, NULL, 0);
	}
	undo_update_text(SIMPLE_CANVAS(widget));
	if(!undo_can_undo(SIMPLE_CANVAS(widget))){
		unsigned int status = file_get_status();

		status = status & FILE_NEED_SAVE ? status ^ 0x02 : status;
		file_set_status(status);
		dibuja_set_title();
	}
}

void undo_redoing(UndoLinear *ul, gpointer undo_data)
{
	UndoData *data = NULL;
	cairo_surface_t *tmp = NULL;
	GtkWidget *widget;
	cairo_t *cr;
	guint file_status = 0;

	data = (UndoData *)undo_data;
	if(!data){
		return;
	}

	widget = g_object_get_data(G_OBJECT(ul), "canvas");
	/* TODO: Dont confuse selection w/image */
	if(undo_surf_is_swappable(data->desc)){
		tmp = simple_canvas_swap_image(SIMPLE_CANVAS(widget), data->surface);
		data->surface = tmp;
	}
	else{
		tmp = data->surface;
		data->surface = dibuja_copy_from_surface(&(data->rect));
		cr = cairo_create(simple_canvas_get_image(SIMPLE_CANVAS(widget)));
		cairo_set_source_surface(cr, tmp, data->rect.x, data->rect.y);
		cairo_paint(cr);
		cairo_destroy(cr);
		cairo_surface_destroy(tmp);
		gdk_window_invalidate_rect(widget->window, NULL, 0);
	}
	undo_update_text(SIMPLE_CANVAS(widget));

	file_status = file_get_status();
	if(!(file_status & FILE_NEED_SAVE)){
		file_set_status(file_status | FILE_NEED_SAVE);
		dibuja_set_title();
	}
}

void undo_delete(UndoLinear *ul, gpointer undo_data)
{
	UndoData *data = NULL;

	data = (UndoData *)undo_data;
	if(data){
		if(data->surface){
			cairo_surface_destroy(data->surface);
		}
		if(data->desc){
			g_free(data->desc);
		}
		g_free(data);
		data = NULL;
	}
}

void redo_delete(UndoLinear *ul, gpointer undo_data)
{
	UndoData *data = NULL;

	data = (UndoData *)undo_data;
	if(data){
		if(data->surface){
			cairo_surface_destroy(data->surface);
		}
		if(data->desc){
			g_free(data->desc);
		}
		g_free(data);
		data = NULL;
	}
}

void undo_do_undo(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	/* Clear selection or curve tool */
	if(dibuja_tool_cancel()){
		GtkWidget *menuitem;
		char *label = NULL;
		UndoData *data = NULL;

		menuitem = get_widget("edit_undo");

		data = undol_get_undo_data(undo);
		if(!data){

			gtk_widget_set_sensitive(get_widget("tb_undo"), FALSE);
			gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), _("_Undo"));
			undo_update_undo_tb_tooltip(NULL);
		}
		else{                  
			if(data->desc){
							   /* Undo plus undo action
					              ex Undo %s -> Undo Rectangle */
				label = g_strdup_printf(_("_Undo %s"), data->desc);
				gtk_menu_item_set_label(GTK_MENU_ITEM(menuitem), label);
				undo_update_undo_tb_tooltip(data->desc);
				if(label)
					g_free(label);
			}
		}
		
		gtk_widget_queue_draw(GTK_WIDGET(canvas));
		return;
	}

	undol_do_undo(undo);
	undo_update_text(canvas);
	undo_update_tb_buttons(canvas);
}

void undo_do_redo(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	dibuja_tool_cancel();

	undol_do_redo(undo);
	undo_update_text(canvas);
	undo_update_tb_buttons(canvas);
}
// TODO: Shouldn't need to use "SimpleCanvas *canvas" to call this func
gboolean undo_can_undo(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return FALSE;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return FALSE;

	return undol_can_undo(undo);
}
// TODO: Shouldn't need to use "SimpleCanvas *canvas" to call this func
gboolean undo_can_redo(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return FALSE;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return FALSE;

	return undol_can_redo(undo);
}

void undo_clear(SimpleCanvas *canvas)
{
	UndoLinear *undo = NULL;

	if(!SIMPLE_IS_CANVAS(canvas)) return ;

	undo = g_object_get_data(G_OBJECT(canvas), "undo-obj");
	if(!UNDO_IS_LINEAR(undo)) return ;

	undol_clear(undo);
	undo_update_text(canvas);
	undo_update_tb_buttons(canvas);
}











